# ===============================================================================
# ===============================================================================
#            CropResize 2023-10-28
# ===============================================================================
# ===============================================================================

# To use wrapper functions in the CropResizeWrappers.py script as resizers for CropResize,
# uncomment the following line:

# import CropResizeWrappers as crw

import math as m
import fractions as fr
from functools import partial
import vapoursynth as vs
core = vs.core

# -------------------------------------------------------------------------------

def CropResize(Source, OutWidth=None, OutHeight=None, \
    CL=None, CR=None, CT=None, CB=None, \
    CropDAR=None, TCRatio=None, CSplit=None, \
    CPreview=None, CLine=None, \
    InDAR=None, InSAR=None, OutDAR=None, OutSAR=None, AutoDAR=None, \
    Mod=None, HMod=None, \
    NoResize=None, ResizeWO=None, \
    Borders=None, CleanBorders=None, BColor=None, Frosty=None, \
    ColorConvert=None, ColorMode=None, \
    Resizer=None, RStr=None, RingRepair=None, \
    GMode=None, Position=None, Info=None):

# -------------------------------------------------------------------------------

    kwargsdict = {'Source': [Source, vs.VideoNode],
      'OutWidth': [OutWidth, int], 'OutHeight': [OutHeight, int],
      'CL': [CL, (int, float)], 'CR': [CR, (int, float)],
      'CT': [CT, (int, float)], 'CB': [CB, (int, float)],
      'CropDAR': [CropDAR, (int, float)], 'TCRatio': [TCRatio, (int, float)],
      'CSplit': [CSplit, int], 'CPreview': [CPreview, int], 'CLine': [CLine, int],
      'InDAR': [InDAR, (int, float)], 'InSAR': [InSAR, (int, float)],
      'OutDAR': [OutDAR, (int, float)], 'OutSAR': [OutSAR, (int, float)],
      'AutoDAR': [AutoDAR, bool], 'Mod': [Mod, int], 'HMod': [HMod, int],
      'NoResize': [NoResize, bool], 'ResizeWO': [ResizeWO, bool],
      'Borders': [Borders, bool], 'CleanBorders': [CleanBorders, bool], 'Frosty': [Frosty, bool],
      'ColorConvert': [ColorConvert, str], 'ColorMode': [ColorMode, str],
      'Resizer': [Resizer, str], 'RStr': [RStr, str], 'RingRepair': [RingRepair, (int, float)],
      'GMode': [GMode, int], 'Position': [Position, bool], 'Info': [Info, bool]}

    for k, v in kwargsdict.items():

        VType = "a video" if (k == 'Source') else \
          "an integer or float" \
          if (k == 'CL') or (k == 'CR') or (k == 'CT') or (k == 'CB') or (k == 'CropDAR') or \
          (k == 'TCRatio') or (k == 'InDAR') or (k == 'InSAR') or (k == 'OutDAR') or \
          (k == 'OutSAR') or (k == 'RingRepair') else \
          "an integer" \
          if (k == 'OutWidth') or (k == 'OutHeight') or (k == 'CPreview') or (k == 'CLine') or \
          (k == 'CSplit') or (k == 'Mod') or (k == 'HMod') or (k == 'GMode') else \
          "True or False" \
          if (k == 'AutoDAR') or (k == 'NoResize') or (k == 'ResizeWO') or (k == 'Borders') or \
          (k == 'CleanBorders') or (k == 'Frosty') or (k == 'Position') or (k == 'Info') else \
          "a string"

        if (VType == "an integer or float") or (VType == "an integer"):

            if not ((v[0] is None) or \
              ((isinstance(v[0], bool) is False) and isinstance(v[0], v[1]))):
                raise vs.Error(f"CropResize: {k} must be {VType}")

        else:

            if not ((v[0] is None) or isinstance(v[0], v[1])):
                raise vs.Error(f"CropResize: {k} must be {VType}")

    Source_Width = Source.width
    F_Source_Width = float(Source_Width)
    Source_Height = Source.height
    F_Source_Height = float(Source_Height)

    if (Source_Width == 0) or (Source_Height == 0):
        raise vs.Error('CropResize: Only video with a constant width and height is supported')

# ===============================================================================
#            Globals
# ===============================================================================

    IsGInfo = G_Info if ('G_Info' in globals()) else False
    IsGResizer = True if ('G_Resizer' in globals()) else False

    try: IsPriorG1 = G1_PriorG1
    except: IsPriorG1 = False
    try: IsPriorG2 = G1_PriorG2
    except: IsPriorG2 = False
    try: IsGCPreview = (G_CPreview > 0)
    except: IsGCPreview = False

    GMode = CR_Default(GMode, 1)

    if not (0 <= GMode <= 2):
        raise vs.Error('CropResize: GMode must be 0, 1 or 2 (1 or 2 for GCropResize)')
    if (GMode == 2) and (IsPriorG2 is False):
        raise vs.Error('CropResize: GMode cannot equal 2 for CropResize unless' + \
                       'GMode=2 for a previous global GCropResize instance')

    IsGMode0 = (GMode == 0)
    IsGMode1 = (GMode == 1) and (IsPriorG1 is True)
    IsGMode2 = (GMode == 2) and (IsPriorG2 is True)
    NoGMode = (GMode == 0) or (IsPriorG1 is False)
    IsGModeText = (IsPriorG1 is True) and ((IsGMode1 is False) or (IsPriorG2 is True))

    GInDAR = 0.0 if (NoGMode is True) else G1_InDAR if (IsGMode1 is True) else G2_InDAR
    GInSAR = 0.0 if (NoGMode is True) else G1_InSAR if (IsGMode1 is True) else G2_InSAR
    GOutDAR = 0.0 if (IsPriorG1 is False) else G1_OutDAR if (IsGMode2 is False) else G2_OutDAR
    GOutSAR = 0.0 if (IsPriorG1 is False) else G1_OutSAR if (IsGMode2 is False) else G2_OutSAR

# ===============================================================================
#            Defaults
# ===============================================================================

    OutWidth = CR_Default(OutWidth, 0)
    OutWidth = OutWidth if (IsPriorG1 is False) or (OutWidth != 0) else \
        G1_OutWidth if (IsGMode2 is False) else G2_OutWidth
    OutHeight = CR_Default(OutHeight, 0)
    OutHeight = OutHeight if (IsPriorG1 is False) or (OutHeight != 0) else \
        G1_OutHeight if (IsGMode2 is False) else G2_OutHeight

    UseWidth = (OutWidth == -1)
    TryWidth = (OutWidth < -1)
    OutWidth = OutWidth if (UseWidth is False) and (TryWidth is False) else \
        0 if (UseWidth is True) else -OutWidth
    UseHeight = (OutHeight == -1)
    TryHeight = (OutHeight < -1)
    OutHeight = OutHeight if (UseHeight is False) and (TryHeight is False) else \
        0 if (UseHeight is True) else -OutHeight

    CL = CR_Default(CL, 0.0 if (NoGMode is True) else G1_CL if (IsGMode1 is True) else G2_CL)
    CR = CR_Default(CR, 0.0 if (NoGMode is True) else G1_CR if (IsGMode1 is True) else G2_CR)
    CT = CR_Default(CT, 0.0 if (NoGMode is True) else G1_CT if (IsGMode1 is True) else G2_CT)
    CB = CR_Default(CB, 0.0 if (NoGMode is True) else G1_CB if (IsGMode1 is True) else G2_CB)
    CRTest = ((CL + CR) < Source_Width)
    CBTest = ((CT + CB) < Source_Height)

    CropDAR = CR_Default(CropDAR, 0.0 if (NoGMode is True) else \
        G1_CropDAR if (IsGMode1 is True) else G2_CropDAR)
    TCRatio = CR_Default(TCRatio, 0.5 if (NoGMode is True) else \
        G1_TCRatio if (IsGMode1 is True) else G2_TCRatio)
    CSplit = CR_Default(CSplit, 1 if (NoGMode is True) else \
        G1_CSplit if (IsGMode1 is True) else G2_CSplit)
    CPreview = CR_Default(CPreview, G_CPreview if (IsGCPreview is True) else \
        0 if (IsPriorG1 is False) else G1_CPreview if (IsGMode2 is False) else G2_CPreview)
    CLine = CR_Default(CLine, CR_CLine(Source_Width, Source_Height) if (IsPriorG1 is False) else \
        G1_CLine if (IsGMode2 is False) else G2_CLine)

    IsInDAR = False if (InDAR is None) else True
    IsInSAR = False if (InSAR is None) else True
    IsOutDAR = False if (OutDAR is None) else True
    IsOutSAR = False if (OutSAR is None) else True
    InDAR = CR_Default(InDAR, 0)
    InSAR = CR_Default(InSAR, 0)
    OutDAR = CR_Default(OutDAR, 0)
    OutSAR = CR_Default(OutSAR, 0)
    AutoDAR = CR_Default(AutoDAR, CR_AutoDAR() if (IsPriorG1 is False) else \
        G1_AutoDAR if (IsGMode2 is False) else G2_AutoDAR)

    IsMod = False if (Mod is None) else True
    Mod = CR_Default(Mod, 4 if (IsPriorG1 is False) else G1_Mod if (IsGMode2 is False) else G2_Mod)
    F_Mod = float(Mod)
    HMod = CR_Default(HMod, Mod if (IsMod is True) else \
        Mod if (IsPriorG1 is False) else G1_HMod if (IsGMode2 is False) else G2_HMod)
    F_HMod = float(HMod)

    NoResize = CR_Default(NoResize, False if (NoGMode is True) else \
        G1_NoResize if (IsGMode1 is True) else G2_NoResize)
    ResizeWO = CR_Default(ResizeWO, False if (NoGMode is True) else \
        G1_ResizeWO if (IsGMode1 is True) else G2_ResizeWO)
    Borders = CR_Default(Borders, False if (NoGMode is True) else \
        G1_Borders if (IsGMode1 is True) else G2_Borders)
    CleanBorders = CR_Default(CleanBorders, False if (NoGMode is True) else \
        G1_CleanBorders if (IsGMode1 is True) else G2_CleanBorders)
    BColor = CR_Default(BColor, None if (NoGMode is True) else \
        G1_BColor if (IsGMode1 is True) else G2_BColor)
    Frosty = CR_Default(Frosty, False if (NoGMode is True) else \
        G1_Frosty if (IsGMode1 is True) else G2_Frosty)

    ColorConvert = CR_Default(ColorConvert, "")
    ColorMode = CR_Default(ColorMode, "")

    IsResizer = False if (Resizer is None) else True
    Resizer = CR_Default(Resizer, G_Resizer if (IsGResizer is True) else \
        CR_Resizer_Default() if (IsPriorG1 is False) else \
        G1_Resizer if (IsGMode2 is False) else G2_Resizer).strip()
    RStr = CR_Default(RStr, "" if (IsResizer is True) or (Resizer == "") else \
        G_RStr if (IsGResizer is True) else \
        CR_RStr_Default() if (IsPriorG1 is False) else \
        G1_RStr if (IsGMode2 is False) else G2_RStr).strip()
    RingRepair = CR_Default(RingRepair, 0 if (IsPriorG1 is False) else \
        G1_RingRepair if (IsGMode2 is False) else G2_RingRepair)

    Position = CR_Default(Position, False if (NoGMode is True) else \
        G1_Position if (IsGMode1 is True) else G2_Position)
    Info = CR_Default(Info, G_Info if (IsGInfo is True) else False if (IsPriorG1 is False) else \
        G1_Info if (IsGMode2 is False) else G2_Info)

# ===============================================================================
#            Miscellaneous
# ===============================================================================

    SCFamily = Source.format.color_family
    SFName = Source.format.name
    Source_Bits = Source.format.bits_per_sample
    Source_Sample_Type = Source.format.sample_type
    FRateNum = Source.fps.numerator
    FRateDen = Source.fps.denominator
    F_FRateNum = float(FRateNum)
    F_FRateDen = float(FRateDen)
    FRate = F_FRateNum / F_FRateDen

    def getCSS(w, h):

        css={
            (1, 1): "420",
            (1, 0): "422",
            (0, 0): "444",
            (2, 2): "410",
            (2, 0): "411",
            (0, 1): "440"}
        sub=(w, h)

        if css.get(sub) is None:
            raise vs.Error('CropResize: Unknown video chroma subsampling')
        else: return css[sub]

    ChromaSS = getCSS(Source.format.subsampling_w, Source.format.subsampling_h)

    WidthMod = \
      4 if (ChromaSS == "411") or (ChromaSS == "410") else \
      2 if (ChromaSS == "422") or (ChromaSS == "420") else 1
    HeightMod = \
      4 if (ChromaSS == "410") else \
      2 if (ChromaSS == "440") or (ChromaSS == "420") else 1

    F_WidthMod = float(WidthMod)
    F_HeightMod = float(HeightMod)

    CroppedMoreLeft = (CL > CR)
    CroppedMoreTop = (CT > CB)
    IsScriptCPreview = (0 < CPreview < 4)
    CC = ColorConvert.lower().strip()
    CM = ColorMode.upper().strip()

# ===============================================================================
#            Configure Resizing
# ===============================================================================

    IsResizer = (Resizer != "")
    IsFResize = Resizer.startswith("f_") or Resizer.startswith("F_")

    Resizer = "Spline36" \
      if (IsResizer is False) or ((len(Resizer) == 2) and (IsFResize is True)) else \
      Resizer.removeprefix("f_").removeprefix("F_") if (IsFResize is True) else Resizer

    Kernel = Resizer.lower()

    IsVKernel = (Kernel == "bicubic") or (Kernel == "bilinear") or (Kernel == "lanczos") or \
      (Kernel == "lanczos3") or (Kernel == "lanczos4") or (Kernel == "point") or \
      (Kernel == "spline16") or (Kernel == "spline36") or (Kernel == "spline64")

    IsFKernel = (IsVKernel is True) or (Kernel == "blackman") or (Kernel == "gauss") or \
      (Kernel == "sinc") or (Kernel == "blackmanminlobe") or (Kernel == "box") or \
      (Kernel == "cubic") or (Kernel == "gaussian") or (Kernel == "linear") or \
      (Kernel == "rect") or (Kernel == "spline")

    New_Kernel = "lanczos" if (Kernel == "lanczos3") or (Kernel == "lanczos4") else Kernel

    Resizer = New_Kernel if (IsFResize is True) else \
      New_Kernel.capitalize() if (IsVKernel is True) else Resizer

    RTaps = ", taps=3" if (Kernel == "lanczos3") and (IsFResize is True) else \
      ", filter_param_a=3" if (Kernel == "lanczos3") else \
      ", taps=4" if (Kernel == "lanczos4") and (IsFResize is True) else \
      ", filter_param_a=4" if (Kernel == "lanczos4") else ""

    ResStr = "" if (RStr == "") else ", " + RStr
    NoResizing = (NoResize is True) or (CleanBorders is True)
    ResizeWO = (ResizeWO is True) and (NoResizing is False)
    FullResizing = (NoResizing is False) and (ResizeWO is False)
    IsAutoDAR = (AutoDAR is True) and (FullResizing is False)

# ===============================================================================
#            Insist
# ===============================================================================

# -------------------------------------------------------------------------------
#            OutWidth & OutHeight
# -------------------------------------------------------------------------------

    if ((UseWidth is True) or (TryWidth is True)) and ((UseHeight is True) or (TryHeight is True)):
        raise vs.Error('CropResize: OutWidth & OutHeight cannot be negative at the same time')
    if ((UseWidth is True) or (TryWidth is True) or (UseHeight is True) or (TryHeight is True)) and \
       (FullResizing is False):
        raise vs.Error('CropResize: OutWidth & OutHeight can only be negative in full resizing mode')
    if (OutWidth != 0) and (WidthMod == 2) and (OutWidth % 2 != 0):
        raise vs.Error(f'CropResize: {SFName} requires a mod2 OutWidth')
    if (OutWidth != 0) and (WidthMod == 4) and (OutWidth % 4 != 0):
        raise vs.Error(f'CropResize: {SFName} requires a mod4 OutWidth')
    if (OutHeight != 0) and (HeightMod == 2) and (OutHeight % 2 != 0):
        raise vs.Error(f'CropResize: {SFName} requires a mod2 OutHeight')
    if (OutHeight != 0) and (HeightMod == 4) and (OutHeight % 4 != 0):
        raise vs.Error(f'CropResize: {SFName} requires a mod4 OutHeight')

# -------------------------------------------------------------------------------
#            Cropping
# -------------------------------------------------------------------------------

    if (CL < 0): raise vs.Error('CropResize: CL must be greater than or equal to zero')
    if (CR < 0): raise vs.Error('CropResize: CR must be greater than or equal to zero')
    if (CT < 0): raise vs.Error('CropResize: CT must be greater than or equal to zero')
    if (CB < 0): raise vs.Error('CropResize: CB must be greater than or equal to zero')
    if (CRTest is False):
        raise vs.Error('CropResize: CL + CR must be less than the source width')
    if (CBTest is False):
        raise vs.Error('CropResize: CT + CB must be less than the source height')
    if (CropDAR < 0):
        raise vs.Error('CropResize: CropDAR must be zero (undefined) or positive (16/9 or 1.7778 etc)')
    if not (0.0 <= TCRatio <= 1.0):
        raise vs.Error('CropResize: The valid range for TCRatio is 0.0 to 1.0')
    if (TCRatio != 0.5) and (FullResizing is False):
        raise vs.Error('CropResize: TCRatio cannot be used unless CropResize is in full resizing mode ' + \
                       '(NoResize, CleanBorders & ResizeWO connot be true)')
    if not (0 < CSplit < 4): raise vs.Error('CropResize: CSplit must be 1, 2 or 3')
    if not (1 <= CR_CPreview() <= 3):
        raise vs.Error('CropResize: The CR_CPreview() function must return 1, 2 or 3')
    if not ((0 <= CPreview <= 3) and ((IsGCPreview is False) or (0 <= G_CPreview <= 3))):
        raise vs.Error('CropResize: CPreview must be 0, 1, 2 or 3')
    if (CLine < 0): raise vs.Error('CropResize: CLine must be greater than zero')
    if SFName.endswith("H") and (CPreview > 0):
        raise vs.Error(f'CropResize: Cropping previews don\'t support {SFName} (CPreview must be 0)')

# -------------------------------------------------------------------------------
#            InDAR / InSAR / OutDAR / OutSAR
# -------------------------------------------------------------------------------

    if (InDAR < 0):
        raise vs.Error('CropResize: InDAR must be zero (undefined) or positive (16/9 or 1.7778 etc)')
    if (InSAR < 0):
        raise vs.Error('CropResize: InSAR must be zero (undefined) or positive (16/9 or 1.7778 etc)')
    if (OutDAR < 0):
        raise vs.Error('CropResize: OutDAR must be zero (undefined) or positive (64/45 or 1.4222 etc)')
    if (OutSAR < 0):
        raise vs.Error('CropResize: OutSAR must be zero (undefined) or positive (64/45 or 1.4222 etc)')
    if (InDAR > 0) and (InSAR > 0):
        raise vs.Error('CropResize: InDAR & InSAR cannot be specified together')
    if (OutDAR > 0) and (OutSAR > 0):
        raise vs.Error('CropResize: OutDAR & OutSAR cannot be specified together')

# -------------------------------------------------------------------------------
#            Mod
# -------------------------------------------------------------------------------

    if not ((Mod == 1) or (Mod == 2) or (Mod == 4) or (Mod == 8) or (Mod == 16)):
        raise vs.Error('CropResize: Mod must be 1, 2, 4, 8 or 16')
    if not ((HMod == 1) or (HMod == 2) or (HMod == 4) or (HMod == 8) or (HMod == 16)):
        raise vs.Error('CropResize: Mod must be 1, 2, 4, 8 or 16')
    if (WidthMod == 2) and (Mod % 2 != 0):
        raise vs.Error(f'CropResize: {SFName} requires a minimum of Mod=2')
    if (WidthMod == 4) and (Mod % 4 != 0):
        raise vs.Error(f'CropResize: {SFName} requires a minimum of Mod=4')
    if (HeightMod == 2) and (HMod % 2 != 0):
        raise vs.Error(f'CropResize: {SFName} requires a minimum of HMod=2')
    if (HeightMod == 4) and (HMod % 4 != 0):
        raise vs.Error(f'CropResize: {SFName} requires a minimum of HMod=4')

# -------------------------------------------------------------------------------
#            NoResize / CleanBorders / ResizeWO / Borders
# -------------------------------------------------------------------------------

    if (NoResize is True) and (ResizeWO is True):
        raise vs.Error('CropResize: NoResize & ResizeWO cannot be True at the same time')
    if (CleanBorders is True) and (ResizeWO is True):
        raise vs.Error('CropResize: CleanBorders & ResizeWO cannot be True at the same time')
    if not (((NoResize is False) and (CleanBorders is False)) or \
      (0 == OutDAR == OutSAR == GOutDAR == GOutSAR)):
        raise vs.Error('CropResize: An OutDAR/OutSAR cannot be specified' + \
                       'when NoResize or CleanBorders are True')
    if not ((ResizeWO is False) or (CropDAR == 0) or \
      (((OutDAR == 0) or (OutDAR >= CropDAR)) and ((GOutDAR == 0) or (GOutDAR >= CropDAR)))):
        raise vs.Error('CropResize: When ResizeWO=True, OutDAR must be >= CropDAR')
    if not ((ResizeWO is False) or ((Borders is False) and (Frosty is False)) or \
      (0 == OutDAR == GOutDAR) or (CropDAR > 0)):
        raise vs.Error('CropResize: When ResizeWO=True, ' + \
                    'borders are enabled & OutDAR > 0, a CropDAR is required')
    if not ((ResizeWO is False) or ((Borders is False) and (Frosty is False)) or \
      (0 == OutDAR == GOutDAR) or (OutHeight == 0)):
        raise vs.Error('CropResize: When ResizeWO=True, ' + \
                       'borders are enabled & OutDAR > 0, OutHeight must be zero')

# -------------------------------------------------------------------------------
#            Resizing / Ringing Repair
# -------------------------------------------------------------------------------

    if (IsFResize is True) and (SFName.endswith("H") or \
      (ChromaSS == "440") or (ChromaSS == "410")):
        raise vs.Error(f'CropResize: FMTConv doesn\'t support {SFName}')
    if (IsFResize is True) and (IsFKernel is False):
        raise vs.Error('CropResize: FMTConv has been specified for resizing (F_' + \
                       Resizer + ') but not with a supported resizer')
    if not (0 <= RingRepair <= 1):
        raise vs.Error('CropResize: RingRepair must be a value ' + \
                       'from 0 (disabled) to 1 (maximum repair)')
    if (RingRepair > 0) and ((NoResize is True) or (CleanBorders is True)):
        raise vs.Error('CropResize: RingRepair cannot be enabled ' + \
                       'when either NoResize or CleanBorders are true')
    if SFName.endswith("H") and (RingRepair > 0):
        raise vs.Error(f'CropResize: {SFName} isn\'t supported for RingRepair')

# -------------------------------------------------------------------------------
#            Info / Position
# -------------------------------------------------------------------------------

    if SFName.endswith("H") and (Info is True):
        raise vs.Error('CropResize: The Vapoursynth Text function doesn\'t support ' + \
                       f'{SFName} (Info must be False)')
    if SFName.endswith("H") and (Position is True):
        raise vs.Error('CropResize: The Vapoursynth Text function doesn\'t support ' + \
                       f'{SFName} (Position must be False)')

# ===============================================================================
#            Widths, Heights & Aspect Ratios
# ===============================================================================

    PropsNum = Source.get_frame(0).props.get('_SARNum', 0)
    PropsDen = Source.get_frame(0).props.get('_SARDen', 0)
    PropsSAR = PropsNum / PropsDen if (PropsNum > 0 < PropsDen) else 0

    IsCropDAR = (CropDAR > 0)
    IsGInDAR = (GInDAR > 0) and (IsInDAR is False) and (IsInSAR is False)
    IsGInSAR = (GInSAR > 0) and (IsInDAR is False) and (IsInSAR is False)
    IsGOutDAR = (GOutDAR > 0) and (IsOutDAR is False) and (IsOutSAR is False)
    IsGOutSAR = (GOutSAR > 0) and (IsOutDAR is False) and (IsOutSAR is False)
    IsPropsSAR = (PropsSAR > 0) and (PropsSAR != 1) and \
      (IsInDAR is False) and (IsInSAR is False) and (IsGInDAR is False) and (IsGInSAR is False)
    IsPropsSAR1 = (PropsSAR == 1) and \
      (IsInDAR is False) and (IsInSAR is False) and (IsGInDAR is False) and (IsGInSAR is False)

    IsInDAR = (InDAR > 0)
    IsInSAR = (InSAR > 0)
    IsOutDAR = (OutDAR > 0)
    IsOutSAR = (OutSAR > 0)

    InputDAR = InDAR if (IsInDAR is True) else GInDAR if (IsGInDAR is True) else 0
    InputSAR = InSAR if (IsInSAR is True) else GInSAR if (IsGInSAR is True) else \
      PropsSAR if (IsPropsSAR is True) else 0

    OutputDAR = OutDAR if (IsOutDAR is True) else GOutDAR if (IsGOutDAR is True) else 0
    OutputSAR = OutSAR if (IsOutSAR is True) else GOutSAR if (IsGOutSAR is True) else 0

    IsInputDAR = (InputDAR > 0)
    IsInputSAR = (InputSAR > 0)
    IsOutputDAR = (OutputDAR > 0)
    IsOutputSAR = (OutputSAR > 0)

    IsBordered = (Borders is True) or (Frosty is True) or (CleanBorders is True) or \
      ((IsCropDAR is True) and (IsOutputDAR is True)) or \
      ((FullResizing is True) and (IsCropDAR is True) and (OutWidth > 0 < OutHeight))

    AnamorphicInput = (IsInputDAR is True) or (IsInputSAR is True)

    AnamorphicOutput = ((NoResizing is True) and (AnamorphicInput is True)) or \
      (IsOutputDAR is True) or (IsOutputSAR is True)

    F_Source_Display_Width = F_Source_Width if (AnamorphicInput is False) else \
      InputDAR * F_Source_Height if (IsInputDAR is True) else \
      F_Source_Width * InputSAR

    Source_StorageAR = F_Source_Width / F_Source_Height

    Source_DAR = Source_StorageAR if (AnamorphicInput is False) else \
      InputDAR if (IsInputDAR is True) else \
      F_Source_Width * InputSAR / F_Source_Height

    Source_SAR = 1.0 if (AnamorphicInput is False) else \
      F_Source_Height * InputDAR / F_Source_Width if (IsInputDAR is True) else InputSAR

    F_Cropped_Source_Width = F_Source_Width - CL - CR
    F_Cropped_Source_Height = F_Source_Height - CT - CB

# ===============================================================================
#            No Resizing
# ===============================================================================

    if (NoResizing is True):

        NR_WMod = Mod if (CleanBorders is False) else WidthMod
        NR_HMod = HMod if (CleanBorders is False) else HeightMod
        F_NR_WMod = F_Mod if (CleanBorders is False) else F_WidthMod
        F_NR_HMod = F_HMod if (CleanBorders is False) else F_HeightMod

        F_NR_CL_Adjust = \
          F_WidthMod - (CL % WidthMod) if (CSplit == 1) and (CL % WidthMod != 0) else 0.0
        F_NR_CT_Adjust = \
          F_HeightMod - (CT % HeightMod) if (CSplit == 1) and (CT % HeightMod != 0) else 0.0
        F_NR_CR_Adjust = \
          F_WidthMod - (CR % WidthMod) if (CSplit == 1) and (CR % WidthMod != 0) else 0.0
        F_NR_CB_Adjust = \
          F_HeightMod - (CB % HeightMod) if (CSplit == 1) and (CB % HeightMod != 0) else 0.0

        NR_Mod_Cropped_Width = \
          m.floor((F_Cropped_Source_Width - F_NR_CL_Adjust - F_NR_CR_Adjust) / F_NR_WMod) * NR_WMod
        NR_Mod_Cropped_Height = \
          m.floor((F_Cropped_Source_Height - F_NR_CT_Adjust - F_NR_CB_Adjust) / F_NR_HMod) * NR_HMod

        F_NR_Mod_Cropped_Width = float(NR_Mod_Cropped_Width)
        F_NR_Mod_Cropped_Height = float(NR_Mod_Cropped_Height)

        NR_Mod_Cropped_StorageAR = F_NR_Mod_Cropped_Width / F_NR_Mod_Cropped_Height
        NR_Mod_Cropped_DAR = NR_Mod_Cropped_StorageAR if (AnamorphicInput is False) else \
          F_NR_Mod_Cropped_Width * F_Source_Height * InputDAR / F_NR_Mod_Cropped_Height / F_Source_Width \
          if (IsInputDAR is True) else F_NR_Mod_Cropped_Width * InputSAR / F_NR_Mod_Cropped_Height

        NR_Width = NR_Mod_Cropped_Width
        F_NR_Width = F_NR_Mod_Cropped_Width
        NR_Height = NR_Mod_Cropped_Height
        F_NR_Height = F_NR_Mod_Cropped_Height

        if (IsCropDAR is True) and (CR_IsEqual(CropDAR, NR_Mod_Cropped_DAR) is False):

            if (CropDAR < NR_Mod_Cropped_DAR):

                F_NR_Width = F_NR_Height * CropDAR / Source_SAR
                F_NR_Width_Adjust = F_NR_Width - (m.floor(F_NR_Width / F_NR_WMod) * NR_WMod)
                F_NR_Width_Adjust_Up = F_NR_Width + F_NR_WMod - F_NR_Width_Adjust
                F_NR_Width_Adjust_Down = F_NR_Width - F_NR_Width_Adjust
                F_NR_Width = \
                  F_NR_Width_Adjust_Down if (F_NR_Width_Adjust_Up > F_NR_Mod_Cropped_Width) else \
                  F_NR_Width_Adjust_Up if (F_NR_Width_Adjust >= (F_NR_WMod / 2.0)) else \
                  F_NR_Width_Adjust_Down
                NR_Width = round(F_NR_Width)

            else:

                F_NR_Height = F_NR_Width / CropDAR * Source_SAR
                F_NR_Height_Adjust = F_NR_Height - (m.floor(F_NR_Height / F_NR_HMod) * NR_HMod)
                F_NR_Height_Adjust_Up = F_NR_Height + F_NR_HMod - F_NR_Height_Adjust
                F_NR_Height_Adjust_Down = F_NR_Height - F_NR_Height_Adjust
                F_NR_Height = \
                  F_NR_Height_Adjust_Down if (F_NR_Height_Adjust_Up > F_NR_Mod_Cropped_Height) else \
                  F_NR_Height_Adjust_Up if (F_NR_Height_Adjust > (F_NR_HMod / 2.0)) else \
                  F_NR_Height_Adjust_Down
                NR_Height = round(F_NR_Height)

# -------------------------------------------------------------------------------
#            No Resizing - Cropping
# -------------------------------------------------------------------------------

        F_NR_Extra_Width_Crop = \
          max(0.0, F_Cropped_Source_Width - F_NR_Width - F_NR_CL_Adjust - F_NR_CR_Adjust)
        F_NR_Extra_Height_Crop = \
          max(0.0, F_Cropped_Source_Height - F_NR_Height - F_NR_CT_Adjust - F_NR_CB_Adjust)

        F_NR_Extra_Crop_LR = 0.0 if (CSplit == 1) else \
          max(0.0, F_NR_Extra_Width_Crop / 2.0)
        F_NR_Extra_Crop_LR = 0.0 if (CSplit == 1) else \
          round(F_NR_Extra_Crop_LR) \
          if CR_IsEqual(F_NR_Extra_Crop_LR, round(F_NR_Extra_Crop_LR)) else \
          F_NR_Extra_Crop_LR

        F_NR_Extra_Crop_TB = 0.0 if (CSplit == 1) else \
          max(0.0, F_NR_Extra_Height_Crop / 2.0)
        F_NR_Extra_Crop_TB = 0.0 if (CSplit == 1) else \
          round(F_NR_Extra_Crop_TB) \
          if CR_IsEqual(F_NR_Extra_Crop_TB, round(F_NR_Extra_Crop_TB)) else \
          F_NR_Extra_Crop_TB

        F_NR_Extra_CL = 0.0 if (CSplit != 1) else \
          max(0.0, m.floor(F_NR_Extra_Width_Crop / 8.0) * 4 \
          if (WidthMod == 4) and (CroppedMoreLeft is True) else \
          m.floor(F_NR_Extra_Width_Crop / 4.0) * 2 \
          if (WidthMod == 2) and (CroppedMoreLeft is True) else \
          m.floor(F_NR_Extra_Width_Crop / 2.0) \
          if (CroppedMoreLeft is True) else \
          round(F_NR_Extra_Width_Crop / 8.0) * 4 \
          if (WidthMod == 4) else \
          round(F_NR_Extra_Width_Crop / 4.0) * 2 \
          if (WidthMod == 2) else \
          round(F_NR_Extra_Width_Crop / 2.0))

        F_NR_Extra_CR = max(0.0, F_NR_Extra_Width_Crop - F_NR_Extra_CL) if (CSplit == 1) else 0.0

        F_NR_Extra_CT = 0.0 if (CSplit != 1) else \
          max(0.0, m.floor(F_NR_Extra_Height_Crop / 8.0) * 4 \
          if (HeightMod == 4) and (CroppedMoreTop is True) else \
          m.floor(F_NR_Extra_Height_Crop / 4.0) * 2 \
          if (HeightMod == 2) and (CroppedMoreTop is True) else \
          m.floor(F_NR_Extra_Height_Crop / 2.0) \
          if (CroppedMoreTop is True) else \
          round(F_NR_Extra_Height_Crop / 8.0) * 4 \
          if (HeightMod == 4) else \
          round(F_NR_Extra_Height_Crop / 4.0) * 2 \
          if (HeightMod == 2) else \
          round(F_NR_Extra_Height_Crop / 2.0))

        F_NR_Extra_CB = max(0.0, F_NR_Extra_Height_Crop - F_NR_Extra_CT) if (CSplit == 1) else 0.0

        F_NR_Extra_CL = F_NR_Extra_CL + F_NR_CL_Adjust
        F_NR_Extra_CR = F_NR_Extra_CR + F_NR_CR_Adjust
        F_NR_Extra_CT = F_NR_Extra_CT + F_NR_CT_Adjust
        F_NR_Extra_CB = F_NR_Extra_CB + F_NR_CB_Adjust

        NR_CL = \
          round(max(0, CL + F_NR_Extra_CL if (CSplit == 1) else \
          CL - (CL % WidthMod) if (CSplit == 2) else 0))
        NR_CT = \
          round(max(0, CT + F_NR_Extra_CT if (CSplit == 1) else \
          CT - (CT % HeightMod) if (CSplit == 2) else 0))
        NR_CR = \
          round(max(0, CR + F_NR_Extra_CR if (CSplit == 1) else \
          CR - (CR % WidthMod) if (CSplit == 2) else 0))
        NR_CB = \
          round(max(0, CB + F_NR_Extra_CB if (CSplit == 1) else \
          CB - (CB % HeightMod) if (CSplit == 2) else 0))

        NR_RCL = \
          0.0 if (CSplit == 1) else \
          max(0.0, CL + F_NR_Extra_Crop_LR - NR_CL if (CSplit == 2) else \
          CL + F_NR_Extra_Crop_LR)
        NR_RCT = \
          0.0 if (CSplit == 1) else \
          max(0.0, CT + F_NR_Extra_Crop_TB - NR_CT if (CSplit == 2) else \
          CT + F_NR_Extra_Crop_TB)
        NR_RCR = \
          0.0 if (CSplit == 1) else \
          max(0.0, CR + F_NR_Extra_Crop_LR - NR_CR if (CSplit == 2) else \
          CR + F_NR_Extra_Crop_LR)
        NR_RCB = \
          0.0 if (CSplit == 1) else \
          max(0.0, CB + F_NR_Extra_Crop_TB - NR_CB if (CSplit == 2) else \
          CB + F_NR_Extra_Crop_TB)

        F_NR_Cropped_Width = \
          F_Cropped_Source_Width - F_NR_Extra_CL - F_NR_Extra_CR \
          if (CSplit == 1) else \
          F_Cropped_Source_Width - F_NR_Extra_Width_Crop
        F_NR_Cropped_Height = \
          F_Cropped_Source_Height - F_NR_Extra_CT - F_NR_Extra_CB \
          if (CSplit == 1) else \
          F_Cropped_Source_Height - F_NR_Extra_Height_Crop

# -------------------------------------------------------------------------------
#            No Resizing - Cropped Aspect Ratios
# -------------------------------------------------------------------------------

        NR_Cropped_StorageAR = F_NR_Cropped_Width / F_NR_Cropped_Height
        NR_Cropped_DAR = NR_Cropped_StorageAR if (AnamorphicInput is False) else \
          F_NR_Cropped_Width * F_Source_Height * InputDAR / F_NR_Cropped_Height / F_Source_Width \
          if (IsInputDAR is True) else F_NR_Cropped_Width * InputSAR / F_NR_Cropped_Height

# ===============================================================================
#            Resize Width Only
# ===============================================================================

    elif (ResizeWO is True):

        F_RWO_CT_Adjust = \
          F_HeightMod - (CT % HeightMod) if (CSplit == 1) and (CT % HeightMod != 0) else 0.0
        F_RWO_CB_Adjust = \
          F_HeightMod - (CB % HeightMod) if (CSplit == 1) and (CB % HeightMod != 0) else 0.0

        F_RWO_Resized_Cropped_Width = F_Cropped_Source_Width \
          if (AnamorphicInput is False) else \
          F_Cropped_Source_Width * F_Source_Height * InputDAR / F_Source_Width \
          if (IsInputDAR is True) else F_Cropped_Source_Width * InputSAR

        RWO_Mod_Cropped_Resized_Width = m.floor(F_RWO_Resized_Cropped_Width / F_Mod) * Mod
        F_RWO_Mod_Cropped_Resized_Width = float(RWO_Mod_Cropped_Resized_Width)

        RWO_Mod_Cropped_Height = \
          m.floor((F_Cropped_Source_Height - F_RWO_CT_Adjust - F_RWO_CB_Adjust) / F_HMod) * HMod
        F_RWO_Mod_Cropped_Height = float(RWO_Mod_Cropped_Height)

        RWO_Resized_Cropped_DAR = F_RWO_Resized_Cropped_Width / F_RWO_Mod_Cropped_Height
        RWO_Resized_Mod_Cropped_DAR = F_RWO_Mod_Cropped_Resized_Width / F_RWO_Mod_Cropped_Height
        RWO_Resizing_DAR = CropDAR if (IsCropDAR is True) else \
          OutputDAR if (IsOutputDAR is True) else RWO_Resized_Mod_Cropped_DAR

        RWO_Width = \
          OutWidth if (OutWidth > 0) and (IsOutputDAR is True) else RWO_Mod_Cropped_Resized_Width
        F_RWO_Width = float(RWO_Width)

        RWO_Height = RWO_Mod_Cropped_Height \
          if (RWO_Resizing_DAR <= RWO_Resized_Cropped_DAR) or \
          CR_IsEqual(RWO_Resizing_DAR, RWO_Resized_Cropped_DAR) else \
          m.floor(F_RWO_Resized_Cropped_Width / RWO_Resizing_DAR / F_HMod) * HMod
        F_RWO_Height = float(RWO_Height)

        RWO_Resizing_SAR = 1.0 if (AnamorphicOutput is False) else \
          F_RWO_Height * OutputDAR / F_RWO_Width if (IsOutputDAR is True) else OutputSAR

        F_RWO_Aspect_Width = F_RWO_Height * RWO_Resizing_DAR / RWO_Resizing_SAR
        F_RWO_Aspect_Width = F_RWO_Aspect_Width \
          if CR_IsEqual(F_RWO_Aspect_Width, round(F_RWO_Aspect_Width)) and \
          (round(F_RWO_Aspect_Width) % WidthMod == 0) else float(round(F_RWO_Aspect_Width))

        F_RWO_Width_Adjust = F_RWO_Aspect_Width - (m.floor(F_RWO_Aspect_Width / F_Mod) * Mod)
        F_RWO_Width_Up = F_RWO_Aspect_Width + F_Mod - F_RWO_Width_Adjust
        F_RWO_Width_Down = F_RWO_Aspect_Width - F_RWO_Width_Adjust

        F_RWO_Adjusted_Width = F_RWO_Width_Down if (F_RWO_Width_Up > F_RWO_Width) else \
          F_RWO_Width_Up if ((F_RWO_Width_Up - F_RWO_Aspect_Width) < (F_Mod / 2.0)) else \
          F_RWO_Width_Down
        RWO_Adjusted_Width = round(F_RWO_Adjusted_Width)

        F_RWO_Aspect_Width = \
          F_RWO_Adjusted_Width if (IsOutputDAR is True) and (IsCropDAR is True) else \
          F_RWO_Aspect_Width if (IsOutputDAR is True) else F_RWO_Adjusted_Width

        F_RWO_Resizing_Width = \
          F_RWO_Adjusted_Width if (IsOutputDAR is True) and (IsCropDAR is True) else \
          F_RWO_Width if (IsOutputDAR is True) else F_RWO_Adjusted_Width
        RWO_Resizing_Width = round(F_RWO_Resizing_Width)

# -------------------------------------------------------------------------------
#            Resize Width Only - Cropping
# -------------------------------------------------------------------------------

        F_RWO_Extra_Width_Crop = \
          max(0.0, F_Cropped_Source_Width - (F_RWO_Aspect_Width * RWO_Resizing_SAR) \
          if (AnamorphicInput is False) else \
          F_Cropped_Source_Width - \
          (F_RWO_Aspect_Width * RWO_Resizing_SAR * F_Source_Width / InputDAR / F_Source_Height) \
          if (IsInputDAR is True) else \
          F_Cropped_Source_Width - (F_RWO_Aspect_Width * RWO_Resizing_SAR / InputSAR))

        F_RWO_Extra_Height_Crop = \
          max(0.0, F_Cropped_Source_Height - F_RWO_Height - F_RWO_CT_Adjust - F_RWO_CB_Adjust)

        F_RWO_Extra_Crop_LR = max(0.0, F_RWO_Extra_Width_Crop / 2.0)
        F_RWO_Extra_Crop_LR = round(F_RWO_Extra_Crop_LR) \
          if CR_IsEqual(F_RWO_Extra_Crop_LR, round(F_RWO_Extra_Crop_LR)) else \
          F_RWO_Extra_Crop_LR

        F_RWO_Extra_Crop_TB = 0.0 if (CSplit == 1) else \
          max(0.0, F_RWO_Extra_Height_Crop / 2.0)
        F_RWO_Extra_Crop_TB = 0.0 if (CSplit == 1) else \
          round(F_RWO_Extra_Crop_TB) \
          if CR_IsEqual(F_RWO_Extra_Crop_TB, round(F_RWO_Extra_Crop_TB)) else \
          F_RWO_Extra_Crop_TB

        RWO_CL = \
          round(max(0, CL + F_RWO_Extra_Crop_LR - ((CL + F_RWO_Extra_Crop_LR) % WidthMod) \
          if (CSplit == 1) else \
          CL - (CL % WidthMod) \
          if (CSplit == 2) else 0))
        RWO_CR = \
          round(max(0, CR + F_RWO_Extra_Crop_LR - ((CR + F_RWO_Extra_Crop_LR) % WidthMod) \
          if (CSplit == 1) else \
          CR - (CR % WidthMod) \
          if (CSplit == 2) else 0))

        RWO_RCL = \
          max(0.0, CL + F_RWO_Extra_Crop_LR - RWO_CL \
          if (CSplit == 1) or (CSplit == 2) else \
          CL + F_RWO_Extra_Crop_LR)
        RWO_RCR = \
          max(0.0, CR + F_RWO_Extra_Crop_LR - RWO_CR \
          if (CSplit == 1) or (CSplit == 2) else \
          CR + F_RWO_Extra_Crop_LR)

        F_RWO_Extra_CT = 0.0 if (CSplit != 1) else \
          max(0.0, m.floor(F_RWO_Extra_Height_Crop / 8.0) * 4 \
          if (HeightMod == 4) and (CroppedMoreTop is True) else \
          m.floor(F_RWO_Extra_Height_Crop / 4.0) * 2 \
          if (HeightMod == 2) and (CroppedMoreTop is True) else \
          m.floor(F_RWO_Extra_Height_Crop / 2.0) \
          if (CroppedMoreTop is True) else \
          round(F_RWO_Extra_Height_Crop / 8.0) * 4 \
          if (HeightMod == 4) else \
          round(F_RWO_Extra_Height_Crop / 4.0) * 2 \
          if (HeightMod == 2) else \
          round(F_RWO_Extra_Height_Crop / 2.0))

        F_RWO_Extra_CB = \
          max(0.0, F_RWO_Extra_Height_Crop - F_RWO_Extra_CT) if (CSplit == 1) else 0.0

        F_RWO_Extra_CT = F_RWO_Extra_CT + F_RWO_CT_Adjust
        F_RWO_Extra_CB = F_RWO_Extra_CB + F_RWO_CB_Adjust

        RWO_CT = \
          round(max(0, CT + F_RWO_Extra_CT if (CSplit == 1) else \
          CT - (CT % HeightMod) if (CSplit == 2) else 0))
        RWO_CB = \
          round(max(0, CB + F_RWO_Extra_CB if (CSplit == 1) else \
          CB - (CB % HeightMod) if (CSplit == 2) else 0))

        RWO_RCT = \
          0.0 if (CSplit == 1) else \
          max(0.0, CT + F_RWO_Extra_Crop_TB - RWO_CT if (CSplit == 2) else \
          CT + F_RWO_Extra_Crop_TB)
        RWO_RCB = \
          0.0 if (CSplit == 1) else \
          max(0.0, CB + F_RWO_Extra_Crop_TB - RWO_CB if (CSplit == 2) else \
          CB + F_RWO_Extra_Crop_TB)

        F_RWO_Cropped_Width = F_Cropped_Source_Width - F_RWO_Extra_Width_Crop
        F_RWO_Cropped_Height = \
         F_Cropped_Source_Height - F_RWO_Extra_CT - F_RWO_Extra_CB \
         if (CSplit == 1) else \
         F_Cropped_Source_Height - F_RWO_Extra_Height_Crop

# -------------------------------------------------------------------------------
#            Resize Width Only - Cropped Aspect Ratios
# -------------------------------------------------------------------------------

        RWO_Cropped_StorageAR = F_RWO_Cropped_Width / F_RWO_Cropped_Height
        RWO_Cropped_DAR = RWO_Cropped_StorageAR if (AnamorphicInput is False) else \
          F_RWO_Cropped_Width * F_Source_Height * InputDAR / F_RWO_Cropped_Height / F_Source_Width \
          if (IsInputDAR is True) else F_RWO_Cropped_Width * InputSAR / F_RWO_Cropped_Height

    else:

# ===============================================================================
#            Full Resizing Defaults
# ===============================================================================

        Cropped_Source_DAR = F_Cropped_Source_Width / F_Cropped_Source_Height \
          if (AnamorphicInput is False) else \
          F_Cropped_Source_Width * F_Source_Height * InputDAR / F_Cropped_Source_Height / F_Source_Width \
          if (IsInputDAR is True) else \
          F_Cropped_Source_Width * InputSAR / F_Cropped_Source_Height

        F_Resized_Cropped_Width = F_Cropped_Source_Width \
          if (AnamorphicInput is False) else \
          F_Cropped_Source_Width * F_Source_Height * InputDAR / F_Source_Width \
          if (IsInputDAR is True) else \
          F_Cropped_Source_Width * InputSAR

        Mod_Cropped_Resized_Width = m.floor(F_Resized_Cropped_Width / F_Mod) * Mod
        F_Mod_Cropped_Resized_Width = float(Mod_Cropped_Resized_Width)

        Mod_Cropped_Width = m.floor(F_Cropped_Source_Width / F_Mod) * Mod
        F_Mod_Cropped_Width = float(Mod_Cropped_Width)
        Mod_Cropped_Height = m.floor(F_Cropped_Source_Height / F_HMod) * HMod
        F_Mod_Cropped_Height = float(Mod_Cropped_Height)

        UseWidth = (UseWidth is True) and ((Mod_Cropped_Width - F_Cropped_Source_Width) == 0)
        UseHeight = (UseHeight is True) and ((Mod_Cropped_Height - F_Cropped_Source_Height) == 0)
        TryWidth = (TryWidth is True) and ((OutWidth - F_Cropped_Source_Width) == 0)
        TryHeight = (TryHeight is True) and ((OutHeight - F_Cropped_Source_Height) == 0)
        KeepWidth = (UseWidth is True) or (TryWidth is True)
        KeepHeight = (UseHeight is True) or (TryHeight is True)

        OutWidth = Mod_Cropped_Width if (UseWidth is True) else OutWidth
        OutHeight = Mod_Cropped_Height if (UseHeight is True) else OutHeight
        F_OutWidth = float(OutWidth)
        F_OutHeight = float(OutHeight)

        Resizing_Width = OutWidth if (OutWidth > 0) else Mod_Cropped_Resized_Width
        F_Resizing_Width = F_OutWidth if (OutWidth > 0) else F_Mod_Cropped_Resized_Width
        Resizing_Height = OutHeight if (OutHeight > 0) else Mod_Cropped_Height
        F_Resizing_Height = F_OutHeight if (OutHeight > 0) else F_Mod_Cropped_Height

        Resizing_StorageAR = F_Resizing_Width / F_Resizing_Height
        Resizing_DAR = Cropped_Source_DAR \
          if (IsOutputDAR is True) and (IsBordered is True) and (IsCropDAR is False) else \
          CropDAR if (IsCropDAR is True) else \
          OutputDAR if (IsOutputDAR is True) else Cropped_Source_DAR
        Resizing_SAR = 1.0 if (AnamorphicOutput is False) else \
          F_Resizing_Height * OutputDAR / F_Resizing_Width \
          if (IsOutputDAR is True) else OutputSAR

        TCRatio = 0.5 if (CropDAR == 0) or (CropDAR <= Cropped_Source_DAR) else TCRatio

# ===============================================================================
#            AutoDAR
# ===============================================================================

        IsAutoDAR = (AutoDAR is True) and (IsCropDAR is False) and (IsOutputDAR is False) and \
          (KeepWidth is False) and (KeepHeight is False) and \
          ((IsBordered is True) or ((OutWidth == 0) or (OutHeight == 0)))

        Aspect169 = 1.739 if (IsBordered is False) else 16.0 / 9.0
        Aspect43 = 1.35 if (IsBordered is False) else 4.0 / 3.0
        AspectSAR = 1.0 if (IsOutputSAR is False) else OutputSAR

        IsAspect169 = (AutoDAR is True) and (Aspect169 < Resizing_DAR < 1.829)
        IsAspect43 = (AutoDAR is True) and (1.27 < Resizing_DAR < Aspect43)
        H169 = 1.0 if (AutoDAR is False) else F_Resizing_Width * AspectSAR * 9.0 / 16.0
        H43 = 1.0 if (AutoDAR is False) else F_Resizing_Width * AspectSAR * 3.0 / 4.0
        W169 = 1.0 if (AutoDAR is False) else F_Resizing_Height * 16.0 / 9.0 / AspectSAR
        W43 = 1.0 if (AutoDAR is False) else F_Resizing_Height * 4.0 / 3.0 / AspectSAR

        IsHeight169 = \
          (IsAspect169 is True) and CR_IsEqual(round(H169), H169) and (round(H169) % HMod == 0)
        IsHeight43 = \
          (IsAspect43 is True) and CR_IsEqual(round(H43), H43) and (round(H43) % HMod == 0)
        IsWidth169 = \
          (IsAspect169 is True) and CR_IsEqual(round(W169), W169) and (round(W169) % Mod == 0)
        IsWidth43 = \
          (IsAspect43 is True) and CR_IsEqual(round(W43), W43) and (round(W43) % Mod == 0)

# ===============================================================================
#            Full Resizing (A) Height Calculated From Width
# ===============================================================================

        FullResizeA = True if (IsBordered is True) and \
          (Resizing_DAR >= (Resizing_StorageAR * Resizing_SAR)) else \
          False if (IsBordered is True) else \
          True if (OutWidth > 0) else False if (OutHeight > 0) else \
          True if (IsHeight169 is True) or (IsHeight43 is True) else \
          False if (IsWidth169 is True) or (IsWidth43 is True) else True

        if (FullResizeA is True):

            F_FRA_Aspect_Height = F_Resizing_Width * Resizing_SAR / Resizing_DAR
            F_FRA_Aspect_Height = float(round(F_FRA_Aspect_Height)) \
              if CR_IsEqual(F_FRA_Aspect_Height, round(F_FRA_Aspect_Height)) and \
            (round(F_FRA_Aspect_Height) % HeightMod == 0) else \
            F_FRA_Aspect_Height

            F_FRA_Height_Adjust = F_FRA_Aspect_Height - (m.floor(F_FRA_Aspect_Height / F_HMod) * HMod)
            F_FRA_Height_Up = F_FRA_Aspect_Height + F_HMod - F_FRA_Height_Adjust
            F_FRA_Height_Down = F_FRA_Aspect_Height - F_FRA_Height_Adjust

            F_FRA_Adjusted_Height1 = F_FRA_Height_Down if (KeepWidth is True) else \
              F_FRA_Height_Up if ((F_FRA_Height_Up - F_FRA_Aspect_Height) < (F_HMod / 2.0)) else \
              F_FRA_Height_Down

            F_FRA_Adjusted_Height2 = round(H169) if (IsHeight169 is True) else \
              round(H43) if (IsHeight43 is True) else F_FRA_Adjusted_Height1

            F_FRA_Aspect_Height = F_Resizing_Height \
              if (IsOutputDAR is True) and (IsBordered is True) and \
              (IsCropDAR is False) and (OutHeight == 0) else \
              F_FRA_Adjusted_Height1 \
              if (IsOutputDAR is True) and (IsBordered is True) and (IsCropDAR is False) else \
              F_FRA_Height_Down if (IsOutputDAR is True) and (IsCropDAR is True) else \
              F_FRA_Aspect_Height if (IsOutputDAR is True) else \
              F_FRA_Adjusted_Height2 if (IsBordered is True) or (OutHeight == 0) else \
              F_Resizing_Height

            F_FRA_Resizing_Height = F_FRA_Aspect_Height \
              if (IsOutputDAR is True) and (IsBordered is True) and (IsCropDAR is False) else \
              F_FRA_Height_Down if (IsOutputDAR is True) and (IsCropDAR is True) else \
              F_Resizing_Height if (IsOutputDAR is True) else F_FRA_Aspect_Height

            FRA_Resizing_Height = round(F_FRA_Resizing_Height)

            FRA_Cropping_DAR = F_Resizing_Width * Resizing_SAR / F_FRA_Aspect_Height

            F_FRA_Extra_Width_Crop = 0.0 if CR_IsEqual(FRA_Cropping_DAR, Cropped_Source_DAR) or \
              (FRA_Cropping_DAR > Cropped_Source_DAR) else \
              max(0.0, F_Cropped_Source_Width - \
              (F_Cropped_Source_Height * F_Resizing_Width * Resizing_SAR / F_FRA_Aspect_Height) \
              if (AnamorphicInput is False) else \
              F_Cropped_Source_Width - \
              (F_Cropped_Source_Height * F_Resizing_Width * \
              F_Source_Width * Resizing_SAR / InputDAR / F_FRA_Aspect_Height / F_Source_Height) \
              if (IsInputDAR is True) else \
              F_Cropped_Source_Width - \
              (F_Cropped_Source_Height * F_Resizing_Width * Resizing_SAR / F_FRA_Aspect_Height / InputSAR))

            F_FRA_Extra_Height_Crop = 0.0 if CR_IsEqual(FRA_Cropping_DAR, Cropped_Source_DAR) or \
              (FRA_Cropping_DAR < Cropped_Source_DAR) else \
              max(0.0, F_Cropped_Source_Height - \
              (F_Cropped_Source_Width * F_FRA_Aspect_Height / F_Resizing_Width / Resizing_SAR) \
              if (AnamorphicInput is False) else \
              F_Cropped_Source_Height - \
              (F_Cropped_Source_Width * F_FRA_Aspect_Height * \
              F_Source_Height * InputDAR / F_Resizing_Width / Resizing_SAR / F_Source_Width) \
              if (IsInputDAR is True) else \
              F_Cropped_Source_Height - \
              (F_Cropped_Source_Width * F_FRA_Aspect_Height * InputSAR / F_Resizing_Width / Resizing_SAR))

# -------------------------------------------------------------------------------
#            Full Resizing (A) Cropping
# -------------------------------------------------------------------------------

            F_FRA_Extra_Crop_LR = max(0.0, F_FRA_Extra_Width_Crop / 2.0)
            F_FRA_Extra_Crop_LR = round(F_FRA_Extra_Crop_LR) \
              if CR_IsEqual(F_FRA_Extra_Crop_LR, round(F_FRA_Extra_Crop_LR)) else \
              F_FRA_Extra_Crop_LR

            F_FRA_Extra_Crop_T = max(0.0, F_FRA_Extra_Height_Crop / 2.0 \
              if (TCRatio == 0.5) else F_FRA_Extra_Height_Crop * TCRatio)
            F_FRA_Extra_Crop_T = round(F_FRA_Extra_Crop_T) \
              if CR_IsEqual(F_FRA_Extra_Crop_T, round(F_FRA_Extra_Crop_T)) else \
              F_FRA_Extra_Crop_T

            F_FRA_Extra_Crop_B = max(0.0, F_FRA_Extra_Height_Crop / 2.0 \
              if (TCRatio == 0.5) else F_FRA_Extra_Height_Crop - F_FRA_Extra_Crop_T)
            F_FRA_Extra_Crop_B = round(F_FRA_Extra_Crop_B) \
              if CR_IsEqual(F_FRA_Extra_Crop_B, round(F_FRA_Extra_Crop_B)) else \
              F_FRA_Extra_Crop_B

            FRA_CL = \
              round(max(0, CL + F_FRA_Extra_Crop_LR - ((CL + F_FRA_Extra_Crop_LR) % WidthMod) \
              if (CSplit == 1) else CL - (CL % WidthMod) if (CSplit == 2) else 0))
            FRA_CR = \
              round(max(0, CR + F_FRA_Extra_Crop_LR - ((CR + F_FRA_Extra_Crop_LR) % WidthMod) \
              if (CSplit == 1) else CR - (CR % WidthMod) if (CSplit == 2) else 0))
            FRA_CT = \
              round(max(0, CT + F_FRA_Extra_Crop_T - ((CT + F_FRA_Extra_Crop_T) % HeightMod) \
              if (CSplit == 1) else CT - (CT % HeightMod) if (CSplit == 2) else 0))
            FRA_CB = \
              round(max(0, CB + F_FRA_Extra_Crop_B - ((CB + F_FRA_Extra_Crop_B) % HeightMod) \
              if (CSplit == 1) else CB - (CB % HeightMod) if (CSplit == 2) else 0))

            FRA_RCL = max(0.0, CL + F_FRA_Extra_Crop_LR - FRA_CL \
              if (CSplit == 1) or (CSplit == 2) else CL + F_FRA_Extra_Crop_LR)
            FRA_RCR = max(0.0, CR + F_FRA_Extra_Crop_LR - FRA_CR \
              if (CSplit == 1) or (CSplit == 2) else CR + F_FRA_Extra_Crop_LR)
            FRA_RCT = max(0.0, CT + F_FRA_Extra_Crop_T - FRA_CT \
              if (CSplit == 1) or (CSplit == 2) else CT + F_FRA_Extra_Crop_T)
            FRA_RCB = max(0.0, CB + F_FRA_Extra_Crop_B - FRA_CB \
              if (CSplit == 1) or (CSplit == 2) else CB + F_FRA_Extra_Crop_B)

            F_FRA_Cropped_Width = F_Cropped_Source_Width - F_FRA_Extra_Width_Crop
            F_FRA_Cropped_Height = F_Cropped_Source_Height - F_FRA_Extra_Height_Crop

# -------------------------------------------------------------------------------
#            Full Resizing (A) Cropped Aspect Ratios
# -------------------------------------------------------------------------------

            FRA_Cropped_StorageAR = F_FRA_Cropped_Width / F_FRA_Cropped_Height
            FRA_Cropped_DAR = FRA_Cropped_StorageAR if (AnamorphicInput is False) else \
              F_FRA_Cropped_Width * F_Source_Height * InputDAR / F_FRA_Cropped_Height / F_Source_Width \
              if (IsInputDAR is True) else F_FRA_Cropped_Width * InputSAR / F_FRA_Cropped_Height

        else:

# ===============================================================================
#            Full Resizing (B) Width Calculated From Height
# ===============================================================================

            F_FRB_Aspect_Width = F_Resizing_Height * Resizing_DAR / Resizing_SAR
            F_FRB_Aspect_Width = \
              float(round(F_FRB_Aspect_Width)) \
              if CR_IsEqual(F_FRB_Aspect_Width, round(F_FRB_Aspect_Width)) and \
              (round(F_FRB_Aspect_Width) % WidthMod == 0) else \
              F_FRB_Aspect_Width

            F_FRB_Width_Adjust = F_FRB_Aspect_Width - (m.floor(F_FRB_Aspect_Width / F_Mod) * Mod)
            F_FRB_Width_Up = F_FRB_Aspect_Width + Mod - F_FRB_Width_Adjust
            F_FRB_Width_Down = F_FRB_Aspect_Width - F_FRB_Width_Adjust

            F_FRB_Adjusted_Width1 = F_FRB_Width_Down if (KeepHeight is True) else \
              F_FRB_Width_Up if ((F_FRB_Width_Up - F_FRB_Aspect_Width) < (F_Mod / 2.0)) else \
              F_FRB_Width_Down

            F_FRB_Adjusted_Width2 = round(W169) if (IsWidth169 is True) else \
              round(W43) if (IsWidth43 is True) else F_FRB_Adjusted_Width1

            F_FRB_Aspect_Width = F_Resizing_Width \
              if (IsOutputDAR is True) and (IsBordered is True) and \
              (IsCropDAR is False) and (OutWidth == 0) else \
              F_FRB_Adjusted_Width1 \
              if (IsOutputDAR is True) and (IsBordered is True) and (IsCropDAR is False) else \
              F_FRB_Width_Down if (IsOutputDAR is True) and (IsCropDAR is True) else \
              F_FRB_Aspect_Width if (IsOutputDAR is True) else \
              F_FRB_Adjusted_Width2 if (IsBordered is True) or (OutWidth == 0) else \
              F_Resizing_Width

            F_FRB_Resizing_Width = F_FRB_Aspect_Width \
              if (IsOutputDAR is True) and (IsBordered is True) and (IsCropDAR is False) else \
              F_FRB_Width_Down if (IsOutputDAR is True) and (IsCropDAR is True) else \
              F_Resizing_Width if (IsOutputDAR is True) else F_FRB_Aspect_Width

            FRB_Resizing_Width = round(F_FRB_Resizing_Width)

            FRB_Cropping_DAR = F_FRB_Aspect_Width * Resizing_SAR / F_Resizing_Height

            F_FRB_Extra_Width_Crop = 0.0 if CR_IsEqual(FRB_Cropping_DAR, Cropped_Source_DAR) or \
              (FRB_Cropping_DAR > Cropped_Source_DAR) else \
              max(0.0, F_Cropped_Source_Width - \
              (F_Cropped_Source_Height * F_FRB_Aspect_Width * Resizing_SAR / F_Resizing_Height) \
              if (AnamorphicInput is False) else \
              F_Cropped_Source_Width - \
              (F_Cropped_Source_Height * F_FRB_Aspect_Width * \
              Resizing_SAR * F_Source_Width / InputDAR / F_Resizing_Height / F_Source_Height) \
              if (IsInputDAR is True) else \
              F_Cropped_Source_Width - \
              (F_Cropped_Source_Height * F_FRB_Aspect_Width * Resizing_SAR / F_Resizing_Height / InputSAR))

            F_FRB_Extra_Height_Crop = 0.0 if CR_IsEqual(FRB_Cropping_DAR, Cropped_Source_DAR) or \
              (FRB_Cropping_DAR < Cropped_Source_DAR) else \
              max(0.0, F_Cropped_Source_Height - \
              (F_Cropped_Source_Width * F_Resizing_Height / F_FRB_Aspect_Width / Resizing_SAR) \
              if (AnamorphicInput is False) else \
              F_Cropped_Source_Height - \
              (F_Cropped_Source_Width * F_Resizing_Height * \
              F_Source_Height * InputDAR / F_FRB_Aspect_Width / Resizing_SAR / F_Source_Width) \
              if (IsInputDAR is True) else \
              F_Cropped_Source_Height - \
              (F_Cropped_Source_Width * F_Resizing_Height * InputSAR / F_FRB_Aspect_Width / Resizing_SAR))

# -------------------------------------------------------------------------------
#            Full Resizing (B) Cropping
# -------------------------------------------------------------------------------

            F_FRB_Extra_Crop_LR = max(0.0, F_FRB_Extra_Width_Crop / 2.0)
            F_FRB_Extra_Crop_LR = round(F_FRB_Extra_Crop_LR) \
              if CR_IsEqual(F_FRB_Extra_Crop_LR, round(F_FRB_Extra_Crop_LR)) else \
              F_FRB_Extra_Crop_LR

            F_FRB_Extra_Crop_T = max(0.0, F_FRB_Extra_Height_Crop / 2.0
              if (TCRatio == 0.5) else F_FRB_Extra_Height_Crop * TCRatio)
            F_FRB_Extra_Crop_T = round(F_FRB_Extra_Crop_T) \
              if CR_IsEqual(F_FRB_Extra_Crop_T, round(F_FRB_Extra_Crop_T)) else \
              F_FRB_Extra_Crop_T

            F_FRB_Extra_Crop_B = max(0.0, F_FRB_Extra_Height_Crop / 2.0 \
              if (TCRatio == 0.5) else F_FRB_Extra_Height_Crop - F_FRB_Extra_Crop_T)
            F_FRB_Extra_Crop_B = round(F_FRB_Extra_Crop_B) \
              if CR_IsEqual(F_FRB_Extra_Crop_B, round(F_FRB_Extra_Crop_B)) else \
              F_FRB_Extra_Crop_B

            FRB_CL = \
              round(max(0, CL + F_FRB_Extra_Crop_LR - ((CL + F_FRB_Extra_Crop_LR) % WidthMod) \
              if (CSplit == 1) else  CL - (CL % WidthMod) if (CSplit == 2) else 0))
            FRB_CR = \
              round(max(0, CR + F_FRB_Extra_Crop_LR - ((CR + F_FRB_Extra_Crop_LR) % WidthMod) \
              if (CSplit == 1) else CR - (CR % WidthMod) if (CSplit == 2) else 0))
            FRB_CT = \
              round(max(0, CT + F_FRB_Extra_Crop_T - ((CT + F_FRB_Extra_Crop_T) % HeightMod) \
              if (CSplit == 1) else CT - (CT % HeightMod) if (CSplit == 2) else 0))
            FRB_CB = \
              round(max(0, CB + F_FRB_Extra_Crop_B - ((CB + F_FRB_Extra_Crop_B) % HeightMod) \
              if (CSplit == 1) else CB - (CB % HeightMod) if (CSplit == 2) else 0))

            FRB_RCL = max(0.0, CL + F_FRB_Extra_Crop_LR - FRB_CL \
              if (CSplit == 1) or (CSplit == 2) else CL + F_FRB_Extra_Crop_LR)
            FRB_RCR = max(0.0, CR + F_FRB_Extra_Crop_LR - FRB_CR \
              if (CSplit == 1) or (CSplit == 2) else CR + F_FRB_Extra_Crop_LR)
            FRB_RCT = max(0.0, CT + F_FRB_Extra_Crop_T - FRB_CT \
              if (CSplit == 1) or (CSplit == 2) else CT + F_FRB_Extra_Crop_T)
            FRB_RCB = max(0.0, CB + F_FRB_Extra_Crop_B - FRB_CB \
              if (CSplit == 1) or (CSplit == 2) else CB + F_FRB_Extra_Crop_B)

            F_FRB_Cropped_Width = F_Cropped_Source_Width - F_FRB_Extra_Width_Crop
            F_FRB_Cropped_Height = F_Cropped_Source_Height - F_FRB_Extra_Height_Crop

# -------------------------------------------------------------------------------
#            Full Resizing (B) Cropped Aspect Ratios
# -------------------------------------------------------------------------------

            FRB_Cropped_StorageAR = F_FRB_Cropped_Width / F_FRB_Cropped_Height
            FRB_Cropped_DAR = FRB_Cropped_StorageAR if (AnamorphicInput is False) else \
              F_FRB_Cropped_Width * F_Source_Height * InputDAR / F_FRB_Cropped_Height / F_Source_Width \
              if (IsInputDAR is True) else F_FRB_Cropped_Width * InputSAR / F_FRB_Cropped_Height

# ===============================================================================
#            Output Width, Height & Aspect Ratios
# ===============================================================================

    CRVideo_Width = \
      NR_Width if (NoResizing is True) else RWO_Resizing_Width if (ResizeWO is True) else \
      Resizing_Width if (FullResizeA is True) else FRB_Resizing_Width

    CRVideo_Height = \
      NR_Height if (NoResizing is True) else RWO_Height if (ResizeWO is True) else \
      FRA_Resizing_Height if (FullResizeA is True) else Resizing_Height

    F_CRVideo_Width = float(CRVideo_Width)
    F_CRVideo_Height = float(CRVideo_Height)
    CRVideo_StorageAR = F_CRVideo_Width / F_CRVideo_Height

    CRVideo_DAR = CRVideo_StorageAR \
      if (NoResizing is True) and (AnamorphicInput is False) else \
      F_CRVideo_Width * F_Source_Height * InputDAR / F_CRVideo_Height / F_Source_Width \
      if (NoResizing is True) and (IsInputDAR is True) else \
      F_CRVideo_Width * Source_SAR / F_CRVideo_Height \
      if (NoResizing is True) else \
      CRVideo_StorageAR \
      if (AnamorphicOutput is False) else \
      F_CRVideo_Width * RWO_Resizing_SAR / F_CRVideo_Height \
      if (ResizeWO is True) else \
      F_CRVideo_Width * Resizing_SAR / F_CRVideo_Height

    Final_Width = Source_Width if (CleanBorders is True) else \
      CRVideo_Width if (IsBordered is False) else \
      RWO_Width if (ResizeWO is True) and (IsOutputDAR is True) else \
      OutWidth if (NoResizing is True) or (ResizeWO is True) and \
      (OutWidth > CRVideo_Width) else \
      Resizing_Width if (FullResizing is True) and \
      (Resizing_StorageAR > CRVideo_StorageAR) and (OutWidth == 0) else \
      OutWidth if (FullResizing is True) and \
      (Resizing_StorageAR > CRVideo_StorageAR) and (OutWidth > CRVideo_Width) \
      else CRVideo_Width

    Final_Height = Source_Height if (CleanBorders is True) else \
      CRVideo_Height if (IsBordered is False) else \
      CRVideo_Height if (ResizeWO is True) and (IsOutputDAR is True) else \
      OutHeight if (NoResizing is True) or (ResizeWO is True) and \
      (OutHeight > CRVideo_Height) else \
      Resizing_Height if (FullResizing is True) and \
      (Resizing_StorageAR < CRVideo_StorageAR) and (OutHeight == 0) else \
      OutHeight if (FullResizing is True) and \
      (Resizing_StorageAR < CRVideo_StorageAR) and (OutHeight > CRVideo_Height) \
      else CRVideo_Height

    F_Final_Width = float(Final_Width)
    F_Final_Height = float(Final_Height)
    Final_StorageAR = F_Final_Width / F_Final_Height

    Final_DAR = Final_StorageAR if (NoResizing is True) and (AnamorphicInput is False) else \
      F_Final_Width * F_Source_Height * InputDAR / F_Final_Height / F_Source_Width \
      if (NoResizing is True) and (IsInputDAR is True) else \
      F_Final_Width * Source_SAR / F_Final_Height if (NoResizing is True) else \
      Final_StorageAR if (AnamorphicOutput is False) else \
      F_Final_Width * RWO_Resizing_SAR / F_Final_Height if (ResizeWO is True) else \
      F_Final_Width * Resizing_SAR / F_Final_Height

    IsAnamorphicInput = \
      (AnamorphicInput is True) and (CR_IsEqual(Source_DAR, Source_StorageAR) is False)
    IsAnamorphicOutput = \
      (AnamorphicOutput is True) and (CR_IsEqual(Final_DAR, Final_StorageAR) is False)

    F_Final_Display_Width = F_Final_Width if (AnamorphicOutput is False) else \
      F_Final_Width * Source_SAR if (NoResizing is True) else \
      F_Final_Width * RWO_Resizing_SAR if (ResizeWO is True) else \
      F_Final_Width * Resizing_SAR

    Final_SAR = 1.0 if (AnamorphicOutput is False) else \
      Source_SAR if (NoResizing is True) else \
      RWO_Resizing_SAR if (ResizeWO is True) else Resizing_SAR


# ===============================================================================
#            Configure Color Conversion
# ===============================================================================

    if (CC != "") or (CM != ""):

        if not ((CC == "") or (CC == "vs") or (CC == "fmtconv")):
            raise vs.Error('CropResize: ColorConvert can only be ' + \
                           '"VS", "FMTConv" or "" (an empty string)')

        try:
            TestFMTConv = Source.fmtc.bitdepth(bits=8)
            IsFMTConv = True
        except:
            IsFMTConv = False

        IsSDSource = (F_Source_Display_Width <= 1056) and (Source_Height < 600)
        IsHDSource = ((F_Source_Display_Width > 1056) or (Source_Height >= 600)) and \
          ((F_Source_Display_Width <= 1920) and (Source_Height <= 1080))
        IsUHDSource = (F_Source_Display_Width > 1920) or (Source_Height > 1080)

        IsSDOut = (F_Final_Display_Width <= 1056) and (Final_Height < 600)
        IsHDOut = ((F_Final_Display_Width > 1056) or (Final_Height >= 600)) and \
          ((F_Final_Display_Width <= 1920) and (Final_Height <= 1080))
        IsUHDOut = (F_Final_Display_Width > 1920) or (Final_Height > 1080)

        CM = CM if (CM != "") else \
          "" if (FullResizing is False) or (SCFamily == vs.GRAY) else \
          "601-709" if (IsSDSource is True) and (IsHDOut is True) else \
          "601-2020" if (IsSDSource is True) and (IsUHDOut is True) else \
          "709-601" if (IsHDSource is True) and (IsSDOut is True) else \
          "709-2020" if (IsHDSource is True) and (IsUHDOut is True) else \
          "2020-601" if (IsUHDSource is True) and (IsSDOut is True) else \
          "2020-709" if (IsUHDSource is True) and (IsHDOut is True) else ""

        Is601NP = (CM == "601N-601P") or (CM == "601P-601N")
        Is601709 = (CM == "709-601") or (CM == "601-709")

        Is601709NP = \
          (CM == "709-601N") or (CM == "601N-709") or \
          (CM == "709-601P") or (CM == "601P-709")
        Is2020 = \
          (CM == "2020-709") or (CM == "709-2020") or \
          (CM == "2020-601") or (CM == "601-2020")
        Is2020NP = \
          (CM == "2020-601N") or (CM == "601N-2020") or \
          (CM == "2020-601P") or (CM == "601P-2020")

        if not ((CM == "") or (Is601NP is True) or (Is601709 is True) or \
          (Is601709NP is True) or (Is2020 is True) or (Is2020NP is True)):
            raise vs.Error('CropResize: ColorMode="' + CM + '" is not a supported color mode')

        CC = CC if (CC != "") else \
          "fmtconv" if (IsFResize is True) or ((IsFMTConv is True) and \
          (ChromaSS != "440") and (ChromaSS != "410")) else "vs"

        CC = "VapourSynth" if (CC == "vs") else "FMTConv" if (CC == "fmtconv") else ""

        IsFResize = (IsFResize is True) or \
          ((CC == "FMTConv") and (CM != "") and (IsFKernel is True) and (RStr == ""))

        Resizer = New_Kernel if (IsFResize is True) else \
          New_Kernel.capitalize() if (IsVKernel is True) else Resizer

        RTaps = ", taps=3" if (Kernel == "lanczos3") and (IsFResize is True) else \
          ", filter_param_a=3" if (Kernel == "lanczos3") else \
          ", taps=4" if (Kernel == "lanczos4") and (IsFResize is True) else \
          ", filter_param_a=4" if (Kernel == "lanczos4") else ""

    IsColorConversion = (CC != "") and (CM != "")

# ===============================================================================
#            Apply Cropping
# ===============================================================================

    CCL = \
      NR_CL if (NoResizing is True) else RWO_CL if (ResizeWO is True) else \
      FRA_CL if (FullResizeA is True) else FRB_CL
    CCR = \
      NR_CR if (NoResizing is True) else RWO_CR if (ResizeWO is True) else \
      FRA_CR if (FullResizeA is True) else FRB_CR
    CCT = \
      NR_CT if (NoResizing is True) else RWO_CT if (ResizeWO is True) else \
      FRA_CT if (FullResizeA is True) else FRB_CT
    CCB = \
      NR_CB if (NoResizing is True) else RWO_CB if (ResizeWO is True) else \
      FRA_CB if (FullResizeA is True) else FRB_CB

    RCL = \
      NR_RCL if (NoResizing is True) else RWO_RCL if (ResizeWO is True) else \
      FRA_RCL if (FullResizeA is True) else FRB_RCL
    RCR = \
      NR_RCR if (NoResizing is True) else RWO_RCR if (ResizeWO is True) else \
      FRA_RCR if (FullResizeA is True) else FRB_RCR
    RCT = \
      NR_RCT if (NoResizing is True) else RWO_RCT if (ResizeWO is True) else \
      FRA_RCT if (FullResizeA is True) else FRB_RCT
    RCB = \
      NR_RCB if (NoResizing is True) else RWO_RCB if (ResizeWO is True) else \
      FRA_RCB if (FullResizeA is True) else FRB_RCB

    F_Cropped_Width = F_Source_Width - CCL - CCR - RCL - RCR
    F_Cropped_Height = F_Source_Height - CCT - CCB - RCT - RCB

    Cropped = Source if (0 == CCL == CCR == CCT == CCB) else \
      core.std.Crop(Source, CCL, CCR, CCT, CCB)

    NoResizer = (NoResizing is True) and \
      ((CSplit == 1) or (0 == NR_RCL == NR_RCR == NR_RCT == NR_RCB))

    UpScale = (NoResizing is False) and \
      ((F_CRVideo_Width * F_CRVideo_Height / F_Cropped_Width / F_Cropped_Height) > 1)

# ===============================================================================
#            Resize
# ===============================================================================

    DoResize = ".fmtc.resample(CRVideo_Width, CRVideo_Height, sx=RCL, sy=RCT, " + \
      "sw=F_Cropped_Width, sh=F_Cropped_Height, kernel=Resizer" + RTaps + ResStr + ")" \
      if (IsFResize is True) else \
      ".resize." + Resizer + "(CRVideo_Width, CRVideo_Height, src_left=RCL, src_top=RCT, " + \
      "src_width=F_Cropped_Width, src_height=F_Cropped_Height" + RTaps + ResStr + ")" \
      if (IsVKernel is True) else \
      Resizer + "(Cropped, CRVideo_Width, CRVideo_Height, src_left=RCL, src_top=RCT, " + \
      "src_width=F_Cropped_Width, src_height=F_Cropped_Height" + ResStr + ")"

    To16Bit = "" if (Source_Bits == 16) else ".fmtc.bitdepth(bits=16)"
    Undo16Bit = "" if (Source_Bits == 16) else ".fmtc.bitdepth(bits=Source_Bits)"

    if (IsColorConversion is False):

        Cropped = Cropped if (RingRepair == 0) else eval("Cropped" + To16Bit)

        Resized = "" if (RingRepair == 0) else \
          eval("Cropped" + DoResize) \
          if (IsFResize is True) or (IsVKernel is True) else eval(DoResize)

        CRVideo = Cropped if (NoResizer is True) else \
          eval("CR_RingRepair(Resized, Cropped, CRVideo_Width, CRVideo_Height, " + \
          "RCL, RCT, F_Cropped_Width, F_Cropped_Height, RingRepair)" + Undo16Bit) \
          if (RingRepair > 0) else \
          eval("Cropped" + To16Bit + DoResize + Undo16Bit) \
          if (IsFResize is True) else \
          eval("Cropped" + DoResize) \
          if (IsVKernel is True) else \
          eval(DoResize)

# ===============================================================================
#            Resize & Color Convert
# ===============================================================================

    else:

        if (CC == "FMTConv") and ((ChromaSS == "440") or (ChromaSS == "410")):
            raise vs.Error(f'CropResize: FMTConv doesn\'t support {SFName} for color conversion')
        if (SCFamily == vs.RGB):
            raise vs.Error('CropResize: RGB isn\'t supported for color conversion')
        if SFName.endswith("H"):
            raise vs.Error(f'CropResize: {SFName} isn\'t supported for color conversion')
        if (SCFamily == vs.GRAY):
            raise vs.Error(f'CropResize: {SFName} video doesn\'t require color conversion')

        MatPrimS = "470bg" if (CM == "601-709") or (CM == "601-2020") or \
          (CM == "601P-601N") or (CM == "601P-709") or (CM == "601P-2020") else \
          "170m" if (CM == "601N-601P") or (CM == "601N-709") or (CM == "601N-2020") else \
          "709" if (CM == "709-601") or (CM == "709-601N") or \
          (CM == "709-601P") or (CM == "709-2020") else \
          "2020" if (CM == "2020-601") or (CM == "2020-601N") or \
          (CM == "2020-601P") or (CM == "2020-709") else ""

        MatPrimD = "470bg" if (CM == "709-601") or (CM == "2020-601") or (CM == "601N-601P") or \
          (CM == "709-601P") or (CM == "2020-601P") else \
          "170m" if (CM == "601P-601N") or (CM == "709-601N") or (CM == "2020-601N") else \
          "709" if (CM == "601-709") or (CM == "601N-709") or \
          (CM == "601P-709") or (CM == "2020-709") else \
          "2020" if (CM == "601-2020") or (CM == "601N-2020") or \
          (CM == "601P-2020") or (CM == "709-2020") else ""

        if (CC == "VapourSynth"):

            MatS = "2020ncl" if (MatPrimS == "2020") else MatPrimS
            MatD = "2020ncl" if (MatPrimD == "2020") else MatPrimD

            TranS = "601" if (MatPrimS == "470bg") or (MatPrimS == "170m") else \
              "2020_10" if (MatPrimS == "2020") else MatPrimS
            TranD = "601" if (MatPrimD == "470bg") or (MatPrimD == "170m") else \
              "2020_10" if (MatPrimD == "2020") else MatPrimD

            MatPrimIntS = 5 if (MatPrimS == "470bg") else 6 if (MatPrimS == "170m") else \
              1 if (MatPrimS == "709") else 9 if (MatPrimS == "2020") else 2

            MatIntS = 9 if (MatS == "2020ncl") else MatPrimIntS
            TranIntS = 6 if (TranS == "601") else 14 if (TranS == "2020_10") else MatPrimIntS

            PropsMatrix = Source.get_frame(0).props.get("_Matrix", None)
            PropsMatrix = None \
              if (PropsMatrix is None) or (PropsMatrix == 2) or (PropsMatrix == 3) else PropsMatrix
            PropsPrim = Source.get_frame(0).props.get("_Primaries", None)
            PropsPrim = \
              None if (PropsPrim is None) or (PropsPrim == 2) or (PropsPrim == 3) else PropsPrim
            PropsTrans = Source.get_frame(0).props.get("_Transfer", None)
            PropsTrans = \
              None if (PropsTrans is None) or (PropsTrans == 2) or (PropsTrans == 3) else PropsTrans

            if (PropsMatrix is not None) and \
              (((PropsMatrix != 5) and (PropsMatrix != 6) and (5 <= MatIntS <= 6)) or \
              ((PropsMatrix != 1) and (MatIntS == 1)) or ((PropsMatrix != 9) and (MatIntS == 9))):
                raise vs.Error('CropResize: The value for color matrix in frame properties ' + \
                               f'({PropsMatrix}) is not ' + MatS + ' but a conversion from ' + \
                               MatS + ' has been attempted (try FMTConv instead)')
            if (PropsPrim is not None) and (Is601709 is False) and \
              (((PropsPrim != 5) and (MatPrimIntS == 5)) or ((PropsPrim != 6) and (MatPrimIntS == 6)) or \
              ((PropsPrim != 1) and (MatPrimIntS == 1)) or ((PropsPrim != 9) and (MatPrimIntS == 9))):
                raise vs.Error('CropResize: The value for color primaries in frame properties ' + \
                             f'({PropsPrim}) is not ' + MatPrimS + ' but a conversion from ' + \
                             MatPrimS + ' has been attempted (try FMTConv instead)')
            if (PropsTrans is not None) and (Is601709 is False) and \
              (((PropsTrans != 1) and (PropsTrans != 6) and (PropsTrans != 14)) and \
              ((TransIntS == 1) or (TransIntS == 6) or (TransIntS == 14))):
                raise vs.Error('CropResize: The value for color transfer in frame properties ' + \
                               f'({PropsTrans}) is not ' + TransS + ' but a conversion from ' + \
                               TransS + ' has been attempted (try FMTConv instead)')

            CMatrix = ".resize.Bicubic(matrix_in_s=MatS, matrix_s=MatD)"
            CMatPrim = \
              ".resize.Bicubic(matrix_in_s=MatS, matrix_s=MatD, " + \
              "primaries_in_s=MatPrimS, primaries_s=MatPrimD, " + \
              "transfer_in_s=TranS, transfer_s=TranD)"

        if (CC == "FMTConv"):

            MatS = "601" if (MatPrimS == "470bg") or (MatPrimS == "170m") else MatPrimS
            MatD = "601" if (MatPrimD == "470bg") or (MatPrimD == "170m") else MatPrimD

            PrimS = "601-625" if (MatPrimS == "470bg") else \
              "601-525" if (MatPrimS == "170m") else MatPrimS
            PrimD = "601-625" if (MatPrimD == "470bg") else \
              "601-525" if (MatPrimD == "170m") else MatPrimD

            TwoZero = "420"
            TwoTwo = "422"
            FourFour = "444"
            Trans = "linear"

            To4 = "" if (ChromaSS == "444") else ".fmtc.resample(css=FourFour)"

            From4 = "" if (ChromaSS == "444") else \
              ".fmtc.resample(css=TwoTwo)" if (ChromaSS == "444") else \
              ".fmtc.resample(css=TwoZero)"

            CMatrix = \
              ".fmtc.matrix(mats=MatS, matd=MatD, fulls=False, fulld=False, bits=16)"
            CMatPrim = \
              ".fmtc.matrix(mat=MatS, fulls=False, fulld=True, bits=16)" + \
              ".fmtc.transfer(transs=MatS, transd=Trans)" + \
              ".fmtc.primaries(prims=PrimS, primd=PrimD)" + \
              ".fmtc.transfer(transs=Trans, transd=MatD)" + \
              ".fmtc.matrix(mat=MatD, fulls=True, fulld=False, bits=16)"

        CConvert = CMatrix if (Is601709 is True) else CMatPrim
        CConvert = CConvert if (CC == "VapourSynth") else To4 + CConvert + From4

        CConverted = eval("Cropped" + To16Bit + CConvert) \
          if (NoResizer is True) or (RingRepair > 0) else ""

        Resized = "" if (NoResizer is True) or (RingRepair == 0) else \
          eval("CConverted" + DoResize) \
          if (IsFResize is True) or (IsVKernel is True) else \
          eval(Resizer + "(CConverted, CRVideo_Width, CRVideo_Height, src_left=RCL, src_top=RCT, " + \
          "src_width=F_Cropped_Width, src_height=F_Cropped_Height" + ResStr + ")")

        CCVideo = "" if (UpScale is False) or \
          (IsFResize is True) or (IsVKernel is True) or \
          (NoResizer is True) or (RingRepair > 0) else \
          eval("Cropped" + To16Bit + CConvert + Undo16Bit)

        CRVideo = eval("CConverted" + Undo16Bit) \
          if (NoResizer is True) else \
          eval("CR_RingRepair(Resized, CConverted, CRVideo_Width, CRVideo_Height, " + \
          "RCL, RCT, F_Cropped_Width, F_Cropped_Height, RingRepair)" + Undo16Bit) \
          if (RingRepair > 0) else \
          eval("Cropped" + To16Bit + CConvert + DoResize + Undo16Bit) \
          if (UpScale is True) and (IsFResize is True) else \
          eval("Cropped" + To16Bit + DoResize + CConvert + Undo16Bit) \
          if (IsFResize is True) else \
          eval("Cropped" + To16Bit + CConvert + Undo16Bit + DoResize) \
          if (UpScale is True) and (IsVKernel is True) else \
          eval("Cropped" + DoResize + To16Bit + CConvert + Undo16Bit) \
          if (IsVKernel is True) else \
          eval(Resizer + "(CCVideo, CRVideo_Width, CRVideo_Height, src_left=RCL, src_top=RCT, " + \
          "src_width=F_Cropped_Width, src_height=F_Cropped_Height" + ResStr + ")" + Undo16Bit) \
          if (UpScale is True) else \
          eval(DoResize + To16Bit + CConvert + Undo16Bit)

# ===============================================================================
#            Add Borders As Required
# ===============================================================================

    if (IsBordered is True) and ((Final_Width > CRVideo_Width) or (Final_Height > CRVideo_Height)):

        if (SCFamily == vs.RGB) and (Frosty is True):
            raise vs.Error('CropResize: RGB is not supported for FrostyBorders')

        Additional_Width = Final_Width - CRVideo_Width
        EvenWidthSplit = (Additional_Width % (WidthMod * 2) == 0)
        F_WidthSplit = Additional_Width / 2.0
        BL_Adjust = F_WidthSplit - (m.floor(F_WidthSplit / F_WidthMod) * WidthMod)
        BL_Adjust_Up = min(Additional_Width, round(F_WidthSplit + WidthMod - BL_Adjust))
        BL_Adjust_Down = F_WidthSplit - BL_Adjust
        BL = round(Additional_Width / 2 if (EvenWidthSplit is True) else \
          BL_Adjust_Up if (CroppedMoreLeft is True) else BL_Adjust_Down)
        BR = Additional_Width - BL

        Additional_Height = Final_Height - CRVideo_Height
        EvenHeightSplit = (Additional_Height % (HeightMod * 2) == 0)
        F_HeightSplit = Additional_Height / 2.0
        BT_Adjust = F_HeightSplit - (m.floor(F_HeightSplit / F_HeightMod) * HeightMod)
        BT_Adjust_Up = min(Additional_Height, round(F_HeightSplit + HeightMod - BT_Adjust))
        BT_Adjust_Down = F_HeightSplit - BT_Adjust
        BT = round(Additional_Height / 2 if (EvenHeightSplit is True) else \
          BT_Adjust_Up if (CroppedMoreTop is True) else BT_Adjust_Down)
        BB = Additional_Height - BT

        Frosty = (Frosty is True) and \
          (((Final_StorageAR >= CRVideo_StorageAR) and (BL > WidthMod < BR)) or \
          ((Final_StorageAR < CRVideo_StorageAR) and (BT > HeightMod < BB)))

        FrostyStr = "" if (Frosty is False) else \
          "CR_FrostyBorders(CRVideo, CRVideo_Width, CRVideo_Height, " + \
          "F_CRVideo_Width, F_CRVideo_Height, BL, 0, BR, 0, WidthMod, HeightMod, " + \
          "F_WidthMod, F_HeightMod, CRVideo_StorageAR, Final_StorageAR, " + \
          "Source_Bits, ChromaSS, SCFamily)" \
          if (Final_StorageAR >= CRVideo_StorageAR) else \
          "CR_FrostyBorders(CRVideo, CRVideo_Width, CRVideo_Height, " + \
          "F_CRVideo_Width, F_CRVideo_Height, 0, BT, 0, BB, WidthMod, HeightMod, " + \
          "F_WidthMod, F_HeightMod, CRVideo_StorageAR, Final_StorageAR, " + \
          "Source_Bits, ChromaSS, SCFamily)"

        CRVideo = CRVideo.std.AddBorders(BL, BR, BT, BB, color=BColor) \
          if (Frosty is False) else \
          eval(FrostyStr) \
          if ((Final_StorageAR >= CRVideo_StorageAR) and (0 == BT == BB)) or \
          ((Final_StorageAR < CRVideo_StorageAR) and (0 == BL == BR)) else \
          eval(FrostyStr + ".std.AddBorders(0, BT, 0, BB, color=BColor)") \
          if (Final_StorageAR >= CRVideo_StorageAR) else \
          eval(FrostyStr + ".std.AddBorders(BL, 0, BR, 0, color=BColor)")

    else:

        BL = BR = BT = BB = 0

# ===============================================================================
#            For Info=True & Cropping Previews
# ===============================================================================

    if (Info is True) or (IsScriptCPreview is True):

        Extra_CL = \
          F_NR_Extra_CL if (NoResizing is True) and (CSplit == 1) else \
          F_NR_Extra_Crop_LR if (NoResizing is True) else \
          F_RWO_Extra_Crop_LR if (ResizeWO is True) else \
          F_FRA_Extra_Crop_LR if (FullResizeA is True) else F_FRB_Extra_Crop_LR
        Extra_CR = \
          F_NR_Extra_CR if (NoResizing is True) and (CSplit == 1) else \
          F_NR_Extra_Crop_LR if (NoResizing is True) else \
          F_RWO_Extra_Crop_LR if (ResizeWO is True) else \
          F_FRA_Extra_Crop_LR if (FullResizeA is True) else F_FRB_Extra_Crop_LR
        Extra_CT = \
          F_NR_Extra_CT if (NoResizing is True) and (CSplit == 1) else \
          F_NR_Extra_Crop_TB if (NoResizing is True) else \
          F_RWO_Extra_CT if (ResizeWO is True) and (CSplit == 1) else \
          F_RWO_Extra_Crop_TB if (ResizeWO is True) else \
          F_FRA_Extra_Crop_T if (FullResizeA is True) else F_FRB_Extra_Crop_T
        Extra_CB = \
          F_NR_Extra_CB if (NoResizing is True) and (CSplit == 1) else \
          F_NR_Extra_Crop_TB if (NoResizing is True) else \
          F_RWO_Extra_CB if (ResizeWO is True) and (CSplit == 1) else \
          F_RWO_Extra_Crop_TB if (ResizeWO is True) else \
          F_FRA_Extra_Crop_B if (FullResizeA is True) else F_FRB_Extra_Crop_B

        Total_CL = CL + Extra_CL
        Total_CR = CR + Extra_CR
        Total_CT = CT + Extra_CT
        Total_CB = CB + Extra_CB

        OutVal = 1 if (Info is True) else 2

        CRVideo_Mod_Text = CR_PicMod(CRVideo_Width, CRVideo_Height, OutVal)

        Final_Mod_Text = CR_PicMod(CRVideo_Width, CRVideo_Height, OutVal) \
          if (IsBordered is False) else CR_PicMod(Final_Width, Final_Height, OutVal)

        Source_Mod_Text = CR_PicMod(Source_Width, Source_Height, OutVal)

# -------------------------------------------------------------------------------
#            For Script Cropping Previews
# -------------------------------------------------------------------------------

    if (Info is False) and (IsScriptCPreview is True):

        SCP_Source = "Source " + \
          ("(Non Anamorphic)\n" if (IsAnamorphicInput is False) else "(Anamorphic)\n") + \
          str(Source_Width) + " x " + str(Source_Height) + Source_Mod_Text + \
          "\nDAR  " + CR_FS(Source_DAR, 6, True) + CR_FTF(Source_DAR) + \
          "\nSAR  " + CR_FS(Source_SAR, 6, True) + \
          (CR_FTF(Source_SAR) if (IsPropsSAR is False) else f" ({PropsNum}:{PropsDen})")

        SCP_SpecifiedC = "\n\n\nSpecified Cropping" + "\nLeft " + CR_FS(CL) + \
          ", Right " + CR_FS(CR) + ", Top " + CR_FS(CT) + ", Bottom " + CR_FS(CB)

        SCP_TotalC = "\n\nTotal Cropping" + "\nLeft " + CR_FS(Total_CL) + \
          ", Right " + CR_FS(Total_CR) + ", Top " + CR_FS(Total_CT) + ", Bottom " + CR_FS(Total_CB)

        SCP_Output = "\n\n\n" + \
          ("Output " if (IsBordered is False) else "Picture Excluding Borders\n") + \
          ("(Non Anamorphic)\n" if (IsAnamorphicOutput is False) else "(Anamorphic)\n") + \
          str(CRVideo_Width) + " x " + str(CRVideo_Height) + CRVideo_Mod_Text + \
          "\nDAR  " + CR_FS(CRVideo_DAR, 6, True) + \
          (CR_FTF(Source_DAR) if \
          (NoResizing is True) and (Final_Width == Source_Width) and \
          (Final_Height == Source_Height) else CR_FTF(CRVideo_DAR)) + \
          "\nSAR  " + CR_FS(Final_SAR, 6, True) + (CR_FTF(Final_SAR) \
          if (NoResizing is False) or (IsPropsSAR is False) else f" ({PropsNum}:{PropsDen})") + "\n"

        SCP_GMode = "" if (IsGModeText is False) else "GMode  " + \
          ("0" if (GMode == 0) else "1" if (GMode == 1) else "2")

# -------------------------------------------------------------------------------
#            Script Cropping Previews
# -------------------------------------------------------------------------------

    if (IsScriptCPreview is True):

        Is444 = (Source.format.subsampling_w == 0 == Source.format.subsampling_h)

        Video444 = Source if (SCFamily != vs.YUV) else \
          core.resize.Bicubic(Source, format=vs.YUV444PS) \
          if (Is444 is False) and SFName.endswith("S") else \
          eval("core.resize.Bicubic(Source, format=vs.YUV444P" + str(Source_Bits) + ")") \
          if (Is444 is False) else Source

        Scale = 2 ** (Video444.format.bits_per_sample - 8)
        MinIn = 16 * Scale
        MaxInL = 235 * Scale
        MaxInC = 240 * Scale
        MaxOut = 255 * Scale

        PL = round(CL)
        PR = round(CR)
        PT = round(CT)
        PB = round(CB)

        CLine_L = min(round(CL), CLine)
        CLine_R = min(round(CR), CLine)
        CLine_T = min(round(CT), CLine)
        CLine_B = min(round(CB), CLine)

        Script_CL = round(Total_CL) if (round(Total_CL) > PL) else 0
        Script_CR = round(Total_CR) if (round(Total_CR) > PR) else 0
        Script_CT = round(Total_CT) if (round(Total_CT) > PT) else 0
        Script_CB = round(Total_CB) if (round(Total_CB) > PB) else 0

        SLine_L = min(Script_CL, CLine)
        SLine_R = min(Script_CR, CLine)
        SLine_T = min(Script_CT, CLine)
        SLine_B = min(Script_CB, CLine)

        Cropping = ".Crop(Script_CL, Script_CR, Script_CT, Script_CB)"

        BorderingA = \
          ".AddBorders(Script_CL-SLine_L, Script_CR-SLine_R, Script_CT-SLine_T, Script_CB-SLine_B)"
        BorderingB = \
          ".AddBorders(Script_CL, Script_CR, Script_CT, Script_CB"

        BVid = core.std.BlankClip(Video444)

        if (CPreview == 1):

            OVid = eval("BVid.std" + Cropping)
            OVid = OVid.std.AddBorders(SLine_L, SLine_R, SLine_T, SLine_B, \
                   color=RGBColor(Video444, 'ADD8E6'))
            OVid = eval("OVid.std" + BorderingA)
            OVid = OVid.std.Crop(PL, PR, PT, PB)
            OVid = OVid.std.AddBorders(CLine_L, CLine_R, CLine_T, CLine_B, \
                   color=RGBColor(Video444, 'FFFF00'))
            OVid = OVid.std.AddBorders(PL-CLine_L, PR-CLine_R, PT-CLine_T, PB-CLine_B)

            MVid = eval("BVid.std" + Cropping)
            MVid = MVid.std.AddBorders(SLine_L, SLine_R, SLine_T, SLine_B, \
                   color=RGBColor(Video444, 'FFFFFF'))
            MVid = eval("MVid.std" + BorderingA)
            MVid = MVid.std.Crop(PL, PR, PT, PB)
            MVid = MVid.std.AddBorders(CLine_L, CLine_R, CLine_T, CLine_B, \
                   color=RGBColor(Video444, 'B1B1B1'))
            MVid = MVid.std.AddBorders(PL-CLine_L, PR-CLine_R, PT-CLine_T, PB-CLine_B)

        elif (CPreview == 2):

            OVid = eval("BVid.std" + Cropping)
            OVid = eval("OVid.std" + BorderingB + ", color=RGBColor(Video444, 'ADD8E6'))")
            OVid = OVid.std.Crop(PL, PR, PT, PB)
            OVid = OVid.std.AddBorders(PL, PR, PT, PB, color=RGBColor(Video444, 'FFFF00'))

            MVid = eval("BVid.std" + Cropping)
            MVid = eval("MVid.std" + BorderingB + ", color=RGBColor(Video444, '7F7F7F'))")
            MVid = MVid.std.Crop(PL, PR, PT, PB)
            MVid = MVid.std.AddBorders(PL, PR, PT, PB, color=RGBColor(Video444, '4B4B4B'))

        else:

            OVid = eval("Video444.std" + Cropping)
            OVid = OVid = eval("OVid.std" + BorderingB + ", color=RGBColor(Video444, 'FF0000'))")
            OVid = OVid.std.Crop(PL, PR, PT, PB)
            OVid = OVid.std.AddBorders(PL, PR, PT, PB)

            MVid = eval("core.std.BlankClip(Video444, color=RGBColor(Video444, 'FFFFFF'))" + \
                   ".std" + Cropping)
            MVid = eval("MVid.std" + BorderingB + ", color=RGBColor(Video444, '4B4B4B'))")
            MVid = MVid.std.Crop(PL, PR, PT, PB)
            MVid = MVid.std.AddBorders(PL, PR, PT, PB)

        MVid = MVid if (SCFamily == vs.RGB) or SFName.endswith("S") else \
               MVid.std.Levels(min_in=MinIn, max_in=MaxInL, min_out=0, max_out=MaxOut, planes=0)
        MVid = MVid if (SCFamily == vs.RGB) or (SCFamily == vs.GRAY) or SFName.endswith("S") else \
               MVid.std.Levels(min_in=MinIn, max_in=MaxInC, min_out=0, max_out=MaxOut, planes=[1, 2])

        ScriptCPreviewVideo = \
          core.std.MaskedMerge(Video444, OVid, mask=MVid, first_plane=True) \
          if (CPreview == 1) or (CPreview == 2) else \
          core.std.MaskedMerge(core.std.Invert(Video444), OVid, mask=MVid, first_plane=True)
        ScriptCPreviewVideo = \
          ScriptCPreviewVideo if (Info is True) else \
          core.text.Text(ScriptCPreviewVideo, \
          SCP_Source + SCP_SpecifiedC + SCP_TotalC + SCP_Output, alignment=5)
        ScriptCPreviewVideo = \
          ScriptCPreviewVideo if (Info is True) or (IsGModeText is False) else \
          core.text.Text(ScriptCPreviewVideo, SCP_GMode, alignment=2)

        ScriptCPreviewVideo = \
          ScriptCPreviewVideo if (CPreview == 1) or (SCFamily != vs.YUV) else \
          eval("core.resize.Bicubic(ScriptCPreviewVideo, format=vs." + SFName + ")") \
          if (Is444 is False) else ScriptCPreviewVideo

# ===============================================================================
#            Display Info
# ===============================================================================

    if (Info is True):

        InfoVideo = ScriptCPreviewVideo if (IsScriptCPreview is True) else CRVideo
        InfoVideo_Width = InfoVideo.width
        InfoVideo_Height = InfoVideo.height

        PaddingL = PaddingR = PaddingT = PaddingB = 0
        Pad_Height = 480 if (IsBordered is True) else 432
        IsPadding = (InfoVideo_Width < 800) or (InfoVideo_Height < Pad_Height)

        if (InfoVideo_Width < 832):
            PaddingLR = 832 - InfoVideo_Width
            EvenWidthSplit = (PaddingLR % (WidthMod * 2) == 0)
            PaddingL = round(PaddingLR / 2.0) if (EvenWidthSplit is True) else \
              m.floor(PaddingLR / 2.0 / WidthMod) * WidthMod
            PaddingR = round(PaddingLR - PaddingL)

        if (InfoVideo_Height < Pad_Height):
            PaddingTB = Pad_Height - InfoVideo_Height
            EvenHeightSplit = (PaddingTB % (HeightMod * 2) == 0)
            PaddingT = round(PaddingTB / 2.0) if (EvenHeightSplit is True) else \
              m.floor(PaddingTB / 2.0 / HeightMod) * HeightMod
            PaddingB = round(PaddingTB - PaddingT)

        if (IsPadding is True):
            InfoVideo = InfoVideo.std.AddBorders(PaddingL, PaddingR, PaddingT, PaddingB, \
              color=RGBColor(InfoVideo, 'FFFFFF'))

        I_D = "\nDisplay Aspect Ratio  "
        I_S = "\nSample Aspect Ratio  "
        I_St = "\nStorage Aspect Ratio  "

# -------------------------------------------------------------------------------
#            Left Column
# -------------------------------------------------------------------------------

        I_Source = \
          f"Source Resolution  {Source_Width} x {Source_Height}" + Source_Mod_Text + \
          I_St + CR_FS(Source_StorageAR, 6, True) + CR_FTF(Source_StorageAR) + \
          I_D + CR_FS(Source_DAR, 6, True) + CR_FTF(Source_DAR) + "\n" + \
          f"{Source_Bits} bit " + \
          ("integer" if (Source_Sample_Type == 0) else "Float") + f"  {SFName}  " + \
          CR_FS(FRate, 4) + "fps " + \
          (CR_FTF(FRate) if CR_IsEqual(round(FRate), FRate) else \
          f"({FRateNum}/{FRateDen})")

        I_SpecifiedC = "\n\nSpecified Cropping\nLeft " + \
          CR_FS(CL) + ", Right " + CR_FS(CR) + \
          ", Top " + CR_FS(CT) + ", Bottom " + CR_FS(CB)

        I_ExtraC = "\nAdditional Script Cropping\n(" + \
          CR_FS(Extra_CL) + ", " + CR_FS(Extra_CR) + ", " + \
          CR_FS(Extra_CT) + ", " + CR_FS(Extra_CB) + ")"

        I_CSplit_Crop = "" if (NoResizer is True) else \
          f"\n\nCrop ({CCL}, {CCR}, {CCT}, {CCB})"

        I_CSplit_Resizer = "" if (NoResizer is True) else \
          "\nResizer Cropping (displayed as L,R,T,B)\n(" + \
          CR_FS(RCL) + ", " + CR_FS(RCR) + ", " + \
          CR_FS(RCT) + ", " + CR_FS(RCB) + ")"

        I_TotalC = "\nTotal Cropping\n(" + \
          CR_FS(Total_CL) + ", " + CR_FS(Total_CR) + ", " + \
          CR_FS(Total_CT) + ", " + CR_FS(Total_CB) + \
          (")\n\n\n\n" if (NoResizer is True) else ")")

        I_Cropped_Resolution = "\n\nCropped Resolution  " + \
          (CR_FS(F_NR_Cropped_Width) + " x " + CR_FS(F_NR_Cropped_Height) \
          if (NoResizing is True) else \
          CR_FS(F_RWO_Cropped_Width) + " x " + CR_FS(F_RWO_Cropped_Height) \
          if (ResizeWO is True) else \
          CR_FS(F_FRA_Cropped_Width) + " x " + CR_FS(F_FRA_Cropped_Height) \
          if (FullResizeA is True) else \
          CR_FS(F_FRB_Cropped_Width) + " x " + CR_FS(F_FRB_Cropped_Height))

        I_Cropped_StorageAR = NR_Cropped_StorageAR if (NoResizing is True) else \
          RWO_Cropped_StorageAR if (ResizeWO is True) else \
          FRA_Cropped_StorageAR if (FullResizeA is True) else FRB_Cropped_StorageAR

        I_Cropped_DAR = NR_Cropped_DAR if (NoResizing is True) else \
          RWO_Cropped_DAR if (ResizeWO is True) else \
          FRA_Cropped_DAR if (FullResizeA is True) else FRB_Cropped_DAR

        I_Cropped_DAR = \
          I_Cropped_DAR if CR_IsEqual(I_Cropped_DAR, CRVideo_DAR) else CRVideo_DAR

        I_Cropped_ARs = \
          I_St + CR_FS(I_Cropped_StorageAR, 6, True) + CR_FTF(I_Cropped_StorageAR) + \
          I_D + CR_FS(I_Cropped_DAR, 6, True) + \
          (CR_FTF(Source_DAR) if \
          (NoResizing is True) and (CRVideo_Width == Source_Width) and \
          (CRVideo_Height == Source_Height) else CR_FTF(I_Cropped_DAR))

        I_CRVideo_ARs = CRVideo_Mod_Text + \
          I_St + CR_FS(CRVideo_StorageAR, 6, True) + CR_FTF(CRVideo_StorageAR) + \
          I_D + CR_FS(CRVideo_DAR, 6, True) + \
          (CR_FTF(Source_DAR) if \
          (NoResizing is True) and (CRVideo_Width == Source_Width) and \
          (CRVideo_Height == Source_Height) else CR_FTF(CRVideo_DAR))

        I_Output = \
          f"\n\nOutput Resolution  {CRVideo_Width} x {CRVideo_Height}" + I_CRVideo_ARs \
          if (IsBordered is False) else \
          f"\n\nOutput Exc Borders  {CRVideo_Width} x {CRVideo_Height}" + I_CRVideo_ARs + \
          f"\n\nAddBorders ({BL}, {BR}, {BT}, {BB})" + \
          f"\nOutput Inc Borders  {Final_Width} x {Final_Height}" + Final_Mod_Text + \
          I_St + CR_FS(Final_StorageAR, 6, True) + CR_FTF(Final_StorageAR) + \
          I_D + CR_FS(Final_DAR, 6, True) + \
          (CR_FTF(Source_DAR) if \
          (NoResizing is True) and (Final_Width == Source_Width) and \
          (Final_Height == Source_Height) else CR_FTF(Final_DAR))

        I_Left_Column = I_Source + I_SpecifiedC + I_ExtraC + I_CSplit_Crop + \
          I_CSplit_Resizer + I_TotalC + I_Cropped_Resolution + I_Cropped_ARs + I_Output

# -------------------------------------------------------------------------------
#            Right Column
# -------------------------------------------------------------------------------

        I_NoRC = (NoResizer is True) or \
          ((ResizeWO is True) and ((CSplit == 1) or (0 == RWO_RCT == RWO_RCB)))

        I_ResizingMode = "\n" + \
          ("Clean Borders" if (CleanBorders is True) and (I_NoRC is True) else \
          "Resampling  (CleanBorders)" if (CleanBorders is True) else \
          "No Resizing" if (NoResizing is True) and (I_NoRC is True) else \
          "Resampling  (NoResize)" if (NoResizing is True) else \
          "Resize Width Only" if (ResizeWO is True) and (I_NoRC is True) else \
          "Resampling Height (ResizeWO)" if (ResizeWO is True) else "Full Resizing Mode")

        I_Resizer = "" if (NoResizer is True) else \
          "\nResizer  " + ("f_" if (IsFResize is True) else "") + Resizer + \
          (" (3 taps)" if (RTaps == ", filter_param_a=3") or (RTaps == ", taps=3") else \
          " (4 taps)" if (RTaps == ", filter_param_a=4") or (RTaps == ", taps=4")  else "")

        I_RStr = "" if (NoResizer is True) or (RStr == "") else "\nRStr  " + RStr

        I_RingRepair = "" if (RingRepair == 0) else \
          "\nRingRepair Weight " + CR_FS(RingRepair, 2, True)

        I_Input_ARs = ("\n\n\n" if (NoResizer is True) else \
          "\n\n" if (RingRepair == 0) and (RStr == "") else "\n") + \
          "\nInput Aspect" + \
          ("  InDAR" + I_D + CR_FS(InDAR, 6, True) + CR_FTF(InDAR) + \
          I_S + CR_FS(Source_SAR, 6, True) + CR_FTF(Source_SAR) \
          if (IsInDAR is True) else \
          "  InSAR" + I_D + CR_FS(Source_DAR, 6, True) + CR_FTF(Source_DAR) + \
          I_S + CR_FS(InSAR, 6, True) + CR_FTF(InSAR) \
          if (IsInSAR is True) else \
          "  Global InDAR" + I_D + CR_FS(GInDAR, 6, True) + CR_FTF(GInDAR) + \
          I_S + CR_FS(Source_SAR, 6, True) + CR_FTF(Source_SAR) \
          if (IsGInDAR is True) else \
          "  Global InSAR" + I_D + CR_FS(Source_DAR, 6, True) + CR_FTF(Source_DAR) + \
          I_S + CR_FS(GInSAR, 6, True) + CR_FTF(GInSAR) \
          if (IsGInSAR is True) else \
          "  FrameProps SAR" + I_D + CR_FS(Source_DAR, 6, True) + CR_FTF(Source_DAR) + \
          I_S + CR_FS(PropsSAR, 6, True) + f" ({PropsNum}:{PropsDen})" \
          if (IsPropsSAR is True) else "") + \
          ("\nNot Anamorphic" \
          if (AnamorphicInput is True) and (IsAnamorphicInput is False) else \
          "  FrameProps SAR" \
          if (AnamorphicInput is False) and (IsPropsSAR1 is True) else \
          "  Not Specified" if (AnamorphicInput is False) else "") + \
          (I_S + CR_FS(Source_SAR, 6, True) + CR_FTF(Source_SAR) \
          if (AnamorphicInput is False) else "") + \
          ("" if (IsAnamorphicInput is False) else "\nDisplay Dimensions  " + \
          CR_FS(F_Source_Display_Width, 2) + " x " + str(Source_Height))

        I_Output_ARs = "\n\nOutput Aspect" + \
          (I_D + CR_FS(Final_DAR, 6, True) + CR_FTF(Source_DAR) + \
          I_S + CR_FS(Final_SAR, 6, True) + \
          (CR_FTF(Final_SAR) if (IsPropsSAR is False) else f" ({PropsNum}:{PropsDen})") \
          if (NoResizing is True) and (IsAnamorphicOutput is True) and \
          (Final_Width == Source_Width) and (Final_Height == Source_Height) else \
          I_D + CR_FS(Final_DAR, 6, True) + CR_FTF(Final_DAR) + \
          I_S + CR_FS(Final_SAR, 6, True) + \
          (CR_FTF(Final_SAR) if (IsPropsSAR is False) else f" ({PropsNum}:{PropsDen})") \
          if (NoResizing is True) and (IsAnamorphicOutput is True) else \
          "  OutDAR" + I_D + CR_FS(OutDAR, 6, True) + CR_FTF(OutDAR) + \
          I_S + CR_FS(Final_SAR, 6, True) + CR_FTF(Final_SAR) \
          if (IsOutDAR is True) else \
          "  OutSAR" + I_D + CR_FS(Final_DAR, 6, True) + CR_FTF(Final_DAR) + \
          I_S + CR_FS(OutSAR, 6, True) + CR_FTF(OutSAR) \
          if (IsOutSAR is True) else \
          "  Global OutDAR" + I_D + CR_FS(GOutDAR, 6, True) + CR_FTF(GOutDAR) + \
          I_S + CR_FS(Final_SAR, 6, True) + CR_FTF(Final_SAR) \
          if (IsGOutDAR is True) else \
          "  Global OutSAR" + I_D + CR_FS(Final_DAR, 6, True) + CR_FTF(Final_DAR) + \
          I_S + CR_FS(GOutSAR, 6, True) + CR_FTF(GOutSAR) \
          if (IsGOutDAR is True) else "") + \
          ("\nNot Anamorphic" \
          if (AnamorphicOutput is True) and (IsAnamorphicOutput is False) else \
          "  Not Specified" \
          if (AnamorphicOutput is False) else "") + \
          (I_S + CR_FS(Final_SAR, 6, True) + CR_FTF(Final_SAR) \
          if (AnamorphicOutput is False) else "") + \
          ("" if (IsAnamorphicOutput is False) else \
          "\nDisplay Dimensions  " + \
          CR_FS(F_Final_Display_Width, 2) + " x " + str(Final_Height))

        LineCond = (IsBordered is False) and (AnamorphicInput is True) and \
          (AnamorphicOutput is True) and (NoResizer is False) and \
          (RingRepair > 0) and (RStr != "")
        I_Align = 0 if (IsPadding is False) else 7 if (IsBordered is True) else 8

        LB1 = "\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n"
        LB2 = "\n\n\n" if (IsBordered is True) else ""
        GModeLB = "\n" if (IsGModeText is False) else ""
        CConvertLB = "\n\n\n" if (CC == "") and (IsBordered is True) else \
          "\n\n\n" if (CC == "") else "\n" if (CM == "") else ""
        PosLB = "\n\n" if (Position is False) else ""
        PadLB = "\n\n" if (IsPadding is False) and (IsBordered is True) else ""
        AutoDARLineBreak = \
          "" if (AutoDAR is False) else LB1 + "\n"
        GModeLineBreak = \
          "" if (IsGModeText is False) else LB1 + LB2 + "\n\n" + CConvertLB + PosLB + PadLB
        CConvertLineBreak = \
          "" if (CC == "") else LB1 + LB2 + "\n\n\n\n" + CConvertLB + PosLB + PadLB
        PosLineBreak1 = \
          "" if (Position is False) else LB1 + LB2 + "\n\n\n\n\n\n\n" + PosLB + PadLB
        PosLineBreak2 = \
          "" if (Position is False) else "   "
        PaddingLB = \
          "" if (IsPadding is False) else LB1 + LB2 + "\n\n\n\n\n\n\n\n\n" + PadLB

        I_PreAutoDAR = 0.0 if (IsAutoDAR is False) else \
          F_Resizing_Width / F_FRA_Adjusted_Height1 \
          if (FullResizeA is True) and (IsOutputSAR is False) else  \
          F_Resizing_Width * OutputSAR / F_FRA_Adjusted_Height1 \
          if (FullResizeA is True) else  \
          F_FRB_Adjusted_Width1 / F_Resizing_Height \
          if (IsOutputSAR is False) else  \
          F_FRB_Adjusted_Width1 * OutputSAR / F_Resizing_Height

        I_AutoDAR = "" if (AutoDAR is False) else \
          AutoDARLineBreak + "AutoDAR  " + \
          ("Not Applicable" \
          if (IsAutoDAR is False) else \
          CR_FS(I_PreAutoDAR, 6) + " -> 16:9" \
          if (FullResizeA is True) and (IsHeight169 is True) and \
          (F_FRA_Adjusted_Height1 != F_FRA_Adjusted_Height2) else \
          CR_FS(I_PreAutoDAR, 6) + " -> 4:3" \
          if (FullResizeA is True) and (IsHeight43 is True) and \
          (F_FRA_Adjusted_Height1 != F_FRA_Adjusted_Height2) else \
          "Not Required"  \
          if (FullResizeA is True) and \
          ((IsHeight169 is True) or (IsHeight43 is True)) else \
          "Out Of Range" if (FullResizeA is True) else \
          CR_FS(I_PreAutoDAR, 6) + " -> 16:9" \
          if (IsWidth169 is True) and \
          (F_FRB_Adjusted_Width1 != F_FRB_Adjusted_Width2) else \
          CR_FS(I_PreAutoDAR, 6) + " -> 4:3" \
          if (IsWidth43 is True) and \
          (F_FRB_Adjusted_Width1 != F_FRB_Adjusted_Width2) else \
          "Not Required" \
          if (IsWidth169 is True) or (IsWidth43 is True) else "Out Of Range")

        I_ColorConvert = "" if (CC == "") else \
          CConvertLineBreak + "Color Conversion  " + \
          ("None Required" if (IsColorConversion is False) else \
          CC + ("" if (IsColorConversion is False) or (IsScriptCPreview is False) else \
          "\nNot Applied To Cropping Preview")) + \
          ("" if (IsColorConversion is False) or (IsScriptCPreview is True) else "\n" + \
          ("Rec.709 -> Rec.2020" if (CM == "709-2020") else \
          "Rec.2020 -> Rec.709" if (CM == "2020-709") else \
          "Rec.601 -> Rec.709" if (CM == "601-709") else \
          "Rec.601 (470bg) -> Rec.709" if (CM == "601P-709") else \
          "Rec.601 (170m) -> Rec.709" if (CM == "601N-709") else \
          "Rec.709 -> Rec.601" if (CM == "709-601") else \
          "Rec.709 -> Rec.601 (470bg)" if (CM == "709-601P") else \
          "Rec.709 -> Rec.601 (170m)" if (CM == "709-601N") else \
          "Rec.2020 -> Rec.601 (470bg)" \
          if (CM == "2020-601") or (CM == "2020-601P") else \
          "Rec.2020 -> Rec.601 (170m)" if (CM == "2020-601N") else \
          "Rec.601 (470bg) -> Rec.2020" \
          if (CM == "601-2020") or (CM == "601P-2020") else \
          "Rec.601 (170m) -> Rec.2020" if (CM == "601N-2020") else \
          "Rec.601 (470bg) -> Rec.601 (170m)" if (CM == "601P-601N") else \
          "Rec.601 (170m) -> Rec.601 (470bg)" if (CM == "601N-601P") else ""))

        I_GMode = "" if (IsGModeText is False) else \
          GModeLineBreak + "GMode  " + \
          ("0" if (GMode == 0) else "1" if (GMode == 1) else "2")

        I_Borders = "" if (IsPadding is False) else \
          PaddingLB + "White Padding Added To Display Info"

        I_Right_Column = "CropResize 2023-10-28" + I_ResizingMode + I_Resizer + \
          I_RStr + I_RingRepair + I_Input_ARs + I_Output_ARs

# -------------------------------------------------------------------------------

        I_VWidth = InfoVideo.width
        I_HalfW = m.floor(I_VWidth / 2.0 / F_WidthMod) * WidthMod
        I_VSplit = round(I_VWidth - 528 if (I_VWidth > 960) else I_HalfW - 36)

        InfoVideoL = core.std.Crop(InfoVideo, 0, I_VSplit, 0, 0)
        InfoVideoL = core.text.Text(InfoVideoL, I_Left_Column)

        InfoVideoR = core.std.Crop(InfoVideo, I_VWidth - I_VSplit, 0, 0, 0)
        InfoVideoR = core.text.Text(InfoVideoR, I_Right_Column)

        InfoVideoR = \
          InfoVideoR if (AutoDAR is False) else core.text.Text(InfoVideoR, I_AutoDAR)
        InfoVideoR = \
          InfoVideoR if (IsGModeText is False) else core.text.Text(InfoVideoR, I_GMode)
        InfoVideoR = \
          InfoVideoR if (CC == "") else core.text.Text(InfoVideoR, I_ColorConvert)
        InfoVideoR = \
          InfoVideoR if (Position is False) else \
          CR_Position(InfoVideoR, FRateNum, FRateDen, LB1=PosLineBreak1, LB2=PosLineBreak2)
        InfoVideoR = \
          InfoVideoR if (I_Align != 7) else core.text.Text(InfoVideoR, I_Borders)

        InfoVideo = \
          core.std.StackHorizontal([InfoVideoL, InfoVideoR])
        InfoVideo = \
          InfoVideo if (I_Align != 8) else core.text.Text(InfoVideo, I_Borders, alignment=I_Align)

# ===============================================================================
#            CropResize Output
# ===============================================================================

    Output = InfoVideo if (Info is True) else \
      ScriptCPreviewVideo if (IsScriptCPreview is True) else CRVideo
    Output = Output if (Info is True) or (Position is False) else \
      CR_Position(Output, FRateNum, FRateDen, Align=8)

# -------------------------------------------------------------------------------
#            Write SAR To Frame Properties
# -------------------------------------------------------------------------------

    if (IsScriptCPreview is False):

        Out_SAR = 1.0 if (IsAnamorphicOutput is False) else \
          PropsSAR if (NoResizing is True) and (IsPropsSAR is True) else \
          OutSAR if (IsOutSAR is True) else GOutSAR if (IsGOutSAR is True) else Final_SAR

        Frac = str(fr.Fraction(Out_SAR).limit_denominator(max_denominator=10000))

        SAR_Num = 1 if (IsAnamorphicOutput is False) else \
          PropsNum if (NoResizing is True) and (IsPropsSAR is True) else int(Frac.split("/")[0])

        SAR_Den = 1 if (IsAnamorphicOutput is False) else \
          PropsDen if (NoResizing is True) and (IsPropsSAR is True) else int(Frac.split("/")[1])

        Output = core.std.SetFrameProp(Output, prop="_SARNum", intval=SAR_Num)
        Output = core.std.SetFrameProp(Output, prop="_SARDen", intval=SAR_Den)

# -------------------------------------------------------------------------------

    return Output

# ===============================================================================
# ===============================================================================
#            Global GCropResize Function
# ===============================================================================
# ===============================================================================

def GCropResize(OutWidth=None, OutHeight=None, \
    CL=None, CR=None, CT=None, CB=None, \
    CropDAR=None, TCRatio=None, CSplit=None, \
    CPreview=None, CLine=None, \
    InDAR=None, InSAR=None, OutDAR=None, OutSAR=None, AutoDAR=None, \
    Mod=None, HMod=None, \
    NoResize=None, ResizeWO=None, \
    Borders=None, CleanBorders=None, BColor=None, Frosty=None, \
    ColorConvert=None, ColorMode=None, \
    Resizer=None, RStr=None, RingRepair=None, \
    GMode=None, Position=None, Info=None):

# -------------------------------------------------------------------------------

    kwargsdict = {'OutWidth': [OutWidth, int], 'OutHeight': [OutHeight, int],
      'CL': [CL, (int, float)], 'CR': [CR, (int, float)],
      'CT': [CT, (int, float)], 'CB': [CB, (int, float)],
      'CropDAR': [CropDAR, (int, float)], 'TCRatio': [TCRatio, (int, float)],
      'CSplit': [CSplit, int], 'CPreview': [CPreview, int], 'CLine': [CLine, int],
      'InDAR': [InDAR, (int, float)], 'InSAR': [InSAR, (int, float)],
      'OutDAR': [OutDAR, (int, float)], 'OutSAR': [OutSAR, (int, float)],
      'AutoDAR': [AutoDAR, bool], 'Mod': [Mod, int], 'HMod': [HMod, int],
      'NoResize': [NoResize, bool], 'ResizeWO': [ResizeWO, bool],
      'Borders': [Borders, bool], 'CleanBorders': [CleanBorders, bool], 'Frosty': [Frosty, bool],
      'ColorConvert': [ColorConvert, str], 'ColorMode': [ColorMode, str],
      'Resizer': [Resizer, str], 'RStr': [RStr, str], 'RingRepair': [RingRepair, (int, float)],
      'GMode': [GMode, int], 'Position': [Position, bool], 'Info': [Info, bool]}

    for k, v in kwargsdict.items():

        VType = "an integer or float" \
          if (k == 'CL') or (k == 'CR') or (k == 'CT') or (k == 'CB') or (k == 'CropDAR') or \
          (k == 'TCRatio') or (k == 'InDAR') or (k == 'InSAR') or (k == 'OutDAR') or \
          (k == 'OutSAR') or (k == 'RingRepair') else \
          "an integer" \
          if (k == 'OutWidth') or (k == 'OutHeight') or (k == 'CPreview') or (k == 'CLine') or \
          (k == 'CSplit') or (k == 'Mod') or (k == 'HMod') or (k == 'GMode') else \
          "True or False" \
          if (k == 'AutoDAR') or (k == 'NoResize') or (k == 'ResizeWO') or (k == 'Borders') or \
          (k == 'CleanBorders') or (k == 'Frosty') or (k == 'Position') or (k == 'Info') else \
          "a string"

        if (VType == "an integer or float") or (VType == "an integer"):

            if not ((v[0] is None) or \
              ((isinstance(v[0], bool) is False) and isinstance(v[0], v[1]))):
                raise vs.Error(f"CropResize: {k} must be {VType}")

        else:

            if not ((v[0] is None) or isinstance(v[0], v[1])):
                raise vs.Error(f"CropResize: {k} must be {VType}")

# -------------------------------------------------------------------------------

    OutWidth = CR_Default(OutWidth, 0)
    OutHeight = CR_Default(OutHeight, 0)
    CPreview = CR_Default(CPreview, 0)
    CropDAR = CR_Default(CropDAR, 0.0)
    InDAR = CR_Default(InDAR, 0.0)
    InSAR = CR_Default(InSAR, 0.0)
    OutDAR = CR_Default(OutDAR, 0.0)
    OutSAR = CR_Default(OutSAR, 0.0)
    GMode = CR_Default(GMode, 1)
    IsMod = (Mod is not None)
    IsResizer = (Resizer is not None)

    if not (0 < GMode <= 2):
        raise vs.Error('GCropResize: GMode must be 1 or 2 (0, 1 or 2 for CropResize)')

    if (CropDAR < 0):
        raise vs.Error('GCropResize: CropDAR must be zero (undefined) or positive (4/3 or 1.3333 etc)')
    if (InDAR < 0):
        raise vs.Error('GCropResize: InDAR must be zero (undefined) or positive (16/9 or 1.7778 etc)')
    if (InSAR < 0):
        raise vs.Error('GCropResize: InSAR must be zero (undefined) or positive (16/9 or 1.7778 etc)')
    if (OutDAR < 0):
        raise vs.Error('GCropResize: OutDAR must be zero (undefined) or positive (64/45 or 1.4222 etc)')
    if (OutSAR < 0):
        raise vs.Error('GCropResize: OutSAR must be zero (undefined) or positive (64/45 or 1.4222 etc)')
    if (InDAR > 0) and (InSAR > 0):
        raise vs.Error('GCropResize: InDAR & InSAR cannot be specified together')
    if (OutDAR > 0) and (OutSAR > 0):
        raise vs.Error('CropResize: OutDAR & OutSAR cannot be specified together')

# -------------------------------------------------------------------------------
#            GMode1
# -------------------------------------------------------------------------------

    if (GMode == 1):

        global G1_PriorG1, \
        G1_OutWidth, G1_OutHeight, G1_CL, G1_CR, G1_CT, G1_CB, \
        G1_CropDAR, G1_TCRatio, G1_CSplit, G1_CPreview, G1_CLine, \
        G1_InDAR, G1_InSAR, G1_OutDAR, G1_OutSAR, G1_AutoDAR, \
        G1_Mod, G1_HMod, G1_NoResize, G1_ResizeWO, \
        G1_Borders, G1_CleanBorders, G1_BColor, G1_Frosty, \
        G1_ColorConvert, G1_ColorMode, \
        G1_Resizer, G1_RStr, G1_RingRepair, G1_Position, G1_Info

        try: IsPriorG1 = G1_PriorG1
        except: IsPriorG1 = False

        G1_PriorG1 = True

        G1_OutWidth = G1_OutWidth if (OutWidth == 0) and (IsPriorG1 is True) else OutWidth
        G1_OutHeight = G1_OutHeight if (OutHeight == 0) and (IsPriorG1 is True) else OutHeight

        G1_CL = CR_Default(CL, 0.0 if (IsPriorG1 is False) else G1_CL)
        G1_CR = CR_Default(CR, 0.0 if (IsPriorG1 is False) else G1_CR)
        G1_CT = CR_Default(CT, 0.0 if (IsPriorG1 is False) else G1_CT)
        G1_CB = CR_Default(CB, 0.0 if (IsPriorG1 is False) else G1_CB)

        G1_CropDAR = CR_Default(CropDAR, 0.0 if (IsPriorG1 is False) else G1_CropDAR)
        G1_TCRatio = CR_Default(TCRatio, 0.5 if (IsPriorG1 is False) else G1_TCRatio)
        G1_CSplit = CR_Default(CSplit, 1 if (IsPriorG1 is False) else G1_CSplit)
        G1_CPreview = CR_Default(CPreview, 0 if (IsPriorG1 is False) else G1_CPreview)
        G1_CLine = CR_Default(CLine, 1 if (IsPriorG1 is False) else G1_CLine)

        G1_InDAR = CR_Default(InDAR, 0.0 if ((InSAR > 0) or (IsPriorG1 is False)) else G1_InDAR)
        G1_InSAR = CR_Default(InSAR, 0.0 if ((InDAR > 0) or (IsPriorG1 is False)) else G1_InSAR)
        G1_OutDAR = CR_Default(OutDAR, 0.0 if ((OutSAR > 0) or (IsPriorG1 is False)) else G1_OutDAR)
        G1_OutSAR = CR_Default(OutSAR, 0.0 if ((OutDAR > 0) or (IsPriorG1 is False)) else G1_OutSAR)
        G1_AutoDAR = CR_Default(AutoDAR, CR_AutoDAR() if (IsPriorG1 is False) else G1_AutoDAR)

        G1_Mod = CR_Default(Mod, 4 if (IsPriorG1 is False) else G1_Mod)
        G1_HMod = CR_Default(HMod, Mod if (IsMod == True) else G1_Mod if (IsPriorG1 is False) else G1_HMod)

        G1_NoResize = CR_Default(NoResize, False if (IsPriorG1 is False) else G1_NoResize)
        G1_ResizeWO = CR_Default(ResizeWO, False if (IsPriorG1 is False) else G1_ResizeWO)

        G1_Borders = CR_Default(Borders, False if (IsPriorG1 is False) else G1_Borders)
        G1_CleanBorders = CR_Default(CleanBorders, False if (IsPriorG1 is False) else G1_CleanBorders)
        G1_BColor = CR_Default(BColor, None if (IsPriorG1 is False) else G1_BColor)
        G1_Frosty = CR_Default(Frosty, False if (IsPriorG1 is False) else G1_Frosty)

        G1_ColorConvert = CR_Default(ColorConvert, \
                          CR_ColorConvert() if (IsPriorG1 is False) else G1_ColorConvert)
        G1_ColorMode = CR_Default(ColorMode, "" if (IsPriorG1 is False) else G1_ColorMode)

        G1_Resizer = CR_Default(Resizer, CR_Resizer_Default() if (IsPriorG1 is False) else G1_Resizer)
        G1_RStr = CR_Default(RStr, "" if (IsResizer is True) else \
                  CR_RStr_Default() if (IsPriorG1 is False) else G1_RStr)
        G1_RingRepair = CR_Default(RingRepair, 0 if (IsPriorG1 is False) else G1_RingRepair)

        G1_Position = CR_Default(Position, False if (IsPriorG1 is False) else G1_Position)
        G1_Info = CR_Default(Info, False if (IsPriorG1 is False) else G1_Info)

# -------------------------------------------------------------------------------
#            GMode2
# -------------------------------------------------------------------------------

    elif (GMode == 2):

        try: IsPriorG1 = G1_PriorG1
        except: IsPriorG1 = False

        if (IsPriorG1 is False):
            raise vs.Error('GCropResize: For the first instance of GCropResize, ' + \
                           'GMode must equal one')

        global G2_PriorG2, \
        G2_OutWidth, G2_OutHeight, G2_CL, G2_CR, G2_CT, G2_CB, \
        G2_CropDAR, G2_TCRatio, G2_CSplit, G2_CPreview, G2_CLine, \
        G2_InDAR, G2_InSAR, G2_OutDAR, G2_OutSAR, G2_AutoDAR, \
        G2_Mod, G2_HMod, G2_NoResize, G2_ResizeWO, \
        G2_Borders, G2_CleanBorders, G2_BColor, G2_Frosty, \
        G2_ColorConvert, G2_ColorMode, \
        G2_Resizer, G2_RStr, G2_RingRepair, G2_Position, G2_Info

        try: IsPriorG2 = G2_PriorG2
        except: IsPriorG2 = False

        G2_PriorG2 = True

        G2_OutWidth = G2_OutWidth if (OutWidth == 0) and (IsPriorG2 is True) else OutWidth
        G2_OutHeight = G2_OutHeight if (OutHeight == 0) and (IsPriorG2 is True) else OutHeight

        G2_CL = CR_Default(CL, 0.0 if (IsPriorG2 is False) else G2_CL)
        G2_CR = CR_Default(CR, 0.0 if (IsPriorG2 is False) else G2_CR)
        G2_CT = CR_Default(CT, 0.0 if (IsPriorG2 is False) else G2_CT)
        G2_CB = CR_Default(CB, 0.0 if (IsPriorG2 is False) else G2_CB)

        G2_CropDAR = CR_Default(CropDAR, 0.0 if (IsPriorG2 is False) else G1_CropDAR)
        G2_TCRatio = CR_Default(TCRatio, 0.5 if (IsPriorG2 is False) else G2_TCRatio)
        G2_CSplit = CR_Default(CSplit, 1 if (IsPriorG2 is False) else G2_CSplit)
        G2_CPreview = CR_Default(CPreview, G1_CPreview if (IsPriorG2 is False) else G2_CPreview)
        G2_CLine = CR_Default(CLine, G1_CLine if (IsPriorG2 is False) else G2_CLine)

        G2_InDAR = CR_Default(InDAR, 0.0 if ((InSAR > 0) or (IsPriorG2 is False)) else G2_InDAR)
        G2_InSAR = CR_Default(InSAR, 0.0 if ((InDAR > 0) or (IsPriorG2 is False)) else G2_InSAR)
        G2_OutDAR = CR_Default(OutDAR, 0.0 if ((OutSAR > 0) or (IsPriorG2 is False)) else G2_OutDAR)
        G2_OutSAR = CR_Default(OutSAR, 0.0 if ((OutDAR > 0) or (IsPriorG2 is False)) else G2_OutSAR)
        G2_AutoDAR = CR_Default(AutoDAR, G1_AutoDAR if (IsPriorG2 is False) else G2_AutoDAR)

        G2_Mod = CR_Default(Mod, G1_Mod if (IsPriorG1 is False) else G2_Mod)
        G2_HMod = CR_Default(HMod, Mod if (IsMod == True) else G1_HMod if (IsPriorG2 is False) else G2_HMod)

        G2_NoResize = CR_Default(NoResize, False if (IsPriorG2 is False) else G2_NoResize)
        G2_ResizeWO = CR_Default(ResizeWO, False if (IsPriorG2 is False) else G2_ResizeWO)

        G2_Borders = CR_Default(Borders, False if (IsPriorG2 is False) else G2_Borders)
        G2_CleanBorders = CR_Default(CleanBorders, False if (IsPriorG2 is False) else G2_CleanBorders)
        G2_BColor = CR_Default(BColor, None if (IsPriorG2 is False) else G2_BColor)
        G2_Frosty = CR_Default(Frosty, False if (IsPriorG2 is False) else G2_Frosty)

        G2_ColorConvert = CR_Default(ColorConvert, \
                          CR_ColorConvert() if (IsPriorG2 is False) else G2_ColorConvert)
        G2_ColorMode = CR_Default(ColorMode, "" if (IsPriorG2 is False) else G2_ColorMode)

        G2_Resizer = CR_Default(Resizer, CR_Resizer_Default() if (IsPriorG2 is False) else G2_Resizer)
        G2_RStr = CR_Default(RStr, "" if (IsResizer is True) else \
                  CR_RStr_Default() if (IsPriorG2 is False) else G2_RStr)
        G2_RingRepair = CR_Default(RingRepair, G1_RingRepair if (IsPriorG2 is False) else G2_RingRepair)

        G2_Position = CR_Default(Position, False if (IsPriorG2 is False) else G2_Position)
        G2_Info = CR_Default(Info, G1_Info if (IsPriorG2 is False) else G2_Info)

# ===============================================================================
# ===============================================================================
#            Wrapper Functions
# ===============================================================================
# ===============================================================================

# -------------------------------------------------------------------------------
#            CropResizex
# -------------------------------------------------------------------------------

def CropResizex(Source, OutWidth=None, OutHeight=None, \
    CL=None, CR=None, CT=None, CB=None, \
    CropDAR=None, TCRatio=None, CSplit=None, \
    CPreview=None, CLine=None, \
    InDAR=None, InSAR=None, OutDAR=None, OutSAR=None, AutoDAR=None, \
    Mod=None, HMod=None, \
    NoResize=None, ResizeWO=None, \
    Borders=None, CleanBorders=None, BColor=None, Frosty=None, \
    ColorConvert=None, ColorMode=None, \
    Resizer=None, RStr=None, RingRepair=None, \
    GMode=None, Position=None, Info=None):

    global G_Resizer, G_RStr
    IsResizer = (Resizer is not None)
    G_Resizer = CR_Default(Resizer, CR_Resizer_x())
    G_RStr = CR_Default(RStr, ("" if (IsResizer is True) else CR_RStr_x()))

    return CropResize(Source, OutWidth, OutHeight, CL, CR, CT, CB, \
      CropDAR, TCRatio, CSplit, CPreview, CLine, InDAR, InSAR, OutDAR, OutSAR, AutoDAR, \
      Mod, HMod, NoResize, ResizeWO, Borders, CleanBorders, BColor, Frosty, \
      ColorConvert, ColorMode, Resizer, RStr, RingRepair, GMode, Position, Info)

# -------------------------------------------------------------------------------
#            pCropResize
# -------------------------------------------------------------------------------

def pCropResize(Source, OutWidth=None, OutHeight=None, \
    CL=None, CR=None, CT=None, CB=None, \
    CropDAR=None, TCRatio=None, CSplit=None, \
    CPreview=None, CLine=None, \
    InDAR=None, InSAR=None, OutDAR=None, OutSAR=None, AutoDAR=None, \
    Mod=None, HMod=None, \
    NoResize=None, ResizeWO=None, \
    Borders=None, CleanBorders=None, BColor=None, Frosty=None, \
    ColorConvert=None, ColorMode=None, \
    Resizer=None, RStr=None, RingRepair=None, \
    GMode=None, Position=None, Info=None):

    global G_CPreview
    G_CPreview = CR_Default(CPreview, CR_CPreview())

    return CropResize(Source, OutWidth, OutHeight, CL, CR, CT, CB, \
      CropDAR, TCRatio, CSplit, CPreview, CLine, InDAR, InSAR, OutDAR, OutSAR, AutoDAR, \
      Mod, HMod, NoResize, ResizeWO, Borders, CleanBorders, BColor, Frosty, \
      ColorConvert, ColorMode, Resizer, RStr, RingRepair, GMode, Position, Info)

# -------------------------------------------------------------------------------
#            iCropResize
# -------------------------------------------------------------------------------

def iCropResize(Source, OutWidth=None, OutHeight=None, \
    CL=None, CR=None, CT=None, CB=None, \
    CropDAR=None, TCRatio=None, CSplit=None, \
    CPreview=None, CLine=None, \
    InDAR=None, InSAR=None, OutDAR=None, OutSAR=None, AutoDAR=None, \
    Mod=None, HMod=None, \
    NoResize=None, ResizeWO=None, \
    Borders=None, CleanBorders=None, BColor=None, Frosty=None, \
    ColorConvert=None, ColorMode=None, \
    Resizer=None, RStr=None, RingRepair=None, \
    GMode=None, Position=None, Info=None):

    global G_Info
    G_Info = CR_Default(Info, True)

    return CropResize(Source, OutWidth, OutHeight, CL, CR, CT, CB, \
      CropDAR, TCRatio, CSplit, CPreview, CLine, InDAR, InSAR, OutDAR, OutSAR, AutoDAR, \
      Mod, HMod, NoResize, ResizeWO, Borders, CleanBorders, BColor, Frosty, \
      ColorConvert, ColorMode, Resizer, RStr, RingRepair, GMode, Position, Info)

# -------------------------------------------------------------------------------
#            piCropResize
# -------------------------------------------------------------------------------

def piCropResize(Source, OutWidth=None, OutHeight=None, \
    CL=None, CR=None, CT=None, CB=None, \
    CropDAR=None, TCRatio=None, CSplit=None, \
    CPreview=None, CLine=None, \
    InDAR=None, InSAR=None, OutDAR=None, OutSAR=None, AutoDAR=None, \
    Mod=None, HMod=None, \
    NoResize=None, ResizeWO=None, \
    Borders=None, CleanBorders=None, BColor=None, Frosty=None, \
    ColorConvert=None, ColorMode=None, \
    Resizer=None, RStr=None, RingRepair=None, \
    GMode=None, Position=None, Info=None):

    global G_CPreview, G_Info
    G_CPreview = CR_Default(CPreview, CR_CPreview())
    G_Info = CR_Default(Info, True)

    return CropResize(Source, OutWidth, OutHeight, CL, CR, CT, CB, \
      CropDAR, TCRatio, CSplit, CPreview, CLine, InDAR, InSAR, OutDAR, OutSAR, AutoDAR, \
      Mod, HMod, NoResize, ResizeWO, Borders, CleanBorders, BColor, Frosty, \
      ColorConvert, ColorMode, Resizer, RStr, RingRepair, GMode, Position, Info)

# -------------------------------------------------------------------------------
#            pCropResizex
# -------------------------------------------------------------------------------

def pCropResizex(Source, OutWidth=None, OutHeight=None, \
    CL=None, CR=None, CT=None, CB=None, \
    CropDAR=None, TCRatio=None, CSplit=None, \
    CPreview=None, CLine=None, \
    InDAR=None, InSAR=None, OutDAR=None, OutSAR=None, AutoDAR=None, \
    Mod=None, HMod=None, \
    NoResize=None, ResizeWO=None, \
    Borders=None, CleanBorders=None, BColor=None, Frosty=None, \
    ColorConvert=None, ColorMode=None, \
    Resizer=None, RStr=None, RingRepair=None, \
    GMode=None, Position=None, Info=None):

    global G_Resizer, G_RStr, G_CPreview
    IsResizer = (Resizer is not None)
    G_Resizer = CR_Default(Resizer, CR_Resizer_x())
    G_RStr = CR_Default(RStr, ("" if (IsResizer is True) else CR_RStr_x()))
    G_CPreview = CR_Default(CPreview, CR_CPreview())

    return CropResize(Source, OutWidth, OutHeight, CL, CR, CT, CB, \
      CropDAR, TCRatio, CSplit, CPreview, CLine, InDAR, InSAR, OutDAR, OutSAR, AutoDAR, \
      Mod, HMod, NoResize, ResizeWO, Borders, CleanBorders, BColor, Frosty, \
      ColorConvert, ColorMode, Resizer, RStr, RingRepair, GMode, Position, Info)

# -------------------------------------------------------------------------------
#            iCropResizex
# -------------------------------------------------------------------------------

def iCropResizex(Source, OutWidth=None, OutHeight=None, \
    CL=None, CR=None, CT=None, CB=None, \
    CropDAR=None, TCRatio=None, CSplit=None, \
    CPreview=None, CLine=None, \
    InDAR=None, InSAR=None, OutDAR=None, OutSAR=None, AutoDAR=None, \
    Mod=None, HMod=None, \
    NoResize=None, ResizeWO=None, \
    Borders=None, CleanBorders=None, BColor=None, Frosty=None, \
    ColorConvert=None, ColorMode=None, \
    Resizer=None, RStr=None, RingRepair=None, \
    GMode=None, Position=None, Info=None):

    global G_Resizer, G_RStr, G_Info
    IsResizer = (Resizer is not None)
    G_Resizer = CR_Default(Resizer, CR_Resizer_x())
    G_RStr = CR_Default(RStr, ("" if (IsResizer is True) else CR_RStr_x()))
    G_Info = CR_Default(Info, True)

    return CropResize(Source, OutWidth, OutHeight, CL, CR, CT, CB, \
      CropDAR, TCRatio, CSplit, CPreview, CLine, InDAR, InSAR, OutDAR, OutSAR, AutoDAR, \
      Mod, HMod, NoResize, ResizeWO, Borders, CleanBorders, BColor, Frosty, \
      ColorConvert, ColorMode, Resizer, RStr, RingRepair, GMode, Position, Info)

# -------------------------------------------------------------------------------
#            piCropResizex
# -------------------------------------------------------------------------------

def piCropResizex(Source, OutWidth=None, OutHeight=None, \
    CL=None, CR=None, CT=None, CB=None, \
    CropDAR=None, TCRatio=None, CSplit=None, \
    CPreview=None, CLine=None, \
    InDAR=None, InSAR=None, OutDAR=None, OutSAR=None, AutoDAR=None, \
    Mod=None, HMod=None, \
    NoResize=None, ResizeWO=None, \
    Borders=None, CleanBorders=None, BColor=None, Frosty=None, \
    ColorConvert=None, ColorMode=None, \
    Resizer=None, RStr=None, RingRepair=None, \
    GMode=None, Position=None, Info=None):

    global G_Resizer, G_RStr, G_CPreview, G_Info
    IsResizer = (Resizer is not None)
    G_Resizer = CR_Default(Resizer, CR_Resizer_x())
    G_RStr = CR_Default(RStr, ("" if (IsResizer is True) else CR_RStr_x()))
    G_CPreview = CR_Default(CPreview, CR_CPreview())
    G_Info = CR_Default(Info, True)

    return CropResize(Source, OutWidth, OutHeight, CL, CR, CT, CB, \
      CropDAR, TCRatio, CSplit, CPreview, CLine, InDAR, InSAR, OutDAR, OutSAR, AutoDAR, \
      Mod, HMod, NoResize, ResizeWO, Borders, CleanBorders, BColor, Frosty, \
      ColorConvert, ColorMode, Resizer, RStr, RingRepair, GMode, Position, Info)

# -------------------------------------------------------------------------------
#            GCropResizex
# -------------------------------------------------------------------------------

def GCropResizex(OutWidth=None, OutHeight=None, \
    CL=None, CR=None, CT=None, CB=None, \
    CropDAR=None, TCRatio=None, CSplit=None, \
    CPreview=None, CLine=None, \
    InDAR=None, InSAR=None, OutDAR=None, OutSAR=None, AutoDAR=None, \
    Mod=None, HMod=None, \
    NoResize=None, ResizeWO=None, \
    Borders=None, CleanBorders=None, BColor=None, Frosty=None, \
    ColorConvert=None, ColorMode=None, \
    Resizer=None, RStr=None, RingRepair=None, \
    GMode=None, Position=None, Info=None):

    IsResizer = (Resizer is not None)
    Resizer = CR_Default(Resizer, CR_Resizer_x())
    RStr = CR_Default(RStr, ("" if (IsResizer is True) else CR_RStr_x()))

    return GCropResize(OutWidth, OutHeight, CL, CR, CT, CB, \
      CropDAR, TCRatio, CSplit, CPreview, CLine, InDAR, InSAR, OutDAR, OutSAR, AutoDAR, \
      Mod, HMod, NoResize, ResizeWO, Borders, CleanBorders, BColor, Frosty, \
      ColorConvert, ColorMode, Resizer, RStr, RingRepair, GMode, Position, Info)

# -------------------------------------------------------------------------------
#            pGCropResize
# -------------------------------------------------------------------------------

def pGCropResize(OutWidth=None, OutHeight=None, \
    CL=None, CR=None, CT=None, CB=None, \
    CropDAR=None, TCRatio=None, CSplit=None, \
    CPreview=None, CLine=None, \
    InDAR=None, InSAR=None, OutDAR=None, OutSAR=None, AutoDAR=None, \
    Mod=None, HMod=None, \
    NoResize=None, ResizeWO=None, \
    Borders=None, CleanBorders=None, BColor=None, Frosty=None, \
    ColorConvert=None, ColorMode=None, \
    Resizer=None, RStr=None, RingRepair=None, \
    GMode=None, Position=None, Info=None):

    CPreview = CR_Default(CPreview, CR_CPreview())

    return GCropResize(OutWidth, OutHeight, CL, CR, CT, CB, \
      CropDAR, TCRatio, CSplit, CPreview, CLine, InDAR, InSAR, OutDAR, OutSAR, AutoDAR, \
      Mod, HMod, NoResize, ResizeWO, Borders, CleanBorders, BColor, Frosty, \
      ColorConvert, ColorMode, Resizer, RStr, RingRepair, GMode, Position, Info)

# -------------------------------------------------------------------------------
#            iGCropResize
# -------------------------------------------------------------------------------

def iGCropResize(OutWidth=None, OutHeight=None, \
    CL=None, CR=None, CT=None, CB=None, \
    CropDAR=None, TCRatio=None, CSplit=None, \
    CPreview=None, CLine=None, \
    InDAR=None, InSAR=None, OutDAR=None, OutSAR=None, AutoDAR=None, \
    Mod=None, HMod=None, \
    NoResize=None, ResizeWO=None, \
    Borders=None, CleanBorders=None, BColor=None, Frosty=None, \
    ColorConvert=None, ColorMode=None, \
    Resizer=None, RStr=None, RingRepair=None, \
    GMode=None, Position=None, Info=None):

    Info = CR_Default(Info, True)

    return GCropResize(OutWidth, OutHeight, CL, CR, CT, CB, \
      CropDAR, TCRatio, CSplit, CPreview, CLine, InDAR, InSAR, OutDAR, OutSAR, AutoDAR, \
      Mod, HMod, NoResize, ResizeWO, Borders, CleanBorders, BColor, Frosty, \
      ColorConvert, ColorMode, Resizer, RStr, RingRepair, GMode, Position, Info)

# -------------------------------------------------------------------------------
#            piGCropResize
# -------------------------------------------------------------------------------

def piGCropResize(OutWidth=None, OutHeight=None, \
    CL=None, CR=None, CT=None, CB=None, \
    CropDAR=None, TCRatio=None, CSplit=None, \
    CPreview=None, CLine=None, \
    InDAR=None, InSAR=None, OutDAR=None, OutSAR=None, AutoDAR=None, \
    Mod=None, HMod=None, \
    NoResize=None, ResizeWO=None, \
    Borders=None, CleanBorders=None, BColor=None, Frosty=None, \
    ColorConvert=None, ColorMode=None, \
    Resizer=None, RStr=None, RingRepair=None, \
    GMode=None, Position=None, Info=None):

    CPreview = CR_Default(CPreview, CR_CPreview())
    Info = CR_Default(Info, True)

    return GCropResize(OutWidth, OutHeight, CL, CR, CT, CB, \
      CropDAR, TCRatio, CSplit, CPreview, CLine, InDAR, InSAR, OutDAR, OutSAR, AutoDAR, \
      Mod, HMod, NoResize, ResizeWO, Borders, CleanBorders, BColor, Frosty, \
      ColorConvert, ColorMode, Resizer, RStr, RingRepair, GMode, Position, Info)

# -------------------------------------------------------------------------------
#            pGCropResizex
# -------------------------------------------------------------------------------

def pGCropResizex(OutWidth=None, OutHeight=None, \
    CL=None, CR=None, CT=None, CB=None, \
    CropDAR=None, TCRatio=None, CSplit=None, \
    CPreview=None, CLine=None, \
    InDAR=None, InSAR=None, OutDAR=None, OutSAR=None, AutoDAR=None, \
    Mod=None, HMod=None, \
    NoResize=None, ResizeWO=None, \
    Borders=None, CleanBorders=None, BColor=None, Frosty=None, \
    ColorConvert=None, ColorMode=None, \
    Resizer=None, RStr=None, RingRepair=None, \
    GMode=None, Position=None, Info=None):

    IsResizer = (Resizer is not None)
    Resizer = CR_Default(Resizer, CR_Resizer_x())
    RStr = CR_Default(RStr, ("" if (IsResizer is True) else CR_RStr_x()))
    CPreview = CR_Default(CPreview, CR_CPreview())

    return GCropResize(OutWidth, OutHeight, CL, CR, CT, CB, \
      CropDAR, TCRatio, CSplit, CPreview, CLine, InDAR, InSAR, OutDAR, OutSAR, AutoDAR, \
      Mod, HMod, NoResize, ResizeWO, Borders, CleanBorders, BColor, Frosty, \
      ColorConvert, ColorMode, Resizer, RStr, RingRepair, GMode, Position, Info)

# -------------------------------------------------------------------------------
#            iGCropResizex
# -------------------------------------------------------------------------------

def iGCropResizex(OutWidth=None, OutHeight=None, \
    CL=None, CR=None, CT=None, CB=None, \
    CropDAR=None, TCRatio=None, CSplit=None, \
    CPreview=None, CLine=None, \
    InDAR=None, InSAR=None, OutDAR=None, OutSAR=None, AutoDAR=None, \
    Mod=None, HMod=None, \
    NoResize=None, ResizeWO=None, \
    Borders=None, CleanBorders=None, BColor=None, Frosty=None, \
    ColorConvert=None, ColorMode=None, \
    Resizer=None, RStr=None, RingRepair=None, \
    GMode=None, Position=None, Info=None):

    IsResizer = (Resizer is not None)
    Resizer = CR_Default(Resizer, CR_Resizer_x())
    RStr = CR_Default(RStr, ("" if (IsResizer is True) else CR_RStr_x()))
    Info = CR_Default(Info, True)

    return GCropResize(OutWidth, OutHeight, CL, CR, CT, CB, \
      CropDAR, TCRatio, CSplit, CPreview, CLine, InDAR, InSAR, OutDAR, OutSAR, AutoDAR, \
      Mod, HMod, NoResize, ResizeWO, Borders, CleanBorders, BColor, Frosty, \
      ColorConvert, ColorMode, Resizer, RStr, RingRepair, GMode, Position, Info)

# -------------------------------------------------------------------------------
#            piGCropResizex
# -------------------------------------------------------------------------------

def piGCropResizex(OutWidth=None, OutHeight=None, \
    CL=None, CR=None, CT=None, CB=None, \
    CropDAR=None, TCRatio=None, CSplit=None, \
    CPreview=None, CLine=None, \
    InDAR=None, InSAR=None, OutDAR=None, OutSAR=None, AutoDAR=None, \
    Mod=None, HMod=None, \
    NoResize=None, ResizeWO=None, \
    Borders=None, CleanBorders=None, BColor=None, Frosty=None, \
    ColorConvert=None, ColorMode=None, \
    Resizer=None, RStr=None, RingRepair=None, \
    GMode=None, Position=None, Info=None):

    IsResizer = (Resizer is not None)
    Resizer = CR_Default(Resizer, CR_Resizer_x())
    RStr = CR_Default(RStr, ("" if (IsResizer is True) else CR_RStr_x()))
    CPreview = CR_Default(CPreview, CR_CPreview())
    Info = CR_Default(Info, True)

    return GCropResize(OutWidth, OutHeight, CL, CR, CT, CB, \
      CropDAR, TCRatio, CSplit, CPreview, CLine, InDAR, InSAR, OutDAR, OutSAR, AutoDAR, \
      Mod, HMod, NoResize, ResizeWO, Borders, CleanBorders, BColor, Frosty, \
      ColorConvert, ColorMode, Resizer, RStr, RingRepair, GMode, Position, Info)

# ===============================================================================
# ===============================================================================
#            Global FrostyB Function
# ===============================================================================
# ===============================================================================

def FrostyB(Frosty=None, FrostyStatic=None, Clone=None, \
    FrostyGradient=None, FrostyPlain=None, Plain=None, \
    BColor=None, BDivider=None, BDColor=None, Feather=None, \
    Bright=None, Cont=None, Sat=None, Blend=None, \
    Blur=None, VBlur=None, Iterations=None, \
    Texture=None, TextureC=None, StaticT=None, TSoft=None, Faster=None):

# -------------------------------------------------------------------------------

    kwargsdict = {'Frosty': [Frosty, bool],
      'FrostyStatic': [FrostyStatic, int], 'Clone': [Clone, int],
      'FrostyGradient': [FrostyGradient, bool], 'FrostyPlain': [FrostyPlain, bool],
      'Plain': [Plain, bool], 'BDivider': [BDivider, int], 'Feather': [Feather, int],
      'Bright': [Bright, (int, float)], 'Cont': [Cont, (int, float)],
      'Sat': [Sat, (int, float)], 'Blend': [Blend, (int, float)],
      'Blur': [Blur, (int, float)], 'VBlur': [VBlur, (int, float)],
      'Iterations': [Iterations, int], 'Texture': [Texture, (int, float)],
      'TextureC': [TextureC, (int, float)], 'StaticT': [StaticT, bool],
      'TSoft': [TSoft, int], 'Faster': [Faster, bool]}

    for k, v in kwargsdict.items():

        VType = "an integer or float" \
          if (k == 'Bright') or (k == 'Cont') or (k == 'Sat') or (k == 'Blend') or \
          (k == 'Blur') or (k == 'VBlur') or (k == 'Texture') or (k == 'TextureC') else \
          "an integer" \
          if (k == 'FrostyStatic') or (k == 'Clone') or (k == 'BDivider') or \
          (k == 'Feather') or (k == 'Iterations') or (k == 'TSoft') else \
          "True or False" \
          if (k == 'Frosty') or (k == 'FrostyGradient') or (k == 'FrostyPlain') or \
          (k == 'Plain') or (k == 'StaticT') or (k == 'Faster') else \
          "a string"

        if (VType == "an integer or float") or (VType == "an integer"):

            if not ((v[0] is None) or \
              ((isinstance(v[0], bool) is False) and isinstance(v[0], v[1]))):
                raise vs.Error(f"CropResize: {k} must be {VType}")

        else:

            if not ((v[0] is None) or isinstance(v[0], v[1])):
                raise vs.Error(f"CropResize: {k} must be {VType}")

# -------------------------------------------------------------------------------

    global G_CRFB, \
    CRFB_Frosty, CRFB_FrostyStatic, CRFB_Clone, CRFB_FrostyGradient, \
    CRFB_FrostyPlain, CRFB_Plain, CRFB_BColor, CRFB_BDivider, CRFB_BDColor, \
    CRFB_Feather, CRFB_Bright, CRFB_Cont, CRFB_Sat, CRFB_Blend, \
    CRFB_Blur, CRFB_VBlur, CRFB_Iterations, CRFB_Deband, \
    CRFB_Texture, CRFB_TextureC, CRFB_StaticT, CRFB_TSoft, CRFB_Faster

    G_CRFB = True
    CRFB_Frosty = CR_Default(Frosty, True)
    CRFB_FrostyStatic = FrostyStatic
    CRFB_Clone = CR_Default(Clone, 0)
    CRFB_FrostyGradient = CR_Default(FrostyGradient, False)
    CRFB_FrostyPlain = CR_Default(FrostyPlain, False)
    CRFB_Plain = CR_Default(Plain, False)
    CRFB_BColor = BColor
    CRFB_BDivider = CR_Default(BDivider, 0)
    CRFB_BDColor = BDColor
    CRFB_Feather = CR_Default(Feather, 0)
    CRFB_Bright = Bright
    CRFB_Cont = Cont
    CRFB_Sat = Sat
    CRFB_Blend = Blend
    CRFB_Blur = Blur
    CRFB_VBlur = VBlur
    CRFB_Iterations = CR_Default(Iterations, 3)
    CRFB_Texture = Texture
    CRFB_TextureC = TextureC
    CRFB_StaticT = CR_Default(StaticT, True)
    CRFB_TSoft = TSoft
    CRFB_Faster = CR_Default(Faster, False)

# ===============================================================================
# ===============================================================================
#            Frosty Borders Function
# ===============================================================================
# ===============================================================================

def CR_FrostyBorders(CRVideo, CRVWidth, CRVHeight, \
    F_CRVWidth, F_CRVHeight, BL, BT, BR, BB, WidthMod, HeightMod, \
    F_WidthMod, F_HeightMod, CRVideo_StorageAR, Final_StorageAR, \
    Source_Bits, ChromaSS, SCFamily):

# -------------------------------------------------------------------------------
#            Defaults
# -------------------------------------------------------------------------------

    IsCRFB = ('G_CRFB' in globals())
    IsPillarBox = (Final_StorageAR >= CRVideo_StorageAR)
    TotalFrames = CRVideo.num_frames
    LastFrame = TotalFrames - 1
    F_VidSize = F_CRVHeight if (IsPillarBox is True) else F_CRVWidth
    Scale = 2 ** (Source_Bits - 8)
    MinIn = 16 * Scale
    MaxInL = 235 * Scale
    MaxInC = 240 * Scale
    MaxOut = 255 * Scale

    Frosty = CRFB_Frosty if (IsCRFB is True) else True
    IsFrostyStatic = True if (IsCRFB is True) and (CRFB_FrostyStatic is not None) else False
    FrostyStatic = CRFB_FrostyStatic if (IsFrostyStatic is True) else 0
    Clone = CRFB_Clone if (IsCRFB is True) else 0
    FrostyGradient = CRFB_FrostyGradient if (IsCRFB is True) else False
    FrostyPlain = CRFB_FrostyPlain if (IsCRFB is True) else False
    Plain = CRFB_Plain if (IsCRFB is True) else False
    BDivider = CRFB_BDivider if (IsCRFB is True) else 0
    BDColor = CRFB_BDColor if (IsCRFB is True) else None
    Feather = CRFB_Feather if (IsCRFB is True) else 0
    IsBlur = True if (IsCRFB is True) and (CRFB_Blur is not None) else False
    IsVBlur = True if (IsCRFB is True) and (CRFB_VBlur is not None) else False
    Iterations = CRFB_Iterations if (IsCRFB is True) else 3
    StaticT = CRFB_StaticT if (IsCRFB is True) else True
    DynamicT = False if (StaticT is True) else True
    Faster = CRFB_Faster if (IsCRFB is True) else False

    IsMod4 = ((IsPillarBox is True) and (WidthMod == 4)) or \
      ((IsPillarBox is False) and (HeightMod == 4))
    BorderMin = 32 if ((Frosty is True) or (IsFrostyStatic is True) or \
      (FrostyGradient is True)) and (IsMod4 is True) else \
      16 if (Frosty is True) or (IsFrostyStatic is True) or (FrostyGradient is True) else \
      16 if ((Clone == 1) or (Clone == 2) or (Clone == 3)) and (IsMod4 is True) else \
      8 if (Clone == 1) or (Clone == 2) or (Clone == 3) else \
      WidthMod if (IsPillarBox is True) else HeightMod
    SizeTest = ((IsPillarBox is True) and (BL >= BorderMin <= BR)) or \
      ((IsPillarBox is False) and (BT >= BorderMin <= BB))
    IsBDivider = (BDivider > 0) and \
      (((IsPillarBox is True) and (BL > BorderMin < BR)) or \
      ((IsPillarBox is False) and (BT > BorderMin < BB)))
    IsFeather = (Feather > 0) and \
      (((IsPillarBox is True) and (BL >= WidthMod <= BR)) or \
      ((IsPillarBox is False) and (BT >= HeightMod <= BB)))

    FrostyPlain = (FrostyPlain is True) and (Plain is False)
    FrostyGradient = (FrostyGradient is True) and (FrostyPlain is False) and \
      (Plain is False) and (SizeTest is True)
    Clone1 = (Clone == 1) or ((Clone == 2) and (IsFeather is False)) and (FrostyGradient is False) and \
      (FrostyPlain is False) and (Plain is False) and (SizeTest is True)
    Clone2 = (Clone == 2) and (IsFeather is True) and (FrostyGradient is False) and \
      (FrostyPlain is False) and (Plain is False) and (SizeTest is True)
    Clone3 = (Clone == 3) and (FrostyGradient is False) and \
      (FrostyPlain is False) and (Plain is False) and (SizeTest is True)
    IsClone = (Clone1 is True) or (Clone2 is True) or (Clone3 is True)
    IsFrostyStatic = (IsFrostyStatic is True) and (FrostyGradient is False) and \
      (FrostyPlain is False) and (Plain is False) and (IsClone is False) and (SizeTest is True)
    Frosty = (Frosty is True) and (FrostyGradient is False) and (FrostyPlain is False) and \
      (Plain is False) and (IsClone is False) and (SizeTest is True)
    IsFrosty = (Frosty is True) or (IsFrostyStatic is True)
    Plain = (Plain is True) or ((FrostyGradient is False) and (FrostyPlain is False) and \
      (IsClone is False) and (IsFrosty is False))
    IsPlain = (Plain is True) or (FrostyPlain is True)

    Feather = 0 if (IsFeather is False) else \
      min(WidthMod if (Feather < WidthMod) else Feather, \
      m.floor(F_CRVWidth / 6.0 / F_WidthMod) * WidthMod) \
      if (IsPillarBox is True) else \
      min(HeightMod if (Feather < HeightMod) else Feather, \
      m.floor(F_CRVHeight / 6.0 / F_HeightMod) * HeightMod)

    BColor = CRFB_BColor if (IsCRFB is True) and (CRFB_BColor is not None) else \
      RGBColor(CRVideo, '008080') if (FrostyGradient is True) or (FrostyPlain is True) else None

    Bright = CRFB_Bright if (IsCRFB is True) and (CRFB_Bright is not None) else \
      0.0 if (IsPillarBox is True) or (Plain is True) else -25.0
    Cont = CRFB_Cont if (IsCRFB is True) and (CRFB_Cont is not None) else \
      1.0 if (IsPillarBox is True) or (Plain is True) else 0.66
    Sat = CRFB_Sat if (IsCRFB is True) and (CRFB_Sat is not None) else \
      1.0 if (Plain is True) else 0.85 if (IsPillarBox is True) else 0.66
    Blend = CRFB_Blend if (IsCRFB is True) and (CRFB_Blend is not None) else \
      0.5 if (IsFrostyStatic is True) else 0.2 if (IsFrosty is True) else 0.0
    IsTweak = (Bright != 0.0) or (Cont != 1.0) or (Sat != 1.0)

    Blur = CRFB_Blur if (IsBlur is True) and (CRFB_Blur >= 0) else \
      -CRFB_Blur * F_VidSize / 720.0  if (IsBlur is True) else \
      F_VidSize / 75.0 if (Clone1 is True) or (Clone2 is True) else \
      F_VidSize / 25.0 if (Clone3 is True) else 0
    VBlur = CRFB_VBlur if (IsVBlur is True) and (CRFB_VBlur >= 0) else \
      -CRFB_VBlur * F_VidSize / 720.0 if (IsVBlur is True) else \
      Blur if (IsBlur is True) or (IsClone is True) else 0

    Texture = round((CRFB_Texture if (IsCRFB is True) and (CRFB_Texture is not None) else \
      2.5 if (IsClone is True) else 5.0 if (Plain is False) else 0) * 64.0)
    TextureC = round((CRFB_TextureC if (IsCRFB is True) and (CRFB_TextureC is not None) else 0) * 64.0)
    TSoft = CRFB_TSoft if (IsCRFB is True) and (CRFB_TSoft is not None) else \
      5 if (Frosty is True) else 1 if (Clone3 is True) else 0
    IsTSoft = (TSoft > 0) and ((Frosty is True) or (IsClone is True))

# ===============================================================================
#            Insist
# ===============================================================================

    if (CRVideo.format.color_family == vs.RGB):
        raise vs.Error('CropResize: RGB is not supported for FrostyBorders')
    if not (CRVideo.format.sample_type == 0):
        raise vs.Error('CropResize: Only integer bit depths ' + \
                       'are supported for FrostyBorders')
    if (CRVideo.format is None):
        raise vs.Error('CropResize: Only clips with a constant format ' + \
                       'are supported for FrostyBorders')

    if not (0 <= FrostyStatic <= LastFrame):
        raise vs.Error('CropResize: The valid range for FrostyStatic ' + \
                       'is from zero to the final frame number (' + str(LastFrame) + ')')

    if not ((IsPillarBox is False) or (BDivider <= 0) or (WidthMod != 2) or (BDivider % 2 == 0)):
        raise vs.Error('CropResize: ' + str(CRVideo.format.name) + \
                       ' requires a Mod2 BDivider for pillarbox borders')
    if not ((IsPillarBox is False) or (BDivider <= 0) or (WidthMod != 4) or (BDivider % 4 == 0)):
        raise vs.Error('CropResize: ' + str(CRVideo.format.name) + \
                       ' requires a Mod4 BDivider for pillarbox borders')
    if not ((IsPillarBox is True) or (BDivider <= 0) or (HeightMod != 2) or (BDivider % 2 == 0)):
        raise vs.Error('CropResize: ' + str(CRVideo.format.name) + \
                       ' requires a Mod2 BDivider for letterbox borders')
    if not ((IsPillarBox is True) or (BDivider <= 0) or (HeightMod != 4) or (BDivider % 4 == 0)):
        raise vs.Error('CropResize: ' + str(CRVideo.format.name) + \
                       ' requires a Mod4 BDivider for letterbox borders')

    if not ((Feather == 0) or (BDivider <= 0)):
        raise vs.Error('CropResize: Feather and BDivider cannot be used together ' + \
                       '(one must equal zero)')
    if not ((Feather == 0) or (FrostyGradient is False)):
        raise vs.Error('CropResize: ' + \
                       'Feather cannot be greater than zero when FrostyGradient=true')

    if not (0 <= Clone <= 3):
        raise vs.Error('CropResize: Clone can only be 0 (disabled), 1, 2 or 3')
    if not (-255 <= Bright <= 255):
        raise vs.Error('CropResize: The valid range for Bright is -255 to 255')
    if not (0 <= Cont <= 10):
        raise vs.Error('CropResize: The valid range for Cont is 0.0 to 10.0')
    if not (0 <= Sat <= 10):
        raise vs.Error('CropResize: The valid range for Sat is 0.0 to 10.0')
    if not (0 <= Blend <= 1):
        raise vs.Error('CropResize: The valid range for Blend is 0.0 to 1.0')
    if not (0 <= Texture) and (0 <= TextureC):
        raise vs.Error('CropResize: Texture & TextureC cannot be negative')
    if not (0 <= TSoft):
        raise vs.Error('CropResize: The valid range for TSoft is 0 to 7')

# ===============================================================================
#            PillarBox Borders
# ===============================================================================

    if (IsPillarBox is True):

        BL = BL - BDivider if (IsBDivider is True) else BL
        BR = BR - BDivider if (IsBDivider is True) else BR
        FL = 0 if (IsFeather is False) else \
          max(BL, m.floor((BL + (Feather * 3)) / F_WidthMod) * WidthMod)
        FR = 0 if (IsFeather is False) else \
          max(BR, m.floor((BR + (Feather * 3)) / F_WidthMod) * WidthMod)

# -------------------------------------------------------------------------------
#            PillarBox Strings
# -------------------------------------------------------------------------------

        CRVideoStr = ",CRVideo," if (IsBDivider is False) else \
          ",CRVideo.std.AddBorders(BDivider, BDivider, 0, 0, color=BDColor),"

        BlurStr1 = \
          ".std.BoxBlur(hradius=Blur, vradius=VBlur, hpasses=Iterations, vpasses=Iterations)"
        BlurStr2 = \
          ".std.BoxBlur(hradius=Blur, vradius=VBlur, " + \
          "hpasses=Iterations, vpasses=Iterations, planes=[0])"
        BlurStr3 = \
          ".std.BoxBlur(hradius=ChromaBlur, vradius=VChromaBlur, " + \
          "hpasses=Iterations, vpasses=Iterations, planes=[1,2])"

        FBordStr = ".std.AddBorders(0, CRVWidth-FL+BL-FR+BR, 0, 0),"
        BicubicStr = ".resize.Bicubic"

        BicubicLStr = "" if (IsPlain is True) else BicubicStr + "(BL, CRVHeight)"
        BicubicRStr = "" if (IsPlain is True) else BicubicStr + "(BR, CRVHeight)"
        FBicubicLStr = "" if (IsFeather is False) else BicubicStr + "(FL, CRVHeight)"
        FBicubicRStr = "" if (IsFeather is False) else BicubicStr + "(FR, CRVHeight)"

        Bicubic3LStr = \
          BicubicStr + "(max(BL, m.floor(WidthCloneBorderL * RatioL / F_WidthMod) * WidthMod), \
          max(CRVHeight, m.floor(CRVHeight * RatioL / F_HeightMod) * HeightMod))" \
          if (Clone3 is True) else ""
        Bicubic3RStr = \
          BicubicStr + "(max(BR, m.floor(WidthCloneBorderR * RatioR / F_WidthMod) * WidthMod), \
          max(CRVHeight, m.floor(CRVHeight * RatioR / F_HeightMod) * HeightMod))" \
          if (Clone3 is True) else ""

        GrainStr = "" if (Texture == 0) and (TextureC == 0) else \
          ".f3kdb.Deband(grainy=Texture, grainc=TextureC, dither_algo=1, " + \
          "dynamic_grain=DynamicT, output_depth=Source_Bits)"

        LoopStr = ".std.Trim(1, 1).std.Loop(TotalFrames)"
        SLoopStr = ".std.Trim(FrostyStatic, last=FrostyStatic).std.Loop(TotalFrames)" \
          if (IsFrostyStatic is True) else ""
        TSoftStr = ".focus2.TemporalSoften2(TSoft, 255, 255, 20, 2)" if (IsTSoft is True) else ""

# -------------------------------------------------------------------------------
#            PillarBox Frosty Borders
# -------------------------------------------------------------------------------

        FrostyCrop = CRVWidth - max(BorderMin, \
          round(F_CRVWidth / (5.0 if (Faster is False) else 10.0) / F_WidthMod) * WidthMod) \
          if (IsFrosty is True) and (IsTSoft is True) else ""

        FrostyLR = eval("core.std.StackHorizontal([CRVideo.std.Crop(0, FrostyCrop, 0, 0), \
          CRVideo.std.Crop(FrostyCrop, 0, 0, 0)])" + TSoftStr) \
          if (IsFrosty is True) and (IsTSoft is True) else ""

        WidthFrostyLR = FrostyLR.width if (IsFrosty is True) and (IsTSoft is True) else 0

        FrostyCrop = WidthFrostyLR - max(BorderMin, \
          round(float(WidthFrostyLR) / (16.0 if (Faster is False) else 8.0) / F_WidthMod) * WidthMod) \
          if (IsFrosty is True) and (IsTSoft is True) else \
          CRVWidth - max(BorderMin, round(F_CRVWidth / 80.0 / F_WidthMod) * WidthMod) \
          if (IsFrosty is True) else 0

        if (IsFrosty is True) and (IsTSoft is True):

            FrostyL = FrostyLR.std.Crop(0, FrostyCrop, 0, 0)
            FrostyR = FrostyLR.std.Crop(FrostyCrop, 0, 0, 0)

        elif (IsFrosty is True):

            FrostyL = CRVideo.std.Crop(0, FrostyCrop, 0, 0)
            FrostyR = CRVideo.std.Crop(FrostyCrop, 0, 0, 0)

        WidthFrostyL = FrostyL.width if (IsFrosty is True) else 0
        WidthFrostyR = FrostyR.width if (IsFrosty is True) else 0

        if (IsFrosty is True):

            FrostyBlendL = FrostyL if (Blend == 0) else \
              FrostyR if (Blend == 1) else \
              core.std.Merge(FrostyL, core.std.FlipHorizontal(FrostyR), weight=Blend)
            FrostyBlendR = FrostyR if (Blend == 0) else \
              FrostyL if (Blend == 1) else \
              core.std.Merge(FrostyR, core.std.FlipHorizontal(FrostyL), weight=Blend)

        Blur = round(Blur if (IsBlur is True) else \
          max(WidthFrostyL, WidthFrostyR) + (0 if (IsFeather is False) else Feather) \
          if (IsFrosty is True) else Blur)
        VBlur = round(VBlur if (IsVBlur is True) else \
          F_CRVHeight / 10.0 if (IsFrosty is True) else VBlur)

        ChromaBlur = Blur if (CRVideo.format.subsampling_w == 0) else \
          round(Blur / 2.0  if (CRVideo.format.subsampling_w == 1) else Blur / 4.0)
        VChromaBlur = VBlur if (CRVideo.format.subsampling_h == 0) else \
          round(VBlur / 2.0 if (CRVideo.format.subsampling_h == 1) else VBlur / 4.0)

        BlurStr1 = "" if (Blur == 0 == VBlur) or (Iterations == 0) or \
          (FrostyGradient is True) or ((IsPlain is True) and (IsFeather is False)) else BlurStr1
        BlurStr2 = "" if (Blur == 0 == VBlur) or (Iterations == 0) or \
          (FrostyGradient is True) or ((IsPlain is True) and (IsFeather is False)) else BlurStr2
        BlurStr3 = "" if (ChromaBlur == 0 == VChromaBlur) or (Iterations == 0) or \
          (FrostyGradient is True) or ((IsPlain is True) and (IsFeather is False)) else BlurStr3

        BlurStr = BlurStr1 if (ChromaSS == "444") or (SCFamily == vs.GRAY) else \
          BlurStr2 + BlurStr3

        if (IsFrosty is True) and (IsTweak is True):

            FrostyL = eval("Tweak(FrostyBlendL, bright=Bright, cont=Cont, sat=Sat)" + \
              BlurStr + BicubicLStr + GrainStr + SLoopStr)
            FrostyR = eval("Tweak(FrostyBlendR, bright=Bright, cont=Cont, sat=Sat)" + \
              BlurStr + BicubicRStr + GrainStr + SLoopStr)

        elif (IsFrosty is True):

            FrostyL = eval("FrostyBlendL" + BlurStr + BicubicLStr + GrainStr + SLoopStr)
            FrostyR = eval("FrostyBlendR" + BlurStr + BicubicRStr + GrainStr + SLoopStr)

        if (IsFrosty is True) and (IsFeather is True) and (IsTweak is True):

            FFrostyL = eval("Tweak(FrostyBlendL, bright=Bright, cont=Cont, sat=Sat)" + \
              BlurStr + FBicubicLStr + GrainStr + SLoopStr)
            FFrostyR = eval("Tweak(FrostyBlendR, bright=Bright, cont=Cont, sat=Sat)" + \
              BlurStr + FBicubicRStr + GrainStr + SLoopStr)

        elif (IsFrosty is True) and (IsFeather is True):

            FFrostyL = eval("FrostyBlendL" + BlurStr + FBicubicLStr + GrainStr + SLoopStr)
            FFrostyR = eval("FrostyBlendR" + BlurStr + FBicubicRStr + GrainStr + SLoopStr)

# -------------------------------------------------------------------------------
#            PillarBox Clone Borders
# -------------------------------------------------------------------------------

        if (IsClone is True) or ((IsFrosty is True) and (IsFeather is True)):

            MaxCloneC = m.floor(F_CRVWidth / 2.0 / F_WidthMod) * WidthMod
            CloneCropL = max(MaxCloneC, CRVWidth - BL)
            CloneCropR = max(MaxCloneC, CRVWidth - BR)
            CloneBorderL = CRVideo.std.Crop(0, CloneCropL, 0, 0)
            CloneBorderR = CRVideo.std.Crop(CloneCropR, 0, 0, 0)
            WidthCloneBorderL = CloneBorderL.width
            WidthCloneBorderR = CloneBorderR.width

            FlipCloneL = core.std.FlipHorizontal(CloneBorderL if (WidthCloneBorderL == BL) else \
              eval("CloneBorderL" + BicubicLStr))
            FlipCloneR = core.std.FlipHorizontal(CloneBorderR if (WidthCloneBorderR == BR) else \
              eval("CloneBorderR" + BicubicRStr))

        if (Clone1 is True) or (Clone3 is True):

            if (IsTSoft is True):

                CloneLR = eval("core.std.StackHorizontal([CloneBorderL, CloneBorderR])" + TSoftStr)
                WidthCloneLR = CloneLR.width

            CloneL = CloneLR.std.Crop(0, WidthCloneLR-WidthCloneBorderL, 0, 0) \
              if (IsTSoft is True) else CloneBorderL
            CloneR = CloneLR.std.Crop(WidthCloneLR-WidthCloneBorderR, 0, 0, 0) \
              if (IsTSoft is True) else CloneBorderR
            CloneBlendL = CloneL if (Blend == 0) else CloneR if (Blend == 1) else \
              Overlay(CloneL, FlipHorizontal(CloneR), Opacity=Blend)
            CloneBlendR = CloneR if (Blend == 0) else CloneL if (Blend == 1) else \
              Overlay(CloneR, FlipHorizontal(CloneL), Opacity=Blend)

        if (Clone1 is True):

            if (IsFeather is True):

                FCloneL = CloneBlendL if (WidthCloneBorderL == BL) else \
                  eval("CloneBlendL" + BicubicLStr)
                FCloneR = CloneBlendR if (WidthCloneBorderR == BR) else \
                  eval("CloneBlendR" + BicubicRStr)

            if (IsFeather is False) and (IsTweak is True):

                CloneL = eval("Tweak(CloneBlendL, bright=Bright, cont=Cont, sat=Sat)" + \
                  BlurStr + GrainStr) \
                  if (WidthCloneBorderL == BL) else \
                  eval("Tweak(CloneBlendL, bright=Bright, cont=Cont, sat=Sat)" + \
                  BlurStr + BicubicLStr + GrainStr)
                CloneR = eval("Tweak(CloneBlendR, bright=Bright, cont=Cont, sat=Sat)" + \
                  BlurStr + GrainStr) \
                  if (WidthCloneBorderR == BR) else \
                  eval("Tweak(CloneBlendR, bright=Bright, cont=Cont, sat=Sat)" + \
                  BlurStr + BicubicRStr + GrainStr)

            elif (IsFeather is False):

                CloneL = eval("CloneBlendL" + BlurStr) \
                  if (WidthCloneBorderL == BL) else \
                  eval("CloneBlendL" + BicubicLStr + GrainStr)
                CloneR = eval("CloneBlendR" + BlurStr) \
                  if (WidthCloneBorderR == BR) else \
                  eval("CloneBlendR" + BicubicRStr + GrainStr)

# -------------------------------------------------------------------------------
#            PillarBox Clone2 Borders
# -------------------------------------------------------------------------------

        if (Clone2 is True):

            Clone2CropL = max(MaxCloneC, CRVWidth - FL)
            Clone2CropR = max(MaxCloneC, CRVWidth - FR)
            Clone2BL = CRVideo.std.Crop(0, Clone2CropL, 0, 0)
            Clone2BR = CRVideo.std.Crop(Clone2CropR, 0, 0, 0)
            WidthClone2BL = Clone2BL.width
            WidthClone2BR = Clone2BR.width

            FlipClone2L = core.std.FlipHorizontal(Clone2BL if (WidthClone2BL == FL) else \
              eval("Clone2BL" + FBicubicLStr))
            FlipClone2R = core.std.FlipHorizontal(Clone2BR if (WidthClone2BR == FR) else \
              eval("Clone2BR" + FBicubicRStr))

            if (IsTSoft is True):

                Clone2LR = eval("core.std.StackHorizontal([Clone2BL, Clone2BR])" + TSoftStr)
                WidthClone2LR = Clone2LR.width if (IsTSoft is True) else 0

            Clone2L = Clone2LR.std.Crop(0, WidthClone2LR-WidthClone2BL, 0, 0) \
              if (IsTSoft is True) else Clone2BL
            Clone2R = Clone2LR.std.Crop(WidthClone2LR-WidthClone2BR, 0, 0, 0) \
              if (IsTSoft is True) else Clone2BR
            Clone2BlendL = Clone2L if (Blend == 0) else Clone2R if (Blend == 1) else \
              core.std.Merge(Clone2L, core.std.FlipHorizontal(Clone2R), weight=Blend)
            Clone2BlendR = Clone2R if (Blend == 0) else Clone2L if (Blend == 1) else \
              core.std.Merge(Clone2R, core.std.FlipHorizontal(Clone2L), weight=Blend)

            if (IsTweak is True):

                Clone2L = eval("Tweak(Clone2BlendL, bright=Bright, cont=Cont, sat=Sat)" + \
                  BlurStr + GrainStr) \
                  if (WidthClone2BL == FL) else \
                  eval("Tweak(Clone2BlendL, bright=Bright, cont=Cont, sat=Sat)" + \
                  BlurStr + FBicubicLStr + GrainStr)
                Clone2R = eval("Tweak(Clone2BlendR, bright=Bright, cont=Cont, sat=Sat)" + \
                  BlurStr + GrainStr) \
                  if (WidthClone2BR == FR) else \
                  eval("Tweak(Clone2BlendR, bright=Bright, cont=Cont, sat=Sat)" + \
                  BlurStr + FBicubicRStr + GrainStr)

            else:

                Clone2L = eval("Clone2BlendL" + BlurStr + GrainStr) \
                  if (WidthClone2BL == FL) else \
                  eval("Clone2BlendL" + BlurStr + FBicubicLStr + GrainStr)
                Clone2R = eval("Clone2BlendR" + BlurStr + GrainStr) \
                  if (WidthClone2BR == FR) else \
                  eval("Clone2BlendR" + BlurStr + FBicubicRStr + GrainStr)

# -------------------------------------------------------------------------------
#            PillarBox Clone3 Borders
# -------------------------------------------------------------------------------

        if (Clone3 is True):

            RatioL = 1.618 if ((WidthCloneBorderL - BL) >= 0) else \
              float(BL) / float(WidthCloneBorderL)
            RatioR = 1.618 if ((WidthCloneBorderR - BR) >= 0) else \
              float(BR) / float(WidthCloneBorderR)

            Clone3L = eval("CloneBlendL" + Bicubic3LStr)
            Clone3R = eval("CloneBlendR" + Bicubic3RStr)
            WidthClone3L = Clone3L.width
            HeightClone3L = Clone3L.height
            WidthClone3R = Clone3R.width
            HeightClone3R = Clone3R.height

            Clone3L_CL = max(0, m.floor((WidthClone3L - BL) / F_WidthMod / 2.0) * WidthMod)
            Clone3L_CR = max(0, WidthClone3L - BL - Clone3L_CL)
            Clone3L_CT = max(0, m.floor((HeightClone3L - CRVHeight) / F_HeightMod / 2.0) * HeightMod)
            Clone3L_CB = max(0, HeightClone3L - CRVHeight - Clone3L_CT)
            Clone3R_CL = max(0, m.floor((WidthClone3R - BR) / F_WidthMod / 2.0) * WidthMod)
            Clone3R_CR = max(0, WidthClone3R - BR - Clone3R_CL)
            Clone3R_CT = max(0, m.floor((HeightClone3R - CRVHeight) / F_HeightMod / 2.0) * HeightMod)
            Clone3R_CB = max(0, HeightClone3R - CRVHeight - Clone3R_CT)

            Clone3L = Clone3L.std.Crop(Clone3L_CL, Clone3L_CR, Clone3L_CT, Clone3L_CB)
            Clone3R = Clone3R.std.Crop(Clone3R_CL, Clone3R_CR, Clone3R_CT, Clone3R_CB)

            if (IsFeather is True):

                FClone3L = Clone3L
                FClone3R = Clone3R

            if (IsFeather is False):

                Clone3L = eval("Tweak(Clone3L, bright=Bright, cont=Cont, sat=Sat)" + \
                  BlurStr + GrainStr) \
                  if (IsTweak is True) else \
                  eval("Clone3L" + BlurStr + GrainStr)
                Clone3R = eval("Tweak(Clone3R, bright=Bright, cont=Cont, sat=Sat)" + \
                  BlurStr + GrainStr) \
                  if (IsTweak is True) else \
                  eval("Clone3R" + BlurStr + GrainStr)

# -------------------------------------------------------------------------------
#            PillarBox FrostyGradient Borders
# -------------------------------------------------------------------------------

        if (FrostyGradient is True):

            GradientBlurStr = \
              ".std.BoxBlur(hradius=40, vradius=0, hpasses=3, vpasses=0)" \
              if (ChromaSS == "444") or (SCFamily == vs.GRAY) else \
              ".std.BoxBlur(hradius=40, vradius=0, hpasses=3, vpasses=0, planes=[0])" + \
              ".std.BoxBlur(hradius=20, vradius=0, hpasses=3, vpasses=0, planes=[1,2])"

            GradientLR = core.std.BlankClip(CRVideo, width=80, height=CRVHeight)
            GradientLR = core.std.AddBorders(GradientLR, 16, 0, 0, 0, color=BColor)
            GradientLR = Tweak(GradientLR, bright=Bright, cont=Cont, sat=Sat) \
              if (IsTweak is True) else GradientLR
            GradientLR = eval("GradientLR" + GradientBlurStr)
            GradientL = eval("core.std.FlipHorizontal(GradientLR)" + BicubicLStr + GrainStr + LoopStr)
            GradientR = eval("GradientLR" + BicubicLStr + GrainStr + LoopStr)

# -------------------------------------------------------------------------------
#            PillarBox Plain Borders
# -------------------------------------------------------------------------------

        if (IsPlain is True) and (BL >= WidthMod):

            PlainL = eval("Tweak(core.std.BlankClip(CRVideo, width=BL, color=BColor), " + \
              "bright=Bright, cont=Cont, sat=Sat)" + GrainStr + LoopStr) \
              if (IsTweak is True) else \
              eval("core.std.BlankClip(CRVideo, width=BL, color=BColor)" + GrainStr + LoopStr)

        if (IsPlain is True) and (BR >= WidthMod):

            PlainR = eval("Tweak(core.std.BlankClip(CRVideo, width=BR, color=BColor), " + \
              "bright=Bright, cont=Cont, sat=Sat)" + GrainStr + LoopStr) \
              if (IsTweak is True) else \
              eval("core.std.BlankClip(CRVideo, width=BR, color=BColor)" + GrainStr + LoopStr)

        if (IsPlain is True) and (IsFeather is True):

            PlainL2 = eval("Tweak(core.std.BlankClip(CRVideo, width=BL, color=BColor), " + \
              "bright=Bright, cont=Cont, sat=Sat)" + LoopStr) \
              if (IsTweak is True) else \
              eval("core.std.BlankClip(CRVideo, width=BL, color=BColor)" + LoopStr)
            PlainR2 = eval("Tweak(core.std.BlankClip(CRVideo, width=BR, color=BColor), " + \
              "bright=Bright, cont=Cont, sat=Sat)" + LoopStr) \
              if (IsTweak is True) else \
              eval("core.std.BlankClip(CRVideo, width=BR, color=BColor)" + LoopStr)

# -------------------------------------------------------------------------------
#            PillarBox Frosty Video
# -------------------------------------------------------------------------------

        if (IsFeather is False):

            FrostyBL = FrostyL if (IsFrosty is True) else \
              CloneL if (Clone1 is True) else \
              Clone3L if (Clone3 is True) else \
              GradientL if (FrostyGradient is True) else \
              PlainL if (IsPlain is True) and (BL >= WidthMod) else None
            FrostyBR = FrostyR if (IsFrosty is True) else \
              CloneR if (Clone1 is True) else \
              Clone3R if (Clone3 is True) else \
              GradientR if (FrostyGradient is True) else \
              PlainR if (IsPlain is True) and (BR >= WidthMod) else None

        else:

            FrostyBL = FlipCloneL if (IsFrosty is True) or \
              (Clone1 is True) or (Clone2 is True) or (Clone3 is True) else \
              PlainL2 if (IsPlain is True) and (BL >= WidthMod) else None
            FrostyBR = FlipCloneR if (IsFrosty is True) or \
              (Clone1 is True) or (Clone2 is True) or (Clone3 is True) else \
              PlainR2 if (IsPlain is True) and (BR >= WidthMod) else None

        FrostyOut = eval("core.std.StackHorizontal([FrostyBL" + CRVideoStr + "FrostyBR])") \
          if (IsFrosty is True) or (IsClone is True) or (FrostyGradient is True) or \
          ((IsPlain is True) and (BL >= WidthMod <= BR)) else \
          core.std.StackHorizontal([FrostyBL, CRVideo]) \
          if (IsPlain is True) and (BL >= WidthMod) else \
          core.std.StackHorizontal([CRVideo, FrostyBR]) \
          if (IsPlain is True) and (BR >= WidthMod) else CRVideo

# -------------------------------------------------------------------------------
#            PillarBox Feather Mask
# -------------------------------------------------------------------------------

        if (IsFeather is True):

            FeatherMask = \
              core.std.BlankClip(CRVideo, width=CRVWidth-(Feather*2))
            FeatherMask = \
              FeatherMask.std.AddBorders(BL+Feather, BR+Feather, 0, 0, color=RGBColor(CRVideo, 'FFFFFF'))
            FeatherMask = \
              FeatherMask.std.Levels(min_in=MinIn, max_in=MaxInL, min_out=0, max_out=MaxOut, planes=0)
            FeatherMask = \
              FeatherMask.std.Levels(min_in=MinIn, max_in=MaxInC, min_out=0, max_out=MaxOut, planes=[1,2])
            FeatherMask = \
              FeatherMask.std.BoxBlur(hradius=Feather/2.25, vradius=0, hpasses=3)
            FeatherMask = \
              eval("FeatherMask" + LoopStr)

# -------------------------------------------------------------------------------
#            PillarBox Feather Clip
# -------------------------------------------------------------------------------

        if (IsFeather is True):

            FeatherClip = core.std.StackHorizontal([\
              FFrostyL.std.AddBorders(0, CRVWidth-FL+BL-FR+BR, 0, 0), FFrostyR]) \
              if (IsFrosty is True) else \
              eval("core.std.StackHorizontal([" + \
              "core.std.StackHorizontal([FCloneL, CRVideo.std.Crop(0, CRVWidth-FL+BL, 0, 0)])" + \
              BlurStr + GrainStr + FBordStr + \
              "core.std.StackHorizontal([CRVideo.std.Crop(CRVWidth-FR+BR, 0, 0, 0), FCloneR])" + \
              BlurStr + GrainStr + "])") \
              if (Clone1 is True) else \
              core.std.StackHorizontal([\
              Clone2L.std.AddBorders(0, CRVWidth-FL+BL-FR+BR, 0, 0), Clone2R]) \
              if (Clone2 is True) else \
              eval("core.std.StackHorizontal([" + \
              "core.std.StackHorizontal([FClone3L, CRVideo.std.Crop(0, CRVWidth-FL+BL, 0, 0)])" + \
              BlurStr + GrainStr + FBordStr + \
              "core.std.StackHorizontal([CRVideo.std.Crop(CRVWidth-FR+BR, 0, 0, 0), FClone3R])" + \
              BlurStr + GrainStr + "])") \
              if (Clone3 is True) else \
              eval("core.std.StackHorizontal([" + \
              "Tweak(core.std.BlankClip(CRVideo, width=FL, color=BColor), " + \
              "bright=Bright, cont=Cont, sat=Sat)" + GrainStr + LoopStr + FBordStr + \
              "Tweak(core.std.BlankClip(CRVideo, width=FR, color=BColor), " + \
              "bright=Bright, cont=Cont, sat=Sat)" + GrainStr + LoopStr + "])") \
              if (IsTweak is True) else \
              eval("core.std.StackHorizontal([" + \
              "core.std.BlankClip(CRVideo, width=FL, color=BColor)" + \
              GrainStr + LoopStr + FBordStr + \
              "core.std.BlankClip(CRVideo, width=FR, color=BColor)" + GrainStr + LoopStr + "])")

# -------------------------------------------------------------------------------
#            PillarBox Output
# -------------------------------------------------------------------------------

        FrostyOut = FrostyOut if (IsFeather is False) else \
          core.std.MaskedMerge(FrostyOut, FeatherClip, mask=FeatherMask, first_plane=True)

    else:

# ===============================================================================
#            LetterBox Borders
# ===============================================================================

        BT = BT - BDivider if (IsBDivider is True) else BT
        BB = BB - BDivider if (IsBDivider is True) else BB
        FT = 0 if (IsFeather is False) else \
          max(BT, m.floor((BT + (Feather * 3)) / F_WidthMod) * WidthMod)
        FB = 0 if (IsFeather is False) else \
          max(BB, m.floor((BB + (Feather * 3)) / F_WidthMod) * WidthMod)

# -------------------------------------------------------------------------------
#            LetterBox Strings
# -------------------------------------------------------------------------------

        CRVideoStr = ",CRVideo," if (IsBDivider is False) else \
          ",CRVideo.std.AddBorders(0, 0, BDivider, BDivider, color=BDColor),"

        BlurStr1 = \
          ".std.BoxBlur(hradius=Blur, vradius=VBlur, hpasses=Iterations, vpasses=Iterations)"
        BlurStr2 = \
          ".std.BoxBlur(hradius=Blur, vradius=VBlur, " + \
          "hpasses=Iterations, vpasses=Iterations, planes=[0])"
        BlurStr3 = \
          ".std.BoxBlur(hradius=ChromaBlur, vradius=VChromaBlur, " + \
          "hpasses=Iterations, vpasses=Iterations, planes=[1,2])"

        FBordStr = ".std.AddBorders(0, 0, 0, CRVHeight-FT+BT-FB+BB),"
        BicubicStr = ".resize.Bicubic"

        BicubicTStr = "" if (IsPlain is True) else BicubicStr + "(CRVWidth, BT)"
        BicubicBStr = "" if (IsPlain is True) else BicubicStr + "(CRVWidth, BB)"
        FBicubicTStr = "" if (IsFeather is False) else BicubicStr + "(CRVWidth, FT)"
        FBicubicBStr = "" if (IsFeather is False) else BicubicStr + "(CRVWidth, FB)"

        Bicubic3TStr = BicubicStr + "(max(CRVWidth, m.floor(CRVWidth * RatioT / F_WidthMod) * WidthMod), \
          max(BT, m.floor(HeightCloneBorderT * RatioT / F_HeightMod) * HeightMod))" \
          if (Clone3 is True) else ""
        Bicubic3BStr = BicubicStr + "(max(CRVWidth, m.floor(CRVWidth * RatioB / F_WidthMod) * WidthMod), \
          max(BB, m.floor(HeightCloneBorderB * RatioB / F_HeightMod) * HeightMod))" \
          if (Clone3 is True) else ""

        GrainStr = "" if (Texture == 0) and (TextureC == 0) else \
          ".f3kdb.Deband(grainy=Texture, grainc=TextureC, dither_algo=1, " + \
          "dynamic_grain=DynamicT, output_depth=Source_Bits)"

        LoopStr = ".std.Trim(1, 1).std.Loop(TotalFrames)"
        SLoopStr = ".std.Trim(FrostyStatic, last=FrostyStatic).std.Loop(TotalFrames)" \
          if (IsFrostyStatic is True) else ""
        TSoftStr = ".focus2.TemporalSoften2(TSoft, 255, 255, 20, 2)" if (IsTSoft is True) else ""

# -------------------------------------------------------------------------------
#            LetterBox Frosty Borders
# -------------------------------------------------------------------------------

        FrostyCrop = CRVHeight - max(BorderMin, \
          round(F_CRVHeight / (5.0 if (Faster is False) else 10.0) / F_HeightMod) * HeightMod) \
          if (IsFrosty is True) and (IsTSoft is True) else ""

        FrostyTB = eval("core.std.StackVertical([CRVideo.std.Crop(0, 0, 0, FrostyCrop), \
          CRVideo.std.Crop(0, 0, FrostyCrop, 0)])" + TSoftStr) \
          if (IsFrosty is True) and (IsTSoft is True) else ""

        HeightFrostyTB = FrostyTB.height if (IsFrosty is True) and (IsTSoft is True) else 0

        FrostyCrop = HeightFrostyTB - max(BorderMin, \
          round(float(HeightFrostyTB) / (16.0 if (Faster is False) else 8.0) / F_HeightMod) * HeightMod) \
          if (IsFrosty is True) and (IsTSoft is True) else \
          CRVHeight - max(BorderMin, round(F_CRVHeight / 80.0 / F_HeightMod) * HeightMod) \
          if (IsFrosty is True) else 0

        if (IsFrosty is True) and (IsTSoft is True):

            FrostyT = FrostyTB.std.Crop(0, 0, 0, FrostyCrop)
            FrostyB = FrostyTB.std.Crop(0, 0, FrostyCrop, 0)

        elif (IsFrosty is True):

            FrostyT = CRVideo.std.Crop(0, 0, 0, FrostyCrop)
            FrostyB = CRVideo.std.Crop(0, 0, FrostyCrop, 0)

        HeightFrostyT = FrostyT.height if (IsFrosty is True) else 0
        HeightFrostyB = FrostyB.height if (IsFrosty is True) else 0

        if (IsFrosty is True):

            FrostyBlendT = FrostyT if (Blend == 0) else \
              FrostyB if (Blend == 1) else \
              core.std.Merge(FrostyT, core.std.FlipVertical(FrostyB), weight=Blend)
            FrostyBlendB = FrostyB if (Blend == 0) else \
              FrostyT if (Blend == 1) else \
              core.std.Merge(FrostyB, core.std.FlipVertical(FrostyT), weight=Blend)

        Blur = round(Blur if (IsBlur is True) else \
          F_CRVWidth / 10.0 if (IsFrosty is True) else Blur)
        VBlur = round(VBlur if (IsVBlur is True) else \
          max(HeightFrostyT, HeightFrostyB) + (0 if (IsFeather is False) else  Feather) \
          if (IsFrosty is True) else VBlur)

        ChromaBlur = Blur if (CRVideo.format.subsampling_w == 0) else \
          round(Blur / 2.0  if (CRVideo.format.subsampling_w == 1) else Blur / 4.0)
        VChromaBlur = VBlur if (CRVideo.format.subsampling_h == 0) else \
          round(VBlur / 2.0 if (CRVideo.format.subsampling_h == 1) else VBlur / 4.0)

        BlurStr1 = "" if (Blur == 0 == VBlur) or (Iterations == 0) or \
          (FrostyGradient is True) or ((IsPlain is True) and (IsFeather is False)) else BlurStr1
        BlurStr2 = "" if (Blur == 0 == VBlur) or (Iterations == 0) or \
          (FrostyGradient is True) or ((IsPlain is True) and (IsFeather is False)) else BlurStr2
        BlurStr3 = "" if (ChromaBlur == 0 == VChromaBlur) or (Iterations == 0) or \
          (FrostyGradient is True) or ((IsPlain is True) and (IsFeather is False)) else BlurStr3

        BlurStr = BlurStr1 if (ChromaSS == "444") or (SCFamily == vs.GRAY) else \
          BlurStr2 + BlurStr3

        if (IsFrosty is True) and (IsTweak is True):

            FrostyT = eval("Tweak(FrostyBlendT, bright=Bright, cont=Cont, sat=Sat)" + \
              BlurStr + BicubicTStr + GrainStr + SLoopStr)
            FrostyB = eval("Tweak(FrostyBlendB, bright=Bright, cont=Cont, sat=Sat)" + \
              BlurStr + BicubicBStr + GrainStr + SLoopStr)

        elif (IsFrosty is True):

            FrostyT = eval("FrostyBlendT" + BlurStr + BicubicTStr + GrainStr + SLoopStr)
            FrostyB = eval("FrostyBlendB" + BlurStr + BicubicBStr + GrainStr + SLoopStr)

        if (IsFrosty is True) and (IsFeather is True) and (IsTweak is True):

            FFrostyT = eval("Tweak(FrostyBlendT, bright=Bright, cont=Cont, sat=Sat)" + \
              BlurStr + FBicubicTStr + GrainStr + SLoopStr)
            FFrostyB = eval("Tweak(FrostyBlendB, bright=Bright, cont=Cont, sat=Sat)" + \
              BlurStr + FBicubicBStr + GrainStr + SLoopStr)

        elif (IsFrosty is True) and (IsFeather is True):

            FFrostyT = eval("FrostyBlendT" + BlurStr + FBicubicTStr + GrainStr + SLoopStr)
            FFrostyB = eval("FrostyBlendB" + BlurStr + FBicubicBStr + GrainStr + SLoopStr)

# -------------------------------------------------------------------------------
#            LetterBox Clone Borders
# -------------------------------------------------------------------------------

        if (IsClone is True) or ((IsFrosty is True) and (IsFeather is True)):

            MaxCloneC = m.floor(F_CRVHeight / 2.0 / F_HeightMod) * HeightMod
            CloneCropT = max(MaxCloneC, CRVHeight - BT)
            CloneCropB = max(MaxCloneC, CRVHeight - BB)
            CloneBorderT = CRVideo.std.Crop(0, 0, 0, CloneCropT)
            CloneBorderB = CRVideo.std.Crop(0, 0, CloneCropB, 0)
            HeightCloneBorderT = CloneBorderT.height
            HeightCloneBorderB = CloneBorderB.height

            FlipCloneT = core.std.FlipVertical(CloneBorderT if (HeightCloneBorderT == BT) else \
              eval("CloneBorderT" + BicubicTStr))
            FlipCloneB = core.std.FlipVertical(CloneBorderB if (HeightCloneBorderB == BB) else \
              eval("CloneBorderB" + BicubicBStr))

        if (Clone1 is True) or (Clone3 is True):

            if (IsTSoft is True):

                CloneTB = eval("core.std.StackVertical([CloneBorderT, CloneBorderB])" + TSoftStr)
                HeightCloneTB = CloneTB.height

            CloneT = CloneTB.std.Crop(0, 0, 0, HeightCloneTB-HeightCloneBorderT) \
              if (IsTSoft is True) else CloneBorderT
            CloneB = CloneTB.std.Crop(0, 0, HeightCloneTB-HeightCloneBorderB, 0) \
              if (IsTSoft is True) else CloneBorderB
            CloneBlendT = CloneT if (Blend == 0) else CloneB if (Blend == 1) else \
              core.std.Merge(CloneT, FlipVertical(CloneB), weight=Blend)
            CloneBlendB = CloneB if (Blend == 0) else CloneT if (Blend == 1) else \
              core.std.Merge(CloneB, FlipVertical(CloneT), weight=Blend)

        if (Clone1 is True):

            if (IsFeather is True):

                FCloneT = CloneBlendT if (HeightCloneBorderT == BT) else \
                  eval("CloneBlendT" + BicubicTStr)
                FCloneB = CloneBlendB if (HeightCloneBorderB == BB) else \
                  eval("CloneBlendB" + BicubicBStr)

            if (IsFeather is False) and (IsTweak is True):

                CloneT = eval("Tweak(CloneBlendT, bright=Bright, cont=Cont, sat=Sat)" + \
                  BlurStr + GrainStr) \
                  if (HeightCloneBorderT == BT) else \
                  eval("Tweak(CloneBlendT, bright=Bright, cont=Cont, sat=Sat)" + \
                  BlurStr + BicubicTStr + GrainStr)
                CloneB = eval("Tweak(CloneBlendB, bright=Bright, cont=Cont, sat=Sat)" + \
                  BlurStr + GrainStr) \
                  if (HeightCloneBorderB == BB) else \
                  eval("Tweak(CloneBlendB, bright=Bright, cont=Cont, sat=Sat)" + \
                  BlurStr + BicubicBStr + GrainStr)

            elif (IsFeather is False):

                CloneT = eval("CloneBlendT" + BlurStr) \
                  if (HeightCloneBorderT == BT) else \
                  eval("CloneBlendT" + BicubicTStr + GrainStr)
                CloneB = eval("CloneBlendB" + BlurStr) \
                  if (HeightCloneBorderB == BB) else \
                  eval("CloneBlendB" + BicubicBStr + GrainStr)

# -------------------------------------------------------------------------------
#            LetterBox Clone2 Borders
# -------------------------------------------------------------------------------

        if (Clone2 is True):

            Clone2CropT = max(MaxCloneC, CRVHeight - FT)
            Clone2CropB = max(MaxCloneC, CRVHeight - FB)
            Clone2BT = CRVideo.std.Crop(0, 0, 0, Clone2CropT)
            Clone2BB = CRVideo.std.Crop(0, 0, Clone2CropB, 0)
            HeightClone2BT = Clone2BT.height
            HeightClone2BB = Clone2BB.height

            FlipClone2T = core.std.FlipVertical(Clone2BT if (HeightClone2BT == FT) else \
              eval("Clone2BT" + FBicubicTStr))
            FlipClone2B = core.std.FlipVertical(Clone2BB if (HeightClone2BB == FB) else \
              eval("Clone2BB" + FBicubicBStr))

            if (IsTSoft is True):

                Clone2TB = eval("core.std.StackVertical([Clone2BT, Clone2BB])" + TSoftStr)
                HeightClone2TB = Clone2TB.height if (IsTSoft is True) else 0

            Clone2T = Clone2TB.std.Crop(0, 0, 0, HeightClone2TB-HeightClone2BT) \
              if (IsTSoft is True) else Clone2BT
            Clone2B = Clone2TB.std.Crop(0, 0, HeightClone2TB-HeightClone2BB, 0) \
              if (IsTSoft is True) else Clone2BB
            Clone2BlendT = Clone2T if (Blend == 0) else Clone2B if (Blend == 1) else \
              core.std.Merge(Clone2T, core.std.FlipVertical(Clone2B), weight=Blend)
            Clone2BlendB = Clone2B if (Blend == 0) else Clone2T if (Blend == 1) else \
              core.std.Merge(Clone2B, core.std.FlipVertical(Clone2T), weight=Blend)

            if (IsTweak is True):

                Clone2T = eval("Tweak(Clone2BlendT, bright=Bright, cont=Cont, sat=Sat)" + \
                  BlurStr + GrainStr) \
                  if (HeightClone2BT == FT) else \
                  eval("Tweak(Clone2BlendT, bright=Bright, cont=Cont, sat=Sat)" + \
                  BlurStr + FBicubicTStr + GrainStr)
                Clone2B = eval("Tweak(Clone2BlendB, bright=Bright, cont=Cont, sat=Sat)" + \
                  BlurStr + GrainStr) \
                  if (HeightClone2BB == FB) else \
                  eval("Tweak(Clone2BlendB, bright=Bright, cont=Cont, sat=Sat)" + \
                  BlurStr + FBicubicBStr + GrainStr)

            else:

                Clone2T = eval("Clone2BlendT" + BlurStr + GrainStr) \
                  if (HeightClone2BT == FT) else \
                  eval("Clone2BlendT" + BlurStr + FBicubicTStr + GrainStr)
                Clone2B = eval("Clone2BlendB" + BlurStr + GrainStr) \
                  if (HeightClone2BB == FB) else \
                  eval("Clone2BlendB" + BlurStr + FBicubicBStr + GrainStr)

# -------------------------------------------------------------------------------
#            LetterBox Clone3 Borders
# -------------------------------------------------------------------------------

        if (Clone3 is True):

            RatioT = 1.618 if ((HeightCloneBorderT - BT) >= 0) else \
              float(BT) / float(HeightCloneBorderT)
            RatioB = 1.618 if ((HeightCloneBorderB - BB) >= 0) else \
              float(BB) / float(HeightCloneBorderB)

            Clone3T = eval("CloneBlendT" + Bicubic3TStr)
            Clone3B = eval("CloneBlendB" + Bicubic3BStr)
            WidthClone3T = Clone3T.width
            HeightClone3T = Clone3T.height
            WidthClone3B = Clone3B.width
            HeightClone3B = Clone3B.height

            Clone3T_CL = max(0, m.floor((WidthClone3T - CRVWidth) / F_WidthMod / 2.0) * WidthMod)
            Clone3T_CR = max(0, WidthClone3T - CRVWidth - Clone3T_CL)
            Clone3T_CT = max(0, m.floor((HeightClone3T - BT) / F_HeightMod / 2.0) * HeightMod)
            Clone3T_CB = max(0, HeightClone3T - BT - Clone3T_CT)
            Clone3B_CL = max(0, m.floor((WidthClone3B - CRVWidth) / F_WidthMod / 2.0) * WidthMod)
            Clone3B_CR = max(0, WidthClone3B - CRVWidth - Clone3B_CL)
            Clone3B_CT = max(0, m.floor((HeightClone3B - BB) / F_HeightMod / 2.0) * HeightMod)
            Clone3B_CB = max(0, HeightClone3B - BB - Clone3B_CT)

            Clone3T = Clone3T.std.Crop(Clone3T_CL, Clone3T_CR, Clone3T_CT, Clone3T_CB)
            Clone3B = Clone3B.std.Crop(Clone3B_CL, Clone3B_CR, Clone3B_CT, Clone3B_CB)

            if (IsFeather is True):

                FClone3T = Clone3T
                FClone3B = Clone3B

            if (IsFeather is False):

                Clone3T = eval("Tweak(Clone3T, bright=Bright, cont=Cont, sat=Sat)" + \
                  BlurStr + GrainStr) \
                  if (IsTweak is True) else \
                  eval("Clone3T" + BlurStr + GrainStr)
                Clone3B = eval("Tweak(Clone3B, bright=Bright, cont=Cont, sat=Sat)" + \
                  BlurStr + GrainStr) \
                  if (IsTweak is True) else \
                  eval("Clone3B" + BlurStr + GrainStr)

# -------------------------------------------------------------------------------
#            LetterBox FrostyGradient Borders
# -------------------------------------------------------------------------------

        if (FrostyGradient is True):

            GradientBlurStr = \
              ".std.BoxBlur(hradius=0, vradius=40, hpasses=0, vpasses=3)" \
              if (ChromaSS == "444") or (SCFamily == vs.GRAY) else \
              ".std.BoxBlur(hradius=0, vradius=40, hpasses=0, vpasses=3, planes=[0])" + \
              ".std.BoxBlur(hradius=0, vradius=20, hpasses=0, vpasses=3, planes=[1,2])"

            GradientTB = core.std.BlankClip(CRVideo, width=CRVWidth, height=80)
            GradientTB = core.std.AddBorders(GradientTB, 0, 0, 16, 0, color=BColor)
            GradientTB = Tweak(GradientTB, bright=Bright, cont=Cont, sat=Sat) \
              if (IsTweak is True) else GradientTB
            GradientTB = eval("GradientTB" + GradientBlurStr)
            GradientT = eval("core.std.FlipVertical(GradientTB)" + BicubicTStr + GrainStr + LoopStr)
            GradientB = eval("GradientTB" + BicubicBStr + GrainStr + LoopStr)

# -------------------------------------------------------------------------------
#            LetterBox Plain Borders
# -------------------------------------------------------------------------------

        if (IsPlain is True) and (BT >= HeightMod):

            PlainT = eval("Tweak(core.std.BlankClip(CRVideo, height=BT, color=BColor), " + \
              "bright=Bright, cont=Cont, sat=Sat)" + GrainStr + LoopStr) \
              if (IsTweak is True) else \
              eval("core.std.BlankClip(CRVideo, height=BT, color=BColor)" + GrainStr + LoopStr)

        if (IsPlain is True) and (BB >= HeightMod):

            PlainB = eval("Tweak(core.std.BlankClip(CRVideo, height=BB, color=BColor), " + \
              "bright=Bright, cont=Cont, sat=Sat)" + GrainStr + LoopStr) \
              if (IsTweak is True) else \
              eval("core.std.BlankClip(CRVideo, height=BB, color=BColor)" + GrainStr + LoopStr)

        if (IsPlain is True) and (IsFeather is True):

            PlainT2 = eval("Tweak(core.std.BlankClip(CRVideo, height=BT, color=BColor), " + \
              "bright=Bright, cont=Cont, sat=Sat)" + LoopStr) \
              if (IsTweak is True) else \
              eval("core.std.BlankClip(CRVideo, height=BT, color=BColor)" + LoopStr)
            PlainB2 = eval("Tweak(core.std.BlankClip(CRVideo, height=BB, color=BColor), " + \
              "bright=Bright, cont=Cont, sat=Sat)" + LoopStr) \
              if (IsTweak is True) else \
              eval("core.std.BlankClip(CRVideo, height=BB, color=BColor)" + LoopStr)

# -------------------------------------------------------------------------------
#            LetterBox Frosty Video
# -------------------------------------------------------------------------------

        if (IsFeather is False):

            FrostyBT = FrostyT if (IsFrosty is True) else \
              CloneT if (Clone1 is True) else \
              Clone3T if (Clone3 is True) else \
              GradientT if (FrostyGradient is True) else \
              PlainT if (IsPlain is True) and (BT >= HeightMod) else None
            FrostyBB = FrostyB if (IsFrosty is True) else \
              CloneB if (Clone1 is True) else \
              Clone3B if (Clone3 is True) else \
              GradientB if (FrostyGradient is True) else \
              PlainB if (IsPlain is True) and (BB >= HeightMod) else None

        else:

            FrostyBT = FlipCloneT if (IsFrosty is True) or \
              (Clone1 is True) or (Clone2 is True) or (Clone3 is True) else \
              PlainT2 if (IsPlain is True) and (BT >= HeightMod) else None
            FrostyBB = FlipCloneB if (IsFrosty is True) or \
              (Clone1 is True) or (Clone2 is True) or (Clone3 is True) else \
              PlainRB if (IsPlain is True) and (BB >= HeightMod) else None

        FrostyOut = eval("core.std.StackVertical([FrostyBT" + CRVideoStr + "FrostyBB])") \
          if (IsFrosty is True) or (IsClone is True) or (FrostyGradient is True) or \
          ((IsPlain is True) and (BT >= HeightMod <= BB)) else \
          core.std.StackVertical([FrostyBT, CRVideo]) \
          if (IsPlain is True) and (BT >= HeightMod) else \
          core.std.StackVertical([CRVideo, FrostyBB]) \
          if (IsPlain is True) and (BB >= HeightMod) else CRVideo

# -------------------------------------------------------------------------------
#            LetterBox Feather Mask
# -------------------------------------------------------------------------------

        if (IsFeather is True):

            FeatherMask = \
              core.std.BlankClip(CRVideo, height=CRVHeight-(Feather*2))
            FeatherMask = \
              FeatherMask.std.AddBorders(0, 0, BT+Feather, BB+Feather, color=RGBColor(CRVideo, 'FFFFFF'))
            FeatherMask = \
              FeatherMask.std.Levels(min_in=MinIn, max_in=MaxInL, min_out=0, max_out=MaxOut, planes=0)
            FeatherMask = \
              FeatherMask.std.Levels(min_in=MinIn, max_in=MaxInC, min_out=0, max_out=MaxOut, planes=[1,2])
            FeatherMask = \
              FeatherMask.std.BoxBlur(hradius=0, vradius=Feather/2.25, vpasses=3)
            FeatherMask = \
              eval("FeatherMask" + LoopStr)

# -------------------------------------------------------------------------------
#            LetterBox Feather Clip
# -------------------------------------------------------------------------------

        if (IsFeather is True):

            FeatherClip = core.std.StackVertical([\
              FFrostyT.std.AddBorders(0, 0, 0, CRVHeight-FT+BT-FB+BB), FFrostyB]) \
              if (IsFrosty is True) else \
              eval("core.std.StackVertical([" + \
              "core.std.StackVertical([FCloneT, CRVideo.std.Crop(0, 0, 0, CRVHeight-FT+BT)])" + \
              BlurStr + GrainStr + FBordStr + \
              "core.std.StackVertical([CRVideo.std.Crop(0, 0, CRVHeight-FB+BB, 0), FCloneB])" + \
              BlurStr + GrainStr + "])") \
              if (Clone1 is True) else \
              core.std.StackVertical([\
              Clone2T.std.AddBorders(0, 0, 0, CRVHeight-FT+BT-FB+BB), Clone2B]) \
              if (Clone2 is True) else \
              eval("core.std.StackVertical([" + \
              "core.std.StackVertical([FClone3T, CRVideo.std.Crop(0, 0, 0, CRVHeight-FT+BT)])" + \
              BlurStr + GrainStr + FBordStr + \
              "core.std.StackVertical([CRVideo.std.Crop(0, 0, CRVHeight-FB+BB, 0), FClone3B])" + \
              BlurStr + GrainStr + "])") \
              if (Clone3 is True) else \
              eval("core.std.StackVertical([" + \
              "Tweak(core.std.BlankClip(CRVideo, height=FT, color=BColor), " + \
              "bright=Bright, cont=Cont, sat=Sat)" + GrainStr + LoopStr + FBordStr + \
              "Tweak(core.std.BlankClip(CRVideo, height=FB, color=BColor), " + \
              "bright=Bright, cont=Cont, sat=Sat)" + GrainStr + LoopStr + "])") \
              if (IsTweak is True) else \
              eval("core.std.StackVertical([" + \
              "core.std.BlankClip(CRVideo, height=FT, color=BColor)" + \
              GrainStr + LoopStr + FBordStr + \
              "core.std.BlankClip(CRVideo, height=FB, color=BColor)" + GrainStr + LoopStr + "])")

# -------------------------------------------------------------------------------
#            LetterBox Output
# -------------------------------------------------------------------------------

        FrostyOut = FrostyOut if (IsFeather is False) else \
          core.std.MaskedMerge(FrostyOut, FeatherClip, mask=FeatherMask, first_plane=True)

# ===============================================================================

    return FrostyOut

# ===============================================================================
# ===============================================================================
#            Helper Functions
# ===============================================================================
# ===============================================================================

# -------------------------------------------------------------------------------
# -------------------------------------------------------------------------------
#            Tweak (copied from adjust.py)
# -------------------------------------------------------------------------------
# -------------------------------------------------------------------------------

def Tweak(clip, hue=None, sat=None, bright=None, cont=None, coring=True):

    if (clip.format.color_family == vs.RGB):
        raise vs.Error('Tweak: RGB is not supported')
    if (clip.format is None):
        raise vs.Error("Tweak: only clips with a constant format are supported")

    if ((hue is not None) or (sat is not None)) and (clip.format.color_family != vs.GRAY):

        hue = CR_Default(hue, 0.0) * m.pi / 180.0
        hue_sin = m.sin(hue)
        hue_cos = m.cos(hue)
        sat = CR_Default(sat, 1.0)
        gray = 128 << (clip.format.bits_per_sample - 8)
        chroma_min = 0
        chroma_max = (2 ** clip.format.bits_per_sample) - 1

        if (coring is True):

            chroma_min = 16 << (clip.format.bits_per_sample - 8)
            chroma_max = 240 << (clip.format.bits_per_sample - 8)

        expr_u = \
          "x {} - {} * y {} - {} * + {} + {} max {} min".format(gray, \
          hue_cos * sat, gray, hue_sin * sat, gray, chroma_min, chroma_max) \
          if (clip.format.sample_type == vs.INTEGER) else \
          "x {} * y {} * + -0.5 max 0.5 min".format(hue_cos * sat, hue_sin * sat)
        expr_v = \
          "y {} - {} * x {} - {} * - {} + {} max {} min".format(gray, \
          hue_cos * sat, gray, hue_sin * sat, gray, chroma_min, chroma_max) \
          if (clip.format.sample_type == vs.INTEGER) else \
          "y {} * x {} * - -0.5 max 0.5 min".format(hue_cos * sat, hue_sin * sat)

        src_u = clip.std.ShufflePlanes(planes=1, colorfamily=vs.GRAY)
        src_v = clip.std.ShufflePlanes(planes=2, colorfamily=vs.GRAY)
        dst_u = core.std.Expr(clips=[src_u, src_v], expr=expr_u)
        dst_v = core.std.Expr(clips=[src_u, src_v], expr=expr_v)

        clip = core.std.ShufflePlanes(clips=[clip, dst_u, dst_v], planes=[0, 0, 0], \
          colorfamily=clip.format.color_family)

    if (bright is not None) or (cont is not None):

        bright = CR_Default(bright, 0.0)
        cont = CR_Default(cont, 1.0)

        if (clip.format.sample_type == vs.INTEGER):

            luma_lut = []
            luma_min = 0
            luma_max = (2 ** clip.format.bits_per_sample) - 1

            if (coring is True):

                luma_min = 16 << (clip.format.bits_per_sample - 8)
                luma_max = 235 << (clip.format.bits_per_sample - 8)

            for i in range(2 ** clip.format.bits_per_sample):

                val = int((i - luma_min) * cont + bright + luma_min + 0.5)
                luma_lut.append(min(max(val, luma_min), luma_max))

            clip = clip.std.Lut(planes=0, lut=luma_lut)

        else:

            expression = "x {} * {} + 0.0 max 1.0 min".format(cont, bright)
            clip = clip.std.Expr(expr=[expression, "", ""])

    return clip

# -------------------------------------------------------------------------------
# -------------------------------------------------------------------------------
#            RGBColor
# -------------------------------------------------------------------------------
# -------------------------------------------------------------------------------

def RGBColor(clip, color=None, matrix=None, range=None):

    if not isinstance(clip, vs.VideoNode):
        raise vs.Error('RGBColor: clip must be a video')
    if not ((color is None) or isinstance(color, str)):
        raise vs.Error('RGBColor: color must be a string, for example \'darkblue\' or \'00008B\'')
    if not ((matrix is None) or isinstance(matrix, str) or \
      (isinstance(matrix, int) and (isinstance(matrix, bool) is False))):
        raise vs.Error('RGBColor: matrix must be an integer or string')
    if not ((range is None) or (isinstance(range, str) and \
      ((range.lower().strip() == 'full') or (range.lower().strip() == 'limited') or \
      (range.lower().strip() == 'f') or (range.lower().strip() == 'l')))):
        raise vs.Error('RGBColor: range must be \'full\' or \'f\', or \'limited\' or \'l\'')
    if (clip.format.name).endswith("H"):
        raise vs.Error(f'RGBColor: {clip.format.name} is not supported')

    color = None if (color is None) else color.lower().strip()

    colors = \
      {'aliceblue'            : 'F0F8FF',    'antiquewhite'         : 'FAEBD7',
       'aqua'                 : '00FFFF',    'aquamarine'           : '7FFFD4',
       'azure'                : 'F0FFFF',    'beige'                : 'F5F5DC',
       'bisque'               : 'FFE4C4',    'black'                : '000000',
       'blanchedalmond'       : 'FFEBCD',    'blue'                 : '0000FF',
       'blueviolet'           : '8A2BE2',    'brown'                : 'A52A2A',
       'burlywood'            : 'DEB887',    'cadetblue'            : '5F9EA0',
       'chartreuse'           : '7FFF00',    'chocolate'            : 'D2691E',
       'coral'                : 'FF7F50',    'cornflowerblue'       : '6495ED',
       'cornsilk'             : 'FFF8DC',    'crimson'              : 'DC143C',
       'cyan'                 : '00FFFF',    'darkblue'             : '00008B',
       'darkcyan'             : '008B8B',    'darkgoldenrod'        : 'B8860B',
       'darkgray'             : 'A9A9A9',    'darkgrey'             : 'A9A9A9',
       'darkgreen'            : '006400',    'darkkhaki'            : 'BDB76B',
       'darkmagenta'          : '8B008B',    'darkolivegreen'       : '556B2F',
       'darkorange'           : 'FF8C00',    'darkorchid'           : '9932CC',
       'darkred'              : '8B0000',    'darksalmon'           : 'E9967A',
       'darkseagreen'         : '8FBC8F',    'darkslateblue'        : '483D8B',
       'darkslategray'        : '2F4F4F',    'darkslategrey'        : '2F4F4F',
       'darkturquoise'        : '00CED1',    'darkviolet'           : '9400D3',
       'deeppink'             : 'FF1493',    'deepskyblue'          : '00BFFF',
       'dimgray'              : '696969',    'dimgrey'              : '696969',
       'dodgerblue'           : '1E90FF',    'firebrick'            : 'B22222',
       'floralwhite'          : 'FFFAF0',    'forestgreen'          : '228B22',
       'fuchsia'              : 'FF00FF',    'gainsboro'            : 'DCDCDC',
       'ghostwhite'           : 'F8F8FF',    'gold'                 : 'FFD700',
       'goldenrod'            : 'DAA520',    'gray'                 : '808080',
       'grey'                 : '808080',    'green'                : '008000',
       'greenyellow'          : 'ADFF2F',    'honeydew'             : 'F0FFF0',
       'hotpink'              : 'FF69B4',    'indianred'            : 'CD5C5C',
       'indigo'               : '4B0082',    'ivory'                : 'FFFFF0',
       'khaki'                : 'F0E68C',    'lavender'             : 'E6E6FA',
       'lavenderblush'        : 'FFF0F5',    'lawngreen'            : '7CFC00',
       'lemonchiffon'         : 'FFFACD',    'lightblue'            : 'ADD8E6',
       'lightcoral'           : 'F08080',    'lightcyan'            : 'E0FFFF',
       'lightgoldenrodyellow' : 'FAFAD2',    'lightgreen'           : '90EE90',
       'lightgrey'            : 'D3D3D3',    'lightpink'            : 'FFB6C1',
       'lightsalmon'          : 'FFA07A',    'lightseagreen'        : '20B2AA',
       'lightskyblue'         : '87CEFA',    'lightslategray'       : '778899',
       'lightslategrey'       : '778899',    'lightsteelblue'       : 'B0C4DE',
       'lightyellow'          : 'FFFFE0',    'lime'                 : '00FF00',
       'limegreen'            : '32CD32',    'linen'                : 'FAF0E6',
       'magenta'              : 'FF00FF',    'maroon'               : '800000',
       'mediumaquamarine'     : '66CDAA',    'mediumblue'           : '0000CD',
       'mediumorchid'         : 'BA55D3',    'mediumpurple'         : '9370DB',
       'mediumseagreen'       : '3CB371',    'mediumslatenlue'      : '7B68EE',
       'mediumspringgreen'    : '00FA9A',    'mediumturquoise'      : '48D1CC',
       'mediumvioletred'      : 'C71585',    'midnightblue'         : '191970',
       'mintcream'            : 'F5FFFA',    'mistyrose'            : 'FFE4E1',
       'moccasin'             : 'FFE4B5',    'navajowhite'          : 'FFDEAD',
       'navy'                 : '000080',    'oldlace'              : 'FDF5E6',
       'olive'                : '808000',    'olivedrab'            : '6B8E23',
       'orange'               : 'FFA500',    'orangered'            : 'FF4500',
       'orchid'               : 'DA70D6',    'palegoldenrod'        : 'EEE8AA',
       'palegreen'            : '98FB98',    'paleturquoise'        : 'AFEEEE',
       'palevioletred'        : 'DB7093',    'papayawhip'           : 'FFEFD5',
       'peachpuff'            : 'FFDAB9',    'peru'                 : 'CD853F',
       'pink'                 : 'FFC0CB',    'plum'                 : 'DDA0DD',
       'powderblue'           : 'B0E0E6',    'purple'               : '800080',
       'red'                  : 'FF0000',    'rosybrown'            : 'BC8F8F',
       'royalblue'            : '4169E1',    'saddlebrown'          : '8B4513',
       'salmon'               : 'FA8072',    'sandybrown'           : 'F4A460',
       'seagreen'             : '2E8B57',    'seashell'             : 'FFF5EE',
       'sienna'               : 'A0522D',    'silver'               : 'C0C0C0',
       'skyblue'              : '87CEEB',    'slateblue'            : '6A5ACD',
       'slategray'            : '708090',    'slategrey'            : '708090',
       'snow'                 : 'FFFAFA',    'springgreen'          : '00FF7F',
       'steelblue'            : '4682B4',    'tan'                  : 'D2B48C',
       'teal'                 : '008080',    'thistle'              : 'D8BFD8',
       'tomato'               : 'FF6347',    'turquoise'            : '40E0D0',
       'violet'               : 'EE82EE',    'wheat'                : 'F5DEB3',
       'white'                : 'FFFFFF',    'whitesmoke'           : 'F5F5F5',
       'yellow'               : 'FFFF00',    'yellowgreen'          : '9ACD32',
       'gray10'               : '191919',    'grey10'               : '191919',
       'gray20'               : '323232',    'grey20'               : '323232',
       'gray30'               : '4B4B4B',    'grey30'               : '4B4B4B',
       'gray40'               : '656565',    'grey40'               : '656565',
       'gray50'               : '7F7F7F',    'grey50'               : '7F7F7F',
       'gray60'               : '989898',    'grey60'               : '989898',
       'gray70'               : 'B1B1B1',    'grey70'               : 'B1B1B1',
       'gray80'               : 'CACACA',    'grey80'               : 'CACACA',
       'gray90'               : 'E4E4E4',    'grey90'               : 'E4E4E4'}

    if (colors.get(color) is None):

        try: v0, v1, v2 = tuple(int(color[i:i+2], 16) for i in (0, 2, 4))
        except: raise vs.Error('RGBColor: Invalid color string specified')

    else:

        v0, v1, v2 = tuple(int(colors.get(color)[i:i+2], 16) for i in (0, 2, 4))

    if (clip.format.color_family != vs.RGB):

        matrix = matrix if (matrix is None) or isinstance(matrix, int) else \
          matrix.lower().strip()

        if (matrix is not None) and isinstance(matrix, str):

            M = \
              1 if (matrix == '709') else \
              4 if (matrix == 'fcc') else \
              5 if (matrix == '470bg') else \
              6 if (matrix == '170m') else \
              7 if (matrix == '240m') else \
              8 if (matrix == 'ycgco') else \
              9 if (matrix == '2020ncl') else \
             10 if (matrix == '2020cl') else -1

        if (matrix is not None) and isinstance(matrix, int):

            M = \
              1 if (matrix == 1) else \
              4 if (matrix == 4) else \
              5 if (matrix == 5) else \
              6 if (matrix == 6) else \
              7 if (matrix == 7) else \
              8 if (matrix == 8) else \
              9 if (matrix == 9) else \
             10 if (matrix == 10) else -1

        if (matrix is not None) and (M == -1):
            raise vs.Error('RGBColor: Unsupported matrix specified')

        MP = clip.get_frame(0).props.get("_Matrix", None)

        if (MP is not None) and (MP > 10):
            raise vs.Error('RGBColor: Video has an unsupported value (' + str(MP) + \
                           ') for "_Matrix" in frame properties')

        MatrixProp = MP if (MP is None) else MP if (MP == 1) or (4 <= MP <= 10) else None

        if (matrix is not None) and (MatrixProp is not None) and (M != MatrixProp):
            raise vs.Error('RGBColor: The value for "_Matrix" (' + str(MatrixProp) + \
                           ') in frame properties doesn\'t match the specified matrix')

        matrix = M if (matrix is not None) else \
          MatrixProp if (MatrixProp is not None) else \
          5 if (clip.width <= 1024) and (clip.height <= 576) else \
          1 if (clip.width <= 1920) and (clip.height <= 1080) else 9

    range = range if (range is None) else range.lower().strip()
    range = range if (range is None) else \
      0 if (range == 'full') or (range == 'f') else \
      1 if (range == 'limited') or (range == 'l') else None

    RP = clip.get_frame(0).props.get("_ColorRange", None)
    RangeProp = RP if (RP is None) else RP if (RP == 0) or (RP == 1) else None

    if (range is not None) and (RangeProp is not None) and (range != RangeProp):
        raise vs.Error('RGBColor: The value for "_ColorRange" (' + str(RangeProp) + \
                         ') in frame properties doesn\'t match the specified range')

    range = range if (range is not None) else RangeProp

    BlankRGBClip = core.std.BlankClip(color=(v0, v1, v2))
    ColorClip = BlankRGBClip.resize.Point(format=clip.format.id, matrix=matrix, range=range)

    f = ColorClip.get_frame(0)
    p0 = f[0][0,0]

    return p0 if (clip.format.color_family == vs.GRAY) else (p0, f[1][0,0], f[2][0,0])

# -------------------------------------------------------------------------------
# -------------------------------------------------------------------------------
#            Kill Globals
# -------------------------------------------------------------------------------
# -------------------------------------------------------------------------------

def GKill():

    global \
    G1_OutWidth, G1_OutHeight, G1_CL, G1_CR, G1_CT, G1_CB, \
    G2_OutWidth, G2_OutHeight, G2_CL, G2_CR, G2_CT, G2_CB, \
    G1_CropDAR, G1_CSplit, G1_CPreview, G1_CLine, \
    G2_CropDAR, G2_CSplit, G2_CPreview, G2_CLine, \
    G1_InDAR, G1_InSAR, G1_OutDAR, G1_OutSAR, G1_AutoDAR, \
    G2_InDAR, G2_InSAR, G2_OutDAR, G2_OutSAR, G2_AutoDAR, \
    G1_Mod, G1_HMod, G1_NoResize, G1_ResizeWO, \
    G2_Mod, G2_HMod, G2_NoResize, G2_ResizeWO, \
    G1_Borders, G1_CleanBorders, G1_BColor, G1_Frosty, G1_ColorConvert, G1_ColorMode, \
    G2_Borders, G2_CleanBorders, G2_BColor, G2_Frosty, G2_ColorConvert, G2_ColorMode, \
    G1_Resizer, G1_RStr, G1_RingRepair, G1_Position, G1_Info, G1_PriorG1, \
    G2_Resizer, G2_RStr, G2_RingRepair, G2_Position, G2_Info, G2_PriorG2, \
    G_CPreview, G_Info, G_Resizer, G_RStr

    G1_OutWidth = G1_OutHeight = G1_CL = G1_CR = G1_CT = G1_CB = G1_CropDAR = \
    G2_OutWidth = G2_OutHeight = G2_CL = G2_CR = G2_CT = G2_CB = G2_CropDAR = \
    G1_CPreview = G1_InDAR = G1_InSAR = G1_OutDAR = G1_OutSAR = G1_RingRepair = \
    G2_CPreview = G2_InDAR = G2_InSAR = G2_OutDAR = G2_OutSAR = G2_RingRepair = \
    G_CPreview = 0

    G1_Mod = G1_HMod = \
    G2_Mod = G2_HMod = 4

    G1_CSplit = G1_CLine = \
    G2_CSplit = G2_CLine = 1

    G1_AutoDAR = G1_NoResize = G1_ResizeWO = G1_Borders = G1_CleanBorders = \
    G2_AutoDAR = G2_NoResize = G2_ResizeWO = G2_Borders = G2_CleanBorders = \
    G1_Frosty = G1_Position = G1_Info = G1_PriorG1 = \
    G2_Frosty = G2_Position = G2_Info = G2_PriorG2 = \
    G_Info = False

    G1_BColor = G2_BColor = None

    G1_ColorConvert = G1_ColorMode = G1_Resizer = G1_RStr = G_Resizer = G_RStr = \
    G2_ColorConvert = G2_ColorMode = G2_Resizer = G2_RStr = ""

# -------------------------------------------------------------------------------

def Killpix():
    global G_CPreview, G_Info, G_Resizer, G_RStr
    G_CPreview = 0
    G_Info = False
    G_Resizer = ""
    G_RStr = ""

# -------------------------------------------------------------------------------
# -------------------------------------------------------------------------------
#            Ringing Repair
# -------------------------------------------------------------------------------
# -------------------------------------------------------------------------------

def CR_RingRepair(Resized, Original, CRVideo_Width, CRVideo_Height, \
    RCL, RCT, F_Cropped_Width, F_Cropped_Height, RingRepair):

    GaussVideo = Original.fmtc.resample(CRVideo_Width, CRVideo_Height, \
    sx=RCL, sy=RCT, sw=F_Cropped_Width, sh=F_Cropped_Height, kernel="gauss", a1=100)

    RepairedVideo = Resized.rgvs.Repair(GaussVideo, 1)
    return RepairedVideo if (RingRepair == 1) else Resized.std.Merge(RepairedVideo, RingRepair)

# -------------------------------------------------------------------------------
# -------------------------------------------------------------------------------
#            Format Strings
# -------------------------------------------------------------------------------
# -------------------------------------------------------------------------------

def CR_FS(Num, Limit=3, Keep=False):

    S = ("{:z02." + str(Limit) + "f}").format(Num)
    if (Limit > 0): S = S.rstrip("0")
    if S.endswith("."): S = S.removesuffix(".") if (Keep is False) else S + "0"
    return S

# -------------------------------------------------------------------------------
# -------------------------------------------------------------------------------
#            Float To Fraction
# -------------------------------------------------------------------------------
# -------------------------------------------------------------------------------

def CR_FTF(Num):

    Frac = " (" + str(fr.Fraction(Num).limit_denominator(max_denominator=10000)) + ")"
    return Frac.replace(")", ":1)") if (Frac.find("/") == -1) else Frac.replace("/", ":")

# -------------------------------------------------------------------------------
# -------------------------------------------------------------------------------
#            Position
# -------------------------------------------------------------------------------
# -------------------------------------------------------------------------------

def CR_Position(Source, FRateNum, FRateDen, LB1="", LB2="\n", Align=None):

    return core.std.FrameEval(Source, partial(CR_Pos, Source=Source, \
      FRateNum=FRateDen, FRateDen=FRateNum, LB1=LB1, LB2=LB2, Align=Align))

def CR_Pos(n, Source, FRateNum, FRateDen, LB1, LB2, Align):

    if (FRateNum > 0 < FRateDen):

        F_Position = n * FRateNum / FRateDen
        HH = m.floor(F_Position / 3600)
        MM = m.floor(F_Position / 60) % 60
        SS = m.floor(F_Position) % 60
        MS = m.floor((F_Position - m.floor(F_Position)) * 1000)
        All = LB1 + str(n) + LB2 + "{:02.0f}:{:02.0f}:{:02.0f}.{:03.0f}".format(HH, MM, SS, MS)

    else: All = LB1 + str(n)

    return core.text.Text(Source, All, alignment=Align)

# -------------------------------------------------------------------------------
# -------------------------------------------------------------------------------
#            Picture Mod
# -------------------------------------------------------------------------------
# -------------------------------------------------------------------------------

def CR_PicMod(Video_Width, Video_Height, Out=1):

    Width_Mod = \
      16 if (Video_Width % 16 == 0) else 8 if (Video_Width % 8 == 0) else \
      4 if (Video_Width % 4 == 0) else 2 if (Video_Width % 2 == 0)  else 1
    Height_Mod = \
      16 if (Video_Height % 16 == 0) else 8 if (Video_Height % 8 == 0) else \
      4 if (Video_Height % 4 == 0) else 2 if (Video_Height % 2 == 0) else 1

    return ("" if (Out == 1) else " ") + f" (Mod {Width_Mod}x{Height_Mod})"

# -------------------------------------------------------------------------------
# -------------------------------------------------------------------------------
#            Default & IsEqual Functions
# -------------------------------------------------------------------------------
# -------------------------------------------------------------------------------

CR_Default = lambda ArgValue, DefaultValue: DefaultValue if (ArgValue is None) else ArgValue
CR_IsEqual = lambda Val1, Val2: (abs(Val1 - Val2) < 0.00005)

# ===============================================================================
# ===============================================================================
#            CropResize Defaults
# ===============================================================================
# ===============================================================================

# Setting these defaults incorrectly will probably break something.
# The functions below are required and cannot be commented out.

def CR_AutoDAR(): return False
def CR_CLine(W, H): return 1 if ((W <= 1920) and (H <= 1080)) else 2
def CR_ColorConvert(): return ""
def CR_CPreview(): return 1

# Default resizer functions
def CR_Resizer_Default(): return ""
def CR_RStr_Default(): return ""

# Alternative resizer functions
def CR_Resizer_x(): return ""
def CR_RStr_x(): return ""

# ===============================================================================
# ===============================================================================
